package sra.jun.collections.sequences;

import sra.smalltalk.*;

/**
 * JunDoubleMatrix class
 * 
 *  @author    nisinaka
 *  @created   1998/10/09 (by nisinaka)
 *  @updated   2002/11/01 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunDoubleMatrix.java,v 6.0 2003/10/28 10:52:02 nisinaka Exp $
 */
public class JunDoubleMatrix extends JunMatrix {
	protected static StSymbol DefalutInverseCalculationMethod = $("GaussElimination");

	/**
	 * Create a new instance of matrix.
	 * 
	 * @param squareSize int
	 */
	public JunDoubleMatrix(int squareSize) {
		super(squareSize);
	}

	/**
	 * Create a new instance of matrix and initialize it with collection.
	 * 
	 * @param squareSize int
	 * @param collection java.lang.Object[]
	 */
	public JunDoubleMatrix(int squareSize, double[] collection) {
		super(squareSize);
		this.collection_(collection);
	}

	/**
	 * Create a new instance of matrix.
	 * 
	 * @param rowSize int
	 * @param columnSize int
	 */
	public JunDoubleMatrix(int rowSize, int columnSize) {
		super(rowSize, columnSize);
	}

	/**
	 * Create a new instance of matrix and initialize it with collection.
	 * 
	 * @param rowSize int
	 * @param columnSize int
	 * @param collection double[]
	 */
	public JunDoubleMatrix(int rowSize, int columnSize, double[] collection) {
		super(rowSize, columnSize);
		this.collection_(collection);
	}

	/**
	 * Set the default inverse calculation method to 'Gauss Elimination'.
	 */
	public static void CalculateInverseByGaussElimination() {
		DefaultInverseCalculationMethod_($("GaussElimination"));
	}

	/**
	 * Set the default inverse calculation method to 'LU Decomposition'.
	 */
	public static void CalculateInverseByLuDecomposition() {
		DefaultInverseCalculationMethod_($("LuDecomposition"));
	}

	/**
	 * Answer the current default method for calculating inverse.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public static StSymbol DefaultInverseCalculationMethod() {
		return DefalutInverseCalculationMethod;
	}

	/**
	 * Create a new unit matrix.
	 * 
	 * @param aClass java.lang.Class
	 * @param squareSize int
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public static JunDoubleMatrix Unit_(Class aClass, int squareSize) {
		JunDoubleMatrix unitMatrix = (JunDoubleMatrix) _New(aClass, new Integer(squareSize));

		for (int i = 0; i < squareSize; i++) {
			for (int j = 0; j < squareSize; j++) {
				unitMatrix._put(i, j, new Double((i == j) ? 1 : 0));
			}
		}

		return unitMatrix;
	}

	/**
	 * Set the new default method for calculating inverse.
	 * 
	 * @param newDefalutInverseCalculationMethod sra.smalltalk.StSymbol
	 */
	protected static void DefaultInverseCalculationMethod_(StSymbol newDefalutInverseCalculationMethod) {
		DefalutInverseCalculationMethod = newDefalutInverseCalculationMethod;
	}

	/**
	 * Answer the value at the specified place as double.
	 * 
	 * @param rowNumber int
	 * @param columnNumber int
	 * 
	 * @return double
	 */
	public double _doubleValueAt(int rowNumber, int columnNumber) {
		return ((Double) this._at(rowNumber, columnNumber)).doubleValue();
	}

	/**
	 * Put a double value at the specified place.
	 * 
	 * @param rowNumber int
	 * @param columnNumber int
	 * @param value double
	 */
	public void _doubleValuePut(int rowNumber, int columnNumber, double value) {
		super._put(rowNumber, columnNumber, new Double(value));
	}

	/**
	 * Answer the sum of this matrix and aMatrix. This method corresponds to
	 * the '+' method in Smalltalk.
	 * 
	 * @param aMatrix sra.jun.collections.sequences.JunDoubleMatrix
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix _plus(JunDoubleMatrix aMatrix) {
		JunDoubleMatrix newMatrix = (JunDoubleMatrix) this.copy();

		for (int i = 0; i < rowSize; i++) {
			for (int j = 0; j < columnSize; j++) {
				newMatrix._doubleValuePut(i, j, (newMatrix._doubleValueAt(i, j) + aMatrix._doubleValueAt(i, j)));
			}
		}

		return newMatrix;
	}

	/**
	 * Convert this matrix as an array of double.
	 * 
	 * @return double[]
	 */
	public double[] asArrayOfDouble() {
		double[] anArray = new double[rowSize * columnSize];
		int index = 0;

		for (int i = 0; i < rowSize; i++) {
			for (int j = 0; j < columnSize; j++) {
				anArray[index] = this._doubleValueAt(i, j);
				index++;
			}
		}

		return anArray;
	}

	/**
	 * Set all elements of the matrix from 'collection'.
	 * 
	 * @param collection double[]
	 */
	public void collection_(double[] collection) {
		int index = 0;

		for (int i = 0; i < rowSize; i++) {
			for (int j = 0; j < columnSize; j++) {
				_matrix[i][j] = new Double(collection[index]);
				index++;
			}
		}
	}

	/**
	 * Answer the inverse of this matrix.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix inverse() {
		if (this.inverseCalculationMethod() == $("GaussElimination")) {
			return this.inverseByGaussElimination();
		} else if (this.inverseCalculationMethod() == $("LuDecomposition")) {
			return this.inverseByLuDecomposition();
		}
		return this.inverseByGaussElimination();
	}

	/**
	 * Calculate the inverse of this matrix by Gauss Elimination method.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix inverseByGaussElimination() {
		if (!this.isSquare()) {
			throw SmalltalkException.Error("the matrix is not square.");
		}

		JunDoubleMatrix copy = (JunDoubleMatrix) this.copy();
		JunDoubleMatrix inverse = Unit_(JunDoubleMatrix.class, rowSize);

		for (int i = 0; i < columnSize; i++) {
			double value = copy._doubleValueAt(i, i);

			if (value == 0) {
				int x = (i >= rowSize) ? (i - 1) : i + 1;
				value = copy._doubleValueAt(x, i);

				if (value != 0) {
					value = 1 / value;
				}

				final int i_ = i;
				final int x_ = x;
				final double value_ = value;
				final JunDoubleMatrix copy_ = copy;
				copy.rowIndex_doJ_(i, new StBlockClosure() {
					public Object value_value_(Object anElement, Object anIndex) {
						double each = ((Double) anElement).doubleValue();
						int j = ((Integer) anIndex).intValue();
						double v = each + (copy_._doubleValueAt(x_, j) * value_);
						copy_._doubleValuePut(i_, j, v);

						return null;
					}
				});

				final JunDoubleMatrix inverse_ = inverse;
				inverse.rowIndex_doJ_(i, new StBlockClosure() {
					public Object value_value_(Object anElement, Object anIndex) {
						double each = ((Double) anElement).doubleValue();
						int j = ((Integer) anIndex).intValue();
						double v = each + (inverse_._doubleValueAt(x_, j) * value_);
						inverse_._doubleValuePut(i_, j, v);

						return null;
					}
				});
			} else {
				final int i_ = i;
				final double value_ = value;
				final JunDoubleMatrix copy_ = copy;
				copy.rowIndex_doJ_(i, new StBlockClosure() {
					public Object value_value_(Object anElement, Object anIndex) {
						double v = ((Double) anElement).doubleValue() / value_;
						copy_._doubleValuePut(i_, ((Integer) anIndex).intValue(), v);

						return null;
					}
				});

				final JunDoubleMatrix inverse_ = inverse;
				inverse.rowIndex_doJ_(i, new StBlockClosure() {
					public Object value_value_(Object anElement, Object anIndex) {
						double v = ((Double) anElement).doubleValue() / value_;
						inverse_._doubleValuePut(i_, ((Integer) anIndex).intValue(), v);

						return null;
					}
				});
			}

			for (int k = 0; k < rowSize; k++) {
				if (k != i) {
					final double value_ = copy._doubleValueAt(k, i);
					final int i_ = i;
					final int k_ = k;
					final JunDoubleMatrix copy_ = copy;
					copy.rowIndex_doJ_(k, new StBlockClosure() {
						public Object value_value_(Object anElement, Object anIndex) {
							double each = ((Double) anElement).doubleValue();
							int j = ((Integer) anIndex).intValue();
							double v = each - (copy_._doubleValueAt(i_, j) * value_);
							copy_._doubleValuePut(k_, j, v);

							return null;
						}
					});

					final JunDoubleMatrix inverse_ = inverse;
					inverse.rowIndex_doJ_(k, new StBlockClosure() {
						public Object value_value_(Object anElement, Object anIndex) {
							double each = ((Double) anElement).doubleValue();
							int j = ((Integer) anIndex).intValue();
							double v = each - (inverse_._doubleValueAt(i_, j) * value_);
							inverse_._doubleValuePut(k_, j, v);

							return null;
						}
					});
				}
			}
		}

		return inverse;
	}

	/**
	 * Calculate the inverse of this matrix by LU Decomposition method.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix inverseByLuDecomposition() {
		if (this.isSquare() == false) {
			throw SmalltalkException.Error("the matrix is not square.");
		}

		Object[] lu = this.luDecompose();
		if (lu == null) {
			throw SmalltalkException.Error("the matrix is singular.");
		}

		JunDoubleMatrix luMatrix = (JunDoubleMatrix) lu[0];
		int[] pivotSequence = (int[]) lu[1];
		int size = this.rowSize();
		JunDoubleMatrix inverse = new JunDoubleMatrix(size);
		for (int columnIndex = 0; columnIndex < size; columnIndex++) {
			for (int i = 0; i < size; i++) {
				int index = pivotSequence[i];
				double value = (index == columnIndex) ? 1 : 0;
				for (int j = 0; j < i; j++) {
					value -= (luMatrix._doubleValueAt(index, j) * inverse._doubleValueAt(j, columnIndex));
				}
				inverse._doubleValuePut(i, columnIndex, value);
			}
			for (int i = size - 1; i >= 0; i--) {
				int index = pivotSequence[i];
				double value = inverse._doubleValueAt(i, columnIndex);
				for (int j = i + 1; j < size; j++) {
					value -= (luMatrix._doubleValueAt(index, j) * inverse._doubleValueAt(j, columnIndex));
				}
				inverse._doubleValuePut(i, columnIndex, value / luMatrix._doubleValueAt(index, i));
			}
		}

		return inverse;
	}

	/**
	 * Answer the current method for calculating inverse.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public StSymbol inverseCalculationMethod() {
		return DefaultInverseCalculationMethod();
	}

	/**
	 * Calculate a product element at i & j with aMatrix.
	 * 
	 * @param i int
	 * @param j int
	 * @param aMatrix sra.jun.collections.sequences.JunMatrix
	 * 
	 * @return java.lang.Object
	 */
	protected Object _calculateProductElementAt(int i, int j, JunMatrix aMatrix) {
		final int j_ = j;
		final JunDoubleMatrix aMatrix_ = (JunDoubleMatrix) aMatrix;
		final StValueHolder answerHolder_ = new StValueHolder(0);
		StBlockClosure aBlock = new StBlockClosure() {
			public Object value_value_(Object value, Object k) {
				double answer = answerHolder_._doubleValue();
				double myValue = ((Double) value).doubleValue();
				double otherValue = aMatrix_._doubleValueAt(((Integer) k).intValue(), j_);
				answer = answer + (myValue * otherValue);
				answerHolder_.value_(answer);

				return null;
			}
		};

		this.rowIndex_doJ_(i, aBlock);

		return answerHolder_.value();
	}

	/**
	 * Calculate the receiver's LU matrix.
	 * 
	 * @return java.lang.Object[]
	 */
	protected Object[] luDecompose() {
		if (this.isSquare() == false) {
			throw SmalltalkException.Error("the matrix is not square.");
		}

		double value;
		double maxValue;
		int size = this.rowSize();
		double[] weights = new double[size];
		int[] pivotSequence = new int[size];
		JunDoubleMatrix luMatrix = (JunDoubleMatrix) this.copy();
		double determinant = 1;
		for (int i = 0; i < size; i++) {
			maxValue = 0;
			for (int j = 0; j < size; j++) {
				value = Math.abs(luMatrix._doubleValueAt(i, j));
				if (maxValue < value) {
					maxValue = value;
				}
			}
			if (maxValue == 0) {
				return null;
			}
			weights[i] = 1 / maxValue;
			pivotSequence[i] = i;
		}

		int maxIndex = -1;
		for (int columnIndex = 0; columnIndex < size; columnIndex++) {
			maxValue = -1;
			for (int i = columnIndex; i < size; i++) {
				int index = pivotSequence[i];
				value = Math.abs(luMatrix._doubleValueAt(index, columnIndex)) * weights[index];
				if (value > maxValue) {
					maxValue = value;
					maxIndex = i;
				}
			}
			int pivotIndex = pivotSequence[maxIndex];
			if (maxIndex != columnIndex) {
				pivotSequence[maxIndex] = pivotSequence[columnIndex];
				pivotSequence[columnIndex] = pivotIndex;
				determinant = -determinant;
			}
			double pivot = luMatrix._doubleValueAt(pivotIndex, columnIndex);
			if (pivot == 0) {
				return null;
			}
			determinant *= pivot;
			for (int i = columnIndex + 1; i < size; i++) {
				value = luMatrix._doubleValueAt(pivotSequence[i], columnIndex);
				value /= pivot;
				luMatrix._doubleValuePut(pivotSequence[i], columnIndex, value);
				for (int j = columnIndex + 1; j < size; j++) {
					luMatrix._doubleValuePut(pivotSequence[i], j, luMatrix._doubleValueAt(pivotSequence[i], j) - (value * luMatrix._doubleValueAt(pivotIndex, j)));
				}
			}
		}

		return new Object[] { luMatrix, pivotSequence, new Double(determinant)};
	}
}
