package sra.jun.dxf;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import sra.jun.geometry.basic.*;
import sra.jun.opengl.display.*;
import sra.jun.opengl.objects.*;
import sra.jun.opengl.projection.JunOpenGLProjector;
import sra.smalltalk.*;

/**
 * JunCADModel class
 * 
 *  @author    nisinaka
 *  @created   2002/07/11 (by nisinaka)
 *  @updated   2003/05/15 (by nisinaka)
 *  @version   406b (with StPL6.0) based on Jun470 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunCADModel.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public class JunCADModel extends JunOpenGLDisplayModel {

	protected Hashtable layerNames;
	protected Vector hiddenObjects = null;
	protected StMenu _layersMenu;

	/**
	 * Menu message: hide all layers.
	 */
	public void allLayersHide() {
		this.toggleAllLayersOff();
		this.changed_($("object"));
	}

	/**
	 * Menu message: show all layers.
	 */
	public void allLayersShow() {
		this.toggleAllLayersOn();
		this.changed_($("object"));
	}

	/**
	 * Answer the default eye point. Assume 2d data, looking down Z axis.
	 * 
	 * @return sra.jun.opengl.objects.Jun3dPoint
	 */
	public Jun3dPoint defaultEyePoint() {
		if (this.openGL3dObject() == null) {
			return new Jun3dPoint(0, 0, 10000);
		}

		Jun3dPoint defaultEyePoint = (Jun3dPoint) this.defaultProjectionTable().get($("eyePoint"));
		if (defaultEyePoint == null) {
			Jun3dBoundingBox box = this.boundingBox();
			double distance = box.depth();
			distance = Math.max(distance * 2, 10000);
			defaultEyePoint = this.defaultSightPoint().plus_(new Jun3dPoint(0, 0, distance));
		}

		return defaultEyePoint;
	}

	/**
	 * Answer the default up vector. Assume 2d data, looking down Z axis.
	 * 
	 * @return sra.jun.opengl.objects.Jun3dPoint
	 */
	public Jun3dPoint defaultUpVector() {
		Jun3dPoint defaultUpVector = (Jun3dPoint) this.defaultProjectionTable().get($("upVector"));
		if (defaultUpVector == null) {
			defaultUpVector = new Jun3dPoint(0, 1, 0);
		}
		return defaultUpVector;
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		JunOpenGLDisplayView defaultView = (JunOpenGLDisplayView) super.defaultView();

		((Component) defaultView.getOpenGLView()).setBackground(Color.black);

		return defaultView;
	}

	/**
	 * Answer my display projector.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjector
	 */
	public JunOpenGLProjector displayProjector() {
		if (openGLProjector == null) {
			openGLProjector = new JunOpenGLProjector();
			openGLProjector.projection_(openGLProjector.projection().asParallelProjection());
			openGLProjector.eyePoint_(this.defaultEyePoint());
			openGLProjector.sightPoint_(this.defaultSightPoint());
			openGLProjector.upVector_(this.defaultUpVector());
			openGLProjector.viewFactor_(this.defaultViewFactor());
		}

		return openGLProjector;
	}

	/**
	 * Answer the current layer names.
	 * 
	 * @return java.util.Hashtable
	 */
	public Hashtable layerNames() {
		return layerNames;
	}

	/**
	 * Set the new layer names.
	 * 
	 * @param newLayerNames java.util.Hashtable
	 */
	public void layerNames_(Hashtable newLayerNames) {
		layerNames = newLayerNames;
	}

	/**
	 * Answer the names of the layers as an array of String.
	 * 
	 * @return java.lang.String[]
	 */
	public String[] layerNameStrings() {
		StBlockClosure sortBlock = new StBlockClosure() {
			public Object value_value_(Object o1, Object o2) {
				return new Boolean(((String) o1).compareTo(o2) < 0);
			}
		};

		StSortedCollection sortedCollection = new StSortedCollection(sortBlock);

		Enumeration keys = this.layerNames().keys();

		while (keys.hasMoreElements()) {
			sortedCollection.add_(keys.nextElement());
		}

		Object[] anArray = sortedCollection._asArray();
		String[] names = new String[anArray.length];
		System.arraycopy(anArray, 0, names, 0, anArray.length);

		return names;
	}

	/**
	 * Menu message: open a DXF file.
	 */
	public void openDXF() {
		File aFile = StDialog.OpenFileDialog_("Please select a DXF file ");

		if (aFile == null) {
			return;
		}

		Hashtable layers = this.readFromDXF_(aFile);

		if (layers == null) {
			return;
		}

		this.setLayerNames_(layers.keys());
		this.setLayerObjects_(layers.elements());
		this.addLayerMenuItems();
		this.resetView();

		if (this.showModel() != null) {
			this.showModel().resetView();
		}
	}

	/**
	 * Show the layers of the parser.
	 * 
	 * @param aParser sra.jun.dxf.JunDXFParser
	 */
	public void openParser(JunDXFParser aParser) {
		Hashtable layers = aParser.layers();

		if (layers == null) {
			return;
		}

		this.setLayerNames_(layers.keys());
		this.setLayerObjects_(layers.elements());
		this.addLayerMenuItems();
		this.resetView();

		if (this.showModel() != null) {
			this.showModel().resetView();
		}
	}

	/**
	 * Answer the default layers menu for JunCADModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultLayersMenu() {
		StMenu defaultLayersMenu = new StMenu("Layers");

		// Show all
		defaultLayersMenu.add(createMenuItem("Show all", "allLayersShow"));

		// Hide all
		defaultLayersMenu.add(createMenuItem("Hide all", "allLayersHide"));

		return defaultLayersMenu;
	}

	/**
	 * Answer my layers menu.
	 *
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _layersMenu() {
		if (_layersMenu == null) {
			_layersMenu = this._defaultLayersMenu();
		}
		return _layersMenu;
	}

	/**
	 * Set the 'Layers' menu to be check-marked lines with layer names  so the
	 * user can turn individual layers on and off.
	 */
	protected void addLayerMenuItems() {
		StMenu layersSubMenu = new StMenu("Layers");
		String[] layerNameStrings = this.layerNameStrings();

		for (int i = 0; i < layerNameStrings.length; i++) {
			String ss = layerNameStrings[i];
			StCheckBoxMenuItem theMenuItem = new StCheckBoxMenuItem(ss, true);
			theMenuItem.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent e) {
					toggleLayer_(((StCheckBoxMenuItem) e.getItem()).getLabel());
				}
			});
			layersSubMenu.add(theMenuItem);
		}

		this._layersMenu().add(layersSubMenu);

		this.menuModel().changed_with_($("added"), layersSubMenu);
	}

	/**
	 * Create the menubar for JunCADModel.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();
		rootMenu.add(this._fileMenu());
		rootMenu.add(this._editMenu());
		rootMenu.add(this._viewMenu());
		rootMenu.add(this._lightMenu());
		rootMenu.add(this._layersMenu());
		rootMenu.add(this._miscMenu());
		return rootMenu;
	}

	/**
	 * Find object in the cache.
	 * 
	 * @param aString java.lang.String
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject hiddenLayerAt_(String aString) {
		Vector obs = this.hiddenObjects();

		for (int i = 0; i < obs.size(); i++) {
			JunOpenGL3dObject x = (JunOpenGL3dObject) obs.get(i);

			if (x.name().equals(aString)) {
				return x;
			}
		}

		return null;
	}

	/**
	 * Cache the layers which have been hidden, in case they get turned on
	 * again.
	 * 
	 * @return java.util.Vector
	 */
	protected Vector hiddenObjects() {
		if (hiddenObjects == null) {
			hiddenObjects = new Vector();
		}

		return hiddenObjects;
	}

	/**
	 * Find object in the layers list.
	 * 
	 * @param aString java.lang.String
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject layerAt_(String aString) {
		JunOpenGL3dObject[] obs = ((JunOpenGL3dCompoundObject) this.displayObject()).components();

		for (int i = 0; i < obs.length; i++) {
			if (obs[i].name().equals(aString)) {
				return obs[i];
			}
		}

		return null;
	}

	/**
	 * Answer the final menu with all the layer labels.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu layersMenu() {
		StMenu daMenu = this._layersMenu();
		if (daMenu == null) {
			return null;
		}

		return (StMenu) daMenu.getItem("Layers");
	}

	/**
	 * Read the DXF file and create layers.
	 * 
	 * @param aFile java.io.File
	 * 
	 * @return java.util.Hashtable
	 */
	protected Hashtable readFromDXF_(File aFile) {
		Hashtable layers = null;

		if (aFile == null) {
			return null;
		}

		JunDXFParser parser = JunDXFParser.OnSmart_(aFile);

		if (parser.parsedOk()) {
			layers = parser.layers();
		}

		return layers;
	}

	/**
	 * Set the layer names. Keep track of which layers are displayed. boolean
	 * values are not currently used...
	 * 
	 * @param names java.util.Enumeration
	 */
	protected void setLayerNames_(Enumeration names) {
		Hashtable table = new Hashtable();

		while (names.hasMoreElements()) {
			table.put(names.nextElement(), Boolean.TRUE);
		}

		this.layerNames_(table);
	}

	/**
	 * Set the layer objects.
	 * 
	 * @param objects java.util.Enumeration
	 */
	protected void setLayerObjects_(Enumeration objects) {
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();

		while (objects.hasMoreElements()) {
			compoundObject.add_((JunOpenGL3dObject) objects.nextElement());
		}

		this.displayObject_(compoundObject);
	}

	/**
	 * Hide all layers.
	 */
	protected void toggleAllLayersOff() {
		String[] names = this.layerNameStrings();

		for (int i = 0; i < names.length; i++) {
			this.toggleLayerOff_(names[i]);
		}
	}

	/**
	 * Show all layers.
	 */
	protected void toggleAllLayersOn() {
		String[] names = this.layerNameStrings();

		for (int i = 0; i < names.length; i++) {
			this.toggleLayerOn_(names[i]);
		}
	}

	/**
	 * Toggle the layer on and off.
	 * 
	 * @param aString java.lang.String
	 */
	protected void toggleLayer_(String aString) {
		JunOpenGL3dObject layer = this.layerAt_(aString);

		if (layer == null) {
			this.toggleLayerOn_(aString);
		} else {
			this.toggleLayerOff_(aString);
		}

		this.changed_($("object"));
	}

	/**
	 * Change the checkmark in the Layers menu.
	 * 
	 * @param aString java.lang.String
	 * @param beOn boolean
	 */
	protected void toggleLayerMenu_(String aString, boolean beOn) {
		StMenu layersMenu = this.layersMenu();
		if (layersMenu == null) {
			return;
		}

		StCheckBoxMenuItem mItem = (StCheckBoxMenuItem) layersMenu.getItem(aString);
		if (mItem != null) {
			if (beOn) {
				mItem.beOn();
			} else {
				mItem.beOff();
			}
		}
	}

	/**
	 * Hide the named layer by removing it from the display object and caching
	 * it.
	 * 
	 * @param aString java.lang.String
	 */
	protected void toggleLayerOff_(String aString) {
		JunOpenGL3dObject layer = this.layerAt_(aString);

		if (layer != null) {
			this.hiddenObjects().add(layer);
			((JunOpenGL3dCompoundObject) this.displayObject()).remove_(layer);
			this.toggleLayerMenu_(aString, false);
		}
	}

	/**
	 * Make the named layer visible again.
	 * 
	 * @param aString java.lang.String
	 */
	protected void toggleLayerOn_(String aString) {
		JunOpenGL3dObject layer = this.hiddenLayerAt_(aString);

		if (layer != null) {
			this.hiddenObjects().remove(layer);
			((JunOpenGL3dCompoundObject) this.displayObject()).add_(layer);
			this.toggleLayerMenu_(aString, true);
		}
	}

	/**
	 * Answer the window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "DXF Viewer";
	}
}
