package sra.jun.geometry.abstracts;

import sra.jun.geometry.basic.Jun2dPoint;
import sra.jun.geometry.basic.Jun3dPoint;
import sra.jun.geometry.coordinate.JunCoordinateSystem;
import sra.jun.geometry.coordinate.JunLocalGeometry;
import sra.jun.geometry.transformations.JunTransformation;
import sra.smalltalk.SmalltalkException;
import sra.smalltalk.StObject;

/**
 * JunGeometry class
 * 
 *  @author    nisinaka
 *  @created   1998/09/29 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on Jun508 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunGeometry.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public abstract class JunGeometry extends StObject {
	public static final double ACCURACY = 1.0e-12d;

	/**
	 * Answer the accuracy of geometries.
	 * 
	 * @return double
	 * 
	 * @category Constants access
	 */
	public static final double Accuracy() {
		return ACCURACY;
	}

	/**
	 * Answer the local geometry class.
	 * 
	 * @return java.lang.Class
	 * 
	 * @category Class accessing
	 */
	private static Class LocalGeometryClass() {
		return JunLocalGeometry.class;
	}

	/**
	 * Convert to a glocal geometry.
	 * 
	 * @return sra.jun.geometry.abstracts.JunGeometry
	 * 
	 * @category converting
	 */
	public JunGeometry asGlobalGeometry() {
		return this;
	}

	/**
	 * Convert this as a local geometry in aJunCoordinateSystem.
	 * 
	 * @param aCoordinateSystem sra.jun.geometry.coordinate.JunCoordinateSystem
	 * 
	 * @return sra.jun.geometry.coordinate.JunLocalGeometry
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 * 
	 * @category converting
	 */
	public JunLocalGeometry asLocalGeometryIn_(JunCoordinateSystem aCoordinateSystem) {
		try {
			JunGeometry localGeometry = aCoordinateSystem.interprete_(this);
			Class aClass = LocalGeometryClass();
			Object[] arguments = { aClass, localGeometry, aCoordinateSystem };

			return (JunLocalGeometry) _PerformWithArguments(aClass, "Geometry_coordinateSystem_", arguments);
		} catch (Exception e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Answer the distance from specified Point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 * 
	 * @category functions (java only)
	 */
	public double distanceFromPoint_(Jun2dPoint aPoint) {
		throw SmalltalkException.SubclassResponsibility();
	}

	/**
	 * Answer the distance from specified Point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 * 
	 * @category functions (java only)
	 */
	public double distanceFromPoint_(Jun3dPoint aPoint) {
		throw SmalltalkException.SubclassResponsibility();
	}

	/**
	 * Answer a new JunGeometry interpreted by aJunCoordinateSystem.
	 * 
	 * @param aJunCoordinateSystem
	 *        sra.jun.geometry.coordinate.JunCoordinateSystem
	 * 
	 * @return sra.jun.geometry.abstracts.JunGeometry
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 * 
	 * @category converting
	 */
	public JunGeometry interpretedBy_(JunCoordinateSystem aJunCoordinateSystem) {
		try {
			return (JunGeometry) this.perform_with_("transform_", aJunCoordinateSystem.globalInvTransformation());
		} catch (Exception e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Answer true if the receiver is a 2d geometry element.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean is2d() {
		return false;
	}

	/**
	 * Answer true if the receiver is a 3d geometry element.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean is3d() {
		return false;
	}

	/**
	 * Answer true if the receiver is an angle geometry element.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isAngle() {
		return false;
	}

	/**
	 * Answer true if the receiver is a curve.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isCurve() {
		return false;
	}

	/**
	 * Answer true if the receiver is a geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isGeometry() {
		return true;
	}

	/**
	 * Answer true if the receiver is a global geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isGlocalGeometry() {
		return true;
	}

	/**
	 * Answer true if the receiver is a line geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isLine() {
		return false;
	}

	/**
	 * Answer true if the receiver is a plane geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isPlane() {
		return false;
	}

	/**
	 * Answer true if the receiver is a point geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isPoint() {
		return false;
	}

	/**
	 * Answer true if the receiver is a polygon geometry.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isPolygon() {
		return false;
	}

	/**
	 * Answer true if the receiver is a surface.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isSurface() {
		return false;
	}

	/**
	 * Answer true if the receiver is a topology.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isTopology() {
		return false;
	}

	/**
	 * Answer the copy of the receiver which is applied the specified
	 * transformation.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.JunTransformation
	 * 
	 * @return sra.jun.geometry.abstracts.JunGeometry
	 * 
	 * @category transforming
	 */
	public JunGeometry transform_(JunTransformation aTransformation) {
		throw SmalltalkException.SubclassResponsibility();
	}
}
