package sra.jun.geometry.basic;

import java.awt.Point;
import java.io.IOException;
import java.io.Writer;

import sra.jun.geometry.transformations.Jun2dTransformation;

/**
 * Jun2dPoint class
 * 
 *  @author    nisinaka
 *  @created   1998/09/29 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on Jun508 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dPoint.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public class Jun2dPoint extends JunPoint {
	/** The <i>x</i> coordinate. */
	protected double x;

	/** The <i>y</i> coordinate. */
	protected double y;

	/**
	 * Create a new Jun2dPoint.
	 * 
	 * @param x double
	 * @param y double
	 * 
	 * @category Instance creation
	 */
	public Jun2dPoint(double x, double y) {
		super();
		this.x = x;
		this.y = y;
	}

	/**
	 * Create a new Jun2dPoint.
	 * 
	 * @param aPoint java.awt.Point
	 * 
	 * @category Instance creation
	 */
	public Jun2dPoint(Point aPoint) {
		super();
		this.x = aPoint.x;
		this.y = aPoint.y;
	}

	/**
	 * Create a new instance of Jun2dPoint and initialize it with two Numbers.
	 * 
	 * @param numberX java.lang.Number
	 * @param numberY java.lang.Number
	 * 
	 * @category Instance creation
	 */
	public Jun2dPoint(Number numberX, Number numberY) {
		this(numberX.doubleValue(), numberY.doubleValue());
	}

	/**
	 * Create a new instance of Jun2dPoint and initialize it with an array of
	 * double.
	 * 
	 * @param anArray double[]
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category Instance creation
	 */
	public static Jun2dPoint FromArray_(double[] anArray) {
		return new Jun2dPoint(anArray[0], anArray[1]);
	}

	/**
	 * Answer a new Jun2dPoint created as a copy of another point.
	 * 
	 * @param aPoint DOCUMENT ME!
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category Instance creation
	 */
	public static Jun2dPoint FromPoint_(Jun2dPoint aPoint) {
		return new Jun2dPoint(aPoint.x, aPoint.y);
	}

	/**
	 * Create a new Jun2dPoint with rho and theta.
	 * 
	 * @param rho double
	 * @param theta double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category Instance creation
	 */
	public static Jun2dPoint Rho_theta_(double rho, double theta) {
		return new Jun2dPoint(rho * Math.cos(theta), rho * Math.sin(theta));
	}

	/**
	 * Answer a new Jun2dPoint with unit values.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category Constants access
	 */
	public static Jun2dPoint Unity() {
		return new Jun2dPoint(1.0d, 1.0d);
	}

	/**
	 * Answer a new Jun2dPoint with zero values.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category Constants access
	 */
	public static Jun2dPoint Zero() {
		return new Jun2dPoint(0.0d, 0.0d);
	}

	/**
	 * Answer the receiver as a java.awt.point.
	 * 
	 * @return java.awt.Point
	 * 
	 * @category converting
	 */
	public Point _toPoint() {
		Jun2dPoint aPoint = this.rounded();
		return new Point((int) aPoint.x(), (int) aPoint.y());
	}

	/**
	 * Answer the absolute Point.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint abs() {
		return new Jun2dPoint(Math.abs(x), Math.abs(y));
	}

	/**
	 * Answer an angle made with three points (including me).
	 * 
	 * @param aPoint1 sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint2 sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category functions
	 */
	public JunAngle angleBetween_and_(Jun2dPoint aPoint1, Jun2dPoint aPoint2) {
		Jun2dPoint v1 = (Jun2dPoint) aPoint1.minus_(this);
		Jun2dPoint v2 = (Jun2dPoint) aPoint2.minus_(this);
		double r1 = v1.rho();
		double r2 = v2.rho();

		if ((r1 < ACCURACY) || (r2 < ACCURACY)) {
			return JunAngle.Zero();
		} else {
			double sin = (((this.y() + aPoint1.y()) * (this.x() - aPoint1.x())) + ((aPoint1.y() + aPoint2.y()) * (aPoint1.x() - aPoint2.x())) + (aPoint2.y() + this.y()) * (aPoint2.x() - this.x())) / r1 / r2;
			sin = Math.min(Math.max(sin, -1.0d), 1.0d);

			if (v1.dotProduct_(v2) >= 0) {
				return JunAngle.FromRad_(Math.asin(sin));
			} else if (Math.abs(sin) < ACCURACY) {
				return JunAngle.FromRad_(Math.PI);
			} else {
				int sign = (sin < 0) ? (-1) : 1;

				return JunAngle.FromRad_((Math.PI - Math.abs(Math.asin(sin))) * sign);
			}
		}
	}

	/**
	 * Answer the receiver as an array of double.
	 * 
	 * @return double[]
	 * 
	 * @category converting
	 */
	public double[] asArray() {
		return new double[] { x, y };
	}

	/**
	 * 
	 * @return Jun2dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun2dPoint ceiling() {
		return new Jun2dPoint(Math.ceil(x), Math.ceil(y));
	}

	/**
	 * Answer the bounding box with a point as the corner.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 * 
	 * @category compatibilities
	 */
	public Jun2dBoundingBox corner_(Jun2dPoint aPoint) {
		return Jun2dBoundingBox.Origin_corner_(this, aPoint);
	}

	/**
	 * Answer the distance from aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double distance_(Jun2dPoint aPoint) {
		return this.minus_(aPoint).rho();
	}

	/**
	 * Answer the result of dividing this point by delta. This method
	 * corresponds to the methd '/' in Smalltalk.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint dividedBy_(double delta) {
		return new Jun2dPoint(this.x / delta, this.y / delta);
	}

	/**
	 * Answer the result of dividing this point by aPoint. This method
	 * corresponds to the methd '/' in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint dividedBy_(Jun2dPoint aPoint) {
		return new Jun2dPoint(this.x / aPoint.x, this.y / aPoint.y);
	}

	/**
	 * Answer the result of a dot product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double dotProduct_(Jun2dPoint aPoint) {
		Jun2dPoint newPoint = this.multipliedBy_(aPoint);

		return newPoint.x + newPoint.y;
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun2dPoint == false) {
			return false;
		}

		Jun2dPoint aPoint = (Jun2dPoint) anObject;

		return (this.x == aPoint.x) && (this.y == aPoint.y);
	}

	/**
	 * Answer the bounding box with a point as the extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 * 
	 * @category compatibilities
	 */
	public Jun2dBoundingBox extent_(Jun2dPoint aPoint) {
		return Jun2dBoundingBox.Origin_extent_(this, aPoint);
	}

	/**
	 * 
	 * @return Jun2dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun2dPoint floor() {
		return new Jun2dPoint(Math.floor(x), Math.floor(y));
	}

	/**
	 * Answer the number representing the ordering of the receiver in the
	 * generality hierarchy.
	 * 
	 * @return int
	 * 
	 * @category coercing
	 */
	public int generality() {
		return 200;
	}

	/**
	 * 
	 * @param aPoint
	 * 
	 * @return Jun2dPoint
	 * 
	 * @category functions
	 */
	public Jun2dPoint grid_(Jun2dPoint aPoint) {
		double newX;
		double newY;

		if (aPoint.x() == 0.0d) {
			newX = 0.0d;
		} else {
			newX = (double) Math.round(this.x() / aPoint.x()) * aPoint.x();
		}
		if (aPoint.y() == 0.0d) {
			newY = 0.0d;
		} else {
			newY = (double) Math.round(this.y() / aPoint.y()) * aPoint.y();
		}
		return new Jun2dPoint(newX, newY);
	}

	/**
	 * Computes a hash code for this object.
	 * 
	 * @return int
	 * 
	 * @category hash function
	 */
	public int hashCode() {
		long bitsX = Double.doubleToLongBits(x);
		int xHash = (int) (bitsX ^ (bitsX >> 32));
		long bitsY = Double.doubleToLongBits(y);
		int yHash = (int) (bitsY ^ (bitsY >> 32));

		return ((xHash >> 2) | yHash);
	}

	/**
	 * Answer the result of the inner product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double innerProduct_(Jun2dPoint aPoint) {
		return this.dotProduct_(aPoint);
	}

	/**
	 * Answer true if this is a kind of 2D geometry object, otherwise false.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean is2d() {
		return true;
	}

	/**
	 * Answer true if the receiver is less than aPoint. This method corresponds
	 * to the method "&lt;" in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean isLessThan_(Jun2dPoint aPoint) {
		return (this.x < aPoint.x) && (this.y < aPoint.y);
	}

	/**
	 * Answer true if the receiver is less than or equal to aPoint. This method
	 * corresponds to the method "&lt;=" in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean isLessThanOrEqualTo_(Jun2dPoint aPoint) {
		return (this.x <= aPoint.x) && (this.y <= aPoint.y);
	}

	/**
	 * Answer true if this point is ZERO, otherwise false.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isZero() {
		return (this.x == 0.0d) && (this.y == 0.0d);
	}

	/**
	 * Set the length from the origin.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category functions
	 */
	public Jun2dPoint length_(double aNumber) {
		double length = this.length();

		if (length <= JunPoint.ACCURACY) {
			return this;
		}

		return this.multipliedBy_(aNumber / length);
	}

	/**
	 * Answer the maximum point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category comparing
	 */
	public Jun2dPoint max_(Jun2dPoint aPoint) {
		return new Jun2dPoint(Math.max(this.x, aPoint.x), Math.max(this.y, aPoint.y));
	}

	/**
	 * Answer the minimum point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category comparing
	 */
	public Jun2dPoint min_(Jun2dPoint aPoint) {
		return new Jun2dPoint(Math.min(this.x, aPoint.x), Math.min(this.y, aPoint.y));
	}

	/**
	 * Answer the difference between this point and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint minus_(double delta) {
		return new Jun2dPoint(this.x - delta, this.y - delta);
	}

	/**
	 * Answer the difference between this point and aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint minus_(Jun2dPoint aPoint) {
		return new Jun2dPoint(this.x - aPoint.x, this.y - aPoint.y);
	}

	/**
	 * Answer the result of multiplying this point by aPoint.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint multipliedBy_(double delta) {
		return new Jun2dPoint(this.x * delta, this.y * delta);
	}

	/**
	 * Answer the result of multiplying this point by aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint multipliedBy_(Jun2dPoint aPoint) {
		return new Jun2dPoint(this.x * aPoint.x, this.y * aPoint.y);
	}

	/**
	 * Answer the negation of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint negated() {
		return Jun2dPoint.Zero().minus_(this);
	}

	/**
	 * Answer the sum of this point and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint plus_(double delta) {
		return new Jun2dPoint(this.x + delta, this.y + delta);
	}

	/**
	 * Answer the sum of this point and aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint plus_(Jun2dPoint aPoint) {
		return new Jun2dPoint(this.x + aPoint.x, this.y + aPoint.y);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 * 
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("(");
		aWriter.write(String.valueOf(this.x));
		aWriter.write(" , ");
		aWriter.write(String.valueOf(this.y));
		aWriter.write(")");
	}

	/**
	 * Answer the product of the Jun2dPoint and the receiver.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double product_(Jun2dPoint aPoint) {
		return (x * aPoint.y) - (y * aPoint.x);
	}

	/**
	 * Answer 1 divided by the receiver.
	 * Provide an error notification if the receiver is 0.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun2dPoint reciprocal() {
		return Unity().dividedBy_(this);
	}

	/**
	 * Answer the rho as of a polar coordinate.
	 * 
	 * @return double
	 * 
	 * @category polar coordinates
	 */
	public double rho() {
		return Math.sqrt(this.dotProduct_(this));
	}

	/**
	 * Answer the new Jun2dPoint which is rotated by aJunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category transforming
	 */
	public Jun2dPoint rotatedBy_(JunAngle anAngle) {
		return this.transform_(Jun2dTransformation.Rotate_(anAngle));
	}

	/**
	 * Answer the rounded Jun2dPoint.
	 * 
	 * @return Jun2dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun2dPoint rounded() {
		return new Jun2dPoint((double) Math.round(this.x), (double) Math.round(this.y));
	}

	/**
	 * Answer the new Jun2dPoint which is scaled by aJunPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category transforming
	 */
	public Jun2dPoint scaledBy_(Jun2dPoint aPoint) {
		return this.transform_(Jun2dTransformation.Scale_(aPoint));
	}

	/**
	 * Answer the theta as of the polar coordinates.
	 * 
	 * @return double
	 * 
	 * @category polar coordinates
	 */
	public double theta() {
		if (x == 0.0d) {
			if (y == 0.0d) {
				return 0.0d;
			}

			if (y >= 0.0d) {
				return Math.PI * 0.5d;
			} else {
				return Math.PI * 1.5d;
			}
		}

		JunAngle theta = JunAngle.FromRad_(Math.atan(y / x));

		if (x >= 0) {
			if (y >= 0) {
				return theta.rad();
			} else {
				return theta.plus_(Math.PI * 2.0d).rad();
			}
		} else {
			return theta.plus_(Math.PI).rad();
		}
	}

	/**
	 * Apply a transformation 'aJunTransformation' to the receiver. This method
	 * is not defined in Smalltalk version.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.Jun2dTransformation
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category transforming
	 */
	public Jun2dPoint transform_(Jun2dTransformation aTransformation) {
		return aTransformation.applyToPoint_(this);
	}

	/**
	 * Answer the new Jun2dPoint which is translated by aJunPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category transforming
	 */
	public Jun2dPoint translatedBy_(Jun2dPoint aPoint) {
		return this.transform_(Jun2dTransformation.Translate_(aPoint));
	}

	/**
	 * Answer the truncated Jun2dPoint.
	 * 
	 * @return Jun2dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun2dPoint truncated() {
		double newX;
		double newY;
		newX = (this.x >= 0) ? Math.ceil(this.x) : Math.floor(this.x);
		newY = (this.y >= 0) ? Math.ceil(this.y) : Math.floor(this.y);

		return new Jun2dPoint(newX, newY);
	}

	/**
	 * Answer the unit vector derived from this point.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @category functions
	 */
	public Jun2dPoint unitVector() {
		return this.length_(1.0d);
	}

	/**
	 * Answer the x coordinate.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double x() {
		return this.x;
	}

	/**
	 * Answer the y coordinate.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double y() {
		return this.y;
	}
}
