package sra.jun.geometry.basic;

import java.io.*;
import sra.jun.geometry.abstracts.*;
import sra.jun.geometry.transformations.Jun3dTransformation;

/**
 * Jun3dBoundingBox class
 * 
 *  @author    nisinaka
 *  @created   1998/10/01 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dBoundingBox.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public class Jun3dBoundingBox extends JunBoundingBox {
	/** The <i>origin</i> point. */
	protected Jun3dPoint origin;

	/** The <i>corner</i> point. */
	protected Jun3dPoint corner;

	/**
	 * Create a new instance of Jun3dBoundingBox and initialize it with the
	 * <code>origin</code> point and the <code>corner</code> point.
	 * 
	 * @param originPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param cornerPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category Instance creation
	 */
	public static Jun3dBoundingBox Origin_corner_(Jun3dPoint originPoint, Jun3dPoint cornerPoint) {
		Jun3dBoundingBox aBoundingBox = new Jun3dBoundingBox();
		aBoundingBox.origin_corner_(originPoint, cornerPoint);

		return aBoundingBox;
	}

	/**
	 * Create a new instance of Jun3dBoundingBox and initialize it with the
	 * <code>origin</code> point and the <code>extent</code> point.
	 * 
	 * @param originPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param extentPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category Instance creation
	 */
	public static Jun3dBoundingBox Origin_extent_(Jun3dPoint originPoint, Jun3dPoint extentPoint) {
		Jun3dBoundingBox aBoundingBox = new Jun3dBoundingBox();
		aBoundingBox.origin_extent_(originPoint, extentPoint);

		return aBoundingBox;
	}

	/**
	 * Create a new instance of Jun3dBoundingBox and initialize it with the
	 * <code>vertex</code> points.
	 * 
	 * @param vertexPoint1 sra.jun.geometry.basic.Jun3dPoint
	 * @param vertexPoint2 sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category Instance creation
	 */
	public static Jun3dBoundingBox Vertex_vertex_(Jun3dPoint vertexPoint1, Jun3dPoint vertexPoint2) {
		Jun3dBoundingBox aBoundingBox = new Jun3dBoundingBox();
		aBoundingBox.vertex_vertex_(vertexPoint1, vertexPoint2);

		return aBoundingBox;
	}

	/**
	 * Answer the center point of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public Jun3dPoint center() {
		return new Jun3dPoint((this.origin.x + this.corner.x) / 2.0d, (this.origin.y + this.corner.y) / 2.0d, (this.origin.z + this.corner.z) / 2.0d);
	}

	/**
	 * Answer true if the receiver contains the bounding box.
	 * 
	 * @param boundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean contains_(Jun3dBoundingBox boundingBox) {
		return origin.isLessThanOrEqualTo_(boundingBox.origin()) && boundingBox.corner().isLessThanOrEqualTo_(corner);
	}

	/**
	 * Answer true if the receiver contains or touches aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean containsOrTouchesPoint_(Jun3dPoint aPoint) {
		return origin.isLessThanOrEqualTo_(aPoint) && aPoint.isLessThanOrEqualTo_(corner);
	}

	/**
	 * Answer true if the receiver contains aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean containsPoint_(Jun3dPoint aPoint) {
		return origin.isLessThanOrEqualTo_(aPoint) && aPoint.isLessThan_(corner);
	}

	/**
	 * Answer the <code>corner</code> point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public Jun3dPoint corner() {
		return this.corner;
	}

	/**
	 * Answer the depth of the receiver.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double depth() {
		return this.corner.z - this.origin.z;
	}

	/**
	 * Answer true if anObject is equal to this bounding box, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun3dBoundingBox == false) {
			return false;
		}

		Jun3dBoundingBox aBoundingBox = (Jun3dBoundingBox) anObject;

		return this.origin.equals(aBoundingBox.origin) && this.corner.equals(aBoundingBox.corner);
	}

	/**
	 * Answer a copy of this object expanded by "delta".
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox expandedBy_(double delta) {
		return Origin_corner_(this.origin.minus_(delta), this.corner.plus_(delta));
	}

	/**
	 * Answer a copy of this object expanded by "delta".
	 * 
	 * @param aBoundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox expandedBy_(Jun3dBoundingBox aBoundingBox) {
		return Origin_corner_(this.origin.minus_(aBoundingBox.origin), this.corner.plus_(aBoundingBox.corner));
	}

	/**
	 * Answer the extent of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public Jun3dPoint extent() {
		return corner.minus_(origin);
	}

	/**
	 * Answer the height of the receiver.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double height() {
		return corner.y - origin.y;
	}

	/**
	 * Answer a copy of this object inset by "delta".
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox insetBy_(double delta) {
		return Origin_corner_(this.origin.plus_(delta), this.corner.minus_(delta));
	}

	/**
	 * Answer a copy of this object inset by "delta".
	 * 
	 * @param aBoundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox insetBy_(Jun3dBoundingBox aBoundingBox) {
		return Origin_corner_(this.origin.plus_(aBoundingBox.origin), this.corner.minus_(aBoundingBox.corner));
	}

	/**
	 * Answer the intersection part with the Jun3dBoundingBox.
	 * 
	 * @param boundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox intersect_(Jun3dBoundingBox boundingBox) {
		return Origin_corner_(origin.max_(boundingBox.origin), corner.min_(boundingBox.corner));
	}

	/**
	 * Answer true if the receiver intersects with the Jun3dBoundingBox.
	 * 
	 * @param aBoundingBox sra.jun.geometry.basic.Jun2dBoundingBox
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean intersects_(Jun3dBoundingBox aBoundingBox) {
		return origin.isLessThan_(aBoundingBox.corner) && aBoundingBox.origin.isLessThan_(corner) && origin.isLessThan_(corner) && aBoundingBox.origin.isLessThan_(aBoundingBox.corner);
	}

	/**
	 * Answer true if the receiver intersects or touches with aBoundingBox.
	 * 
	 * @param aBoundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean intersectsOrTouches_(Jun3dBoundingBox aBoundingBox) {
		return this.intersects_(aBoundingBox) || this.touches_(aBoundingBox);
	}

	/**
	 * Answer true if the receiver is a 3d geometry element.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean is3d() {
		return true;
	}

	/**
	 * Merge the Jun3dBoundingBox and the receiver.
	 * 
	 * @param boundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category functions
	 */
	public Jun3dBoundingBox merge_(Jun3dBoundingBox boundingBox) {
		return Origin_corner_(this.origin.min_(boundingBox.origin), this.corner.max_(boundingBox.corner));
	}

	/**
	 * Answer the <code>origin</code> point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public Jun3dPoint origin() {
		return this.origin;
	}

	/**
	 * Set the <code>origin</code> point and the <code>corner</code> point.
	 * 
	 * @param originPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param cornerPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public void origin_corner_(Jun3dPoint originPoint, Jun3dPoint cornerPoint) {
		this.origin = originPoint;
		this.corner = cornerPoint;
	}

	/**
	 * Set the <code>origin</code> point and the <code>corner</code> point. The
	 * <code>corner</code> point is calculated with the <code>extent</code>.
	 * 
	 * @param originPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param extentPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public void origin_extent_(Jun3dPoint originPoint, Jun3dPoint extentPoint) {
		this.origin = originPoint;
		this.corner = originPoint.plus_(extentPoint);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 * 
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("(");
		aWriter.write(this.origin.printString());
		aWriter.write(" corner: ");
		aWriter.write(this.corner.printString());
		aWriter.write(")");
	}

	/**
	 * Answer the copy of the receiver rotated by the specified angle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.abstracts.JunGeometry
	 * 
	 * @category transforming
	 */
	public JunGeometry rotatedBy_(JunAngle anAngle) {
		return this.transform_(Jun3dTransformation.Rotate_(anAngle));
	}

	/**
	 * Answer the copy of the receiver scaled by the specified extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category transforming
	 */
	public Jun3dBoundingBox scaledBy_(Jun3dPoint aPoint) {
		return (Jun3dBoundingBox) this.transform_(Jun3dTransformation.Scale_(aPoint));
	}

	/**
	 * Answer true if the receiver touches with aBoundingBox.
	 * 
	 * @param aBoundingBox sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean touches_(Jun3dBoundingBox aBoundingBox) {
		return origin.isLessThanOrEqualTo_(aBoundingBox.corner) && aBoundingBox.origin.isLessThanOrEqualTo_(corner) && origin.isLessThanOrEqualTo_(corner) && aBoundingBox.origin.isLessThanOrEqualTo_(aBoundingBox.corner);
	}

	/**
	 * Answer the copy of the receiver translated by the specified extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category transforming
	 */
	public Jun3dBoundingBox translatedBy_(Jun3dPoint aPoint) {
		return (Jun3dBoundingBox) this.transform_(Jun3dTransformation.Translate_(aPoint));
	}

	/**
	 * Set the <code>origin</code> point and the <code>corner</code> point. Two
	 * points are compared and the smaller one will be the
	 * <code>origin</code>.
	 * 
	 * @param aPoint1 sra.jun.geometry.basic.Jun3dPoint
	 * @param aPoint2 sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category accessing
	 */
	public void vertex_vertex_(Jun3dPoint aPoint1, Jun3dPoint aPoint2) {
		this.origin = aPoint1.min_(aPoint2);
		this.corner = aPoint1.max_(aPoint2);
	}

	/**
	 * Answer my volume.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double volume() {
		return this.width() * this.height() * this.depth();
	}

	/**
	 * Answer the width of the receiver.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double width() {
		return this.corner.x - this.origin.x;
	}
}
