package sra.jun.geometry.basic;

import java.io.*;
import sra.jun.geometry.curves.Jun3dLine;
import sra.jun.geometry.transformations.Jun3dTransformation;

/**
 * Jun3dPoint class
 * 
 *  @author    nisinaka
 *  @created   1998/09/30 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dPoint.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public class Jun3dPoint extends JunPoint {
	/** The <i>x</i> coordinate. */
	protected double x;

	/** The <i>y</i> coordinate. */
	protected double y;

	/** The <i>z</i> coordinate. */
	protected double z;

	/**
	 * Create a new Jun3dPoint.
	 * 
	 * @param x double
	 * @param y double
	 * @param z double
	 * 
	 * @category Instance creation
	 */
	public Jun3dPoint(double x, double y, double z) {
		super();
		this.x = x;
		this.y = y;
		this.z = z;
	}

	/**
	 * Create a new instance of Jun3dPoint and initialize it with three
	 * Numbers.
	 * 
	 * @param numberX java.lang.Number
	 * @param numberY java.lang.Number
	 * @param numberZ java.lang.Number
	 * 
	 * @category Instance creation
	 */
	public Jun3dPoint(Number numberX, Number numberY, Number numberZ) {
		this(numberX.doubleValue(), numberY.doubleValue(), numberZ.doubleValue());
	}

	/**
	 * Create a new Jun3dPoint with the Jun2dPoint and z-coordinate.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param z double
	 * 
	 * @category Instance creation
	 */
	public Jun3dPoint(Jun2dPoint aPoint, double z) {
		this(aPoint.x(), aPoint.y(), z);
	}

	/**
	 * Create a new instance of Jun3dPoint and initialize it with a Jun2dPoint
	 * and a Number.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param numberZ java.lang.Number
	 * 
	 * @category Instance creation
	 */
	public Jun3dPoint(Jun2dPoint aPoint, Number numberZ) {
		this(aPoint, numberZ.doubleValue());
	}

	/**
	 * Answer the new Jun3dPoint which is the same kind of point as the
	 * receiver.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category Coercing
	 */
	public static Jun3dPoint Coerce_(Jun3dPoint aPoint) {
		return new Jun3dPoint(aPoint.x, aPoint.y, aPoint.z);
	}

	/**
	 * Create a new instance of Jun3dPoint and initialize it with an array of
	 * double.
	 * 
	 * @param anArray double[]
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category Instance creation
	 */
	public static Jun3dPoint FromArray_(double[] anArray) {
		return new Jun3dPoint(anArray[0], anArray[1], anArray[2]);
	}

	/**
	 * Answer a new Jun3dPoint created as a copy of another point.
	 * 
	 * @param aPoint DOCUMENT ME!
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category Instance creation
	 */
	public static Jun3dPoint FromPoint_(Jun3dPoint aPoint) {
		return new Jun3dPoint(aPoint.x, aPoint.y, aPoint.z);
	}

	/**
	 * Answer a new Jun3dPoint with unit values.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category Constants access
	 */
	public static Jun3dPoint Unity() {
		return new Jun3dPoint(1.0d, 1.0d, 1.0d);
	}

	/**
	 * Answer a new Jun3dPoint with zero values.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category Constants access
	 */
	public static Jun3dPoint Zero() {
		return new Jun3dPoint(0.0d, 0.0d, 0.0d);
	}

	/**
	 * Answer the absolute Point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint abs() {
		return new Jun3dPoint(Math.abs(x), Math.abs(y), Math.abs(z));
	}

	/**
	 * Answer the receiver as an array of double.
	 * 
	 * @return double[]
	 * 
	 * @category converting
	 */
	public double[] asArray() {
		return new double[] { x, y, z };
	}

	/**
	 * Answer the bounding box with a point as the corner.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category compatibilities
	 */
	public Jun3dBoundingBox corner_(Jun3dPoint aPoint) {
		return Jun3dBoundingBox.Origin_corner_(this, aPoint);
	}

	/**
	 * Answer the distance from aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double distance_(Jun3dPoint aPoint) {
		return this.minus_(aPoint).rho();
	}

	/**
	 * Answer the result of dividing this point by delta. This method
	 * corresponds to the methd '/' in Smalltalk.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint dividedBy_(double delta) {
		return new Jun3dPoint(this.x / delta, this.y / delta, this.z / delta);
	}

	/**
	 * Answer the result of dividing this point by aPoint. This method
	 * corresponds to the methd '/' in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint dividedBy_(Jun3dPoint aPoint) {
		return new Jun3dPoint(this.x / aPoint.x, this.y / aPoint.y, this.z / aPoint.z);
	}

	/**
	 * Answer the result of a dot product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double dotProduct_(Jun3dPoint aPoint) {
		Jun3dPoint newPoint = this.multipliedBy_(aPoint);

		return newPoint.x + newPoint.y + newPoint.z;
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun3dPoint == false) {
			return false;
		}

		Jun3dPoint aPoint = (Jun3dPoint) anObject;

		return (this.x == aPoint.x) && (this.y == aPoint.y) && (this.z == aPoint.z);
	}

	/**
	 * Answer the bounding box with a point as the extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 * 
	 * @category compatibilities
	 */
	public Jun3dBoundingBox extent_(Jun3dPoint aPoint) {
		return Jun3dBoundingBox.Origin_extent_(this, aPoint);
	}

	/**
	 * Answer the number representing the ordering of the receiver in the
	 * generality hierarchy.
	 * 
	 * @return int
	 * 
	 * @category coercing
	 */
	public int generality() {
		return 220;
	}

	/**
	 * Computes a hash code for this object.
	 * 
	 * @return int
	 * 
	 * @category hash function
	 */
	public int hashCode() {
		long bitsX = Double.doubleToLongBits(x);
		int xHash = (int) (bitsX ^ (bitsX >> 32));
		long bitsY = Double.doubleToLongBits(y);
		int yHash = (int) (bitsY ^ (bitsY >> 32));
		long bitsZ = Double.doubleToLongBits(z);
		int zHash = (int) (bitsZ ^ (bitsZ >> 32));

		return (((xHash >> 2) | yHash) >> 2 | zHash);
	}

	/**
	 * Answer the result of the inner product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 * 
	 * @category functions
	 */
	public double innerProduct_(Jun3dPoint aPoint) {
		return this.dotProduct_(aPoint);
	}

	/**
	 * Answer the intersecting point with the Jun3dLine.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category functions
	 */
	public Jun3dPoint intersectingPointWithLine_(Jun3dLine aLine) {
		if (aLine.containsPoint_(this)) {
			return this;
		} else {
			return null;
		}
	}

	/**
	 * Answer true if this is a kind of 3D geometry object, otherwise false.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean is3d() {
		return true;
	}

	/**
	 * Answer true if the receiver is less than aPoint. This method corresponds
	 * to the method "&lt;" in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean isLessThan_(Jun3dPoint aPoint) {
		return (this.x < aPoint.x) && (this.y < aPoint.y) && (this.z < aPoint.z);
	}

	/**
	 * Answer true if the receiver is less than or equal to aPoint. This method
	 * corresponds to the method "&lt;=" in Smalltalk.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean isLessThanOrEqualTo_(Jun3dPoint aPoint) {
		return (this.x <= aPoint.x) && (this.y <= aPoint.y) && (this.z <= aPoint.z);
	}

	/**
	 * Answer true if this point is ZERO, otherwise false.
	 * 
	 * @return boolean
	 * 
	 * @category testing
	 */
	public boolean isZero() {
		return (this.x == 0.0d) && (this.y == 0.0d) && (this.z == 0.0d);
	}

	/**
	 * Set the length from the origin.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category functions
	 */
	public Jun3dPoint length_(double aNumber) {
		double length = this.length();

		if (length <= JunPoint.ACCURACY) {
			return this;
		}

		return this.multipliedBy_(aNumber / length);
	}

	/**
	 * Answer the maximum point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category comparing
	 */
	public Jun3dPoint max_(Jun3dPoint aPoint) {
		return new Jun3dPoint(Math.max(this.x, aPoint.x), Math.max(this.y, aPoint.y), Math.max(this.z, aPoint.z));
	}

	/**
	 * Answer the minimum point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category comparing
	 */
	public Jun3dPoint min_(Jun3dPoint aPoint) {
		return new Jun3dPoint(Math.min(this.x, aPoint.x), Math.min(this.y, aPoint.y), Math.min(this.z, aPoint.z));
	}

	/**
	 * Answer the difference between this point and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint minus_(double delta) {
		return new Jun3dPoint(this.x - delta, this.y - delta, this.z - delta);
	}

	/**
	 * Answer the difference between this point and aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint minus_(Jun3dPoint aPoint) {
		return new Jun3dPoint(this.x - aPoint.x, this.y - aPoint.y, this.z - aPoint.z);
	}

	/**
	 * Answer the result of multiplying this point by aPoint.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint multipliedBy_(double delta) {
		return new Jun3dPoint(this.x * delta, this.y * delta, this.z * delta);
	}

	/**
	 * Answer the result of multiplying this point by aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint multipliedBy_(Jun3dPoint aPoint) {
		return new Jun3dPoint(this.x * aPoint.x, this.y * aPoint.y, this.z * aPoint.z);
	}

	/**
	 * Answer the negation of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category testing
	 */
	public Jun3dPoint negated() {
		return Jun3dPoint.Zero().minus_(this);
	}

	/**
	 * Answer the sum of this point and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint plus_(double delta) {
		return new Jun3dPoint(this.x + delta, this.y + delta, this.z + delta);
	}

	/**
	 * Answer the sum of this point and aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint plus_(Jun3dPoint aPoint) {
		return new Jun3dPoint(this.x + aPoint.x, this.y + aPoint.y, this.z + aPoint.z);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 * 
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("(");
		aWriter.write(String.valueOf(this.x));
		aWriter.write(" , ");
		aWriter.write(String.valueOf(this.y));
		aWriter.write(" , ");
		aWriter.write(String.valueOf(this.z));
		aWriter.write(")");
	}

	/**
	 * Answer the result of a product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category functions
	 */
	public Jun3dPoint product_(Jun3dPoint aPoint) {
		return this.vectorProduct_(aPoint);
	}

	/**
	 * Answer 1 divided by the receiver.
	 * Provide an error notification if the receiver is 0.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category arithmetic
	 */
	public Jun3dPoint reciprocal() {
		return Unity().dividedBy_(this);
	}

	/**
	 * Answer the rho as of a polar coordinate.
	 * 
	 * @return double
	 * 
	 * @category polar coordinates
	 */
	public double rho() {
		return Math.sqrt(this.dotProduct_(this));
	}

	/**
	 * Answer the new Jun3dPoint which is rotated by aJunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category transforming
	 */
	public Jun3dPoint rotatedBy_(JunAngle anAngle) {
		return this.transform_(Jun3dTransformation.Rotate_(anAngle));
	}

	/**
	 * Answer the rounded Jun3dPoint.
	 * 
	 * @return Jun3dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun3dPoint rounded() {
		return new Jun3dPoint((double) Math.round(this.x), (double) Math.round(this.y), (double) Math.round(this.z));
	}

	/**
	 * Answer the new Jun3dPoint which is scaled by aJunPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category transforming
	 */
	public Jun3dPoint scaledBy_(Jun3dPoint aPoint) {
		return this.transform_(Jun3dTransformation.Scale_(aPoint));
	}

	/**
	 * Apply a transformation 'aJunTransformation' to the receiver. This method
	 * is not defined in Smalltalk version.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.Jun3dTransformation
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category transforming
	 */
	public Jun3dPoint transform_(Jun3dTransformation aTransformation) {
		return aTransformation.applyToPoint_(this);
	}

	/**
	 * Answer the new Jun3dPoint which is translated by aJunPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category transforming
	 */
	public Jun3dPoint translatedBy_(Jun3dPoint aPoint) {
		return this.transform_(Jun3dTransformation.Translate_(aPoint));
	}

	/**
	 * Answer the truncated Jun3dPoint.
	 * 
	 * @return Jun3dPoint
	 * 
	 * @category truncation and round off
	 */
	public Jun3dPoint truncated() {
		double newX;
		double newY;
		double newZ;
		newX = (this.x >= 0) ? Math.ceil(this.x) : Math.floor(this.x);
		newY = (this.y >= 0) ? Math.ceil(this.y) : Math.floor(this.y);
		newZ = (this.z >= 0) ? Math.ceil(this.z) : Math.floor(this.z);

		return new Jun3dPoint(newX, newY, newZ);
	}

	/**
	 * Answer the unit vector derived from this point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category functions
	 */
	public Jun3dPoint unitVector() {
		return this.length_(1.0d);
	}

	/**
	 * Answer the result of a vector product with aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @category functions
	 */
	public Jun3dPoint vectorProduct_(Jun3dPoint aPoint) {
		return new Jun3dPoint((y * aPoint.z) - (z * aPoint.y), (z * aPoint.x) - (x * aPoint.z), (x * aPoint.y) - (y * aPoint.x));
	}

	/**
	 * Answer the x coordinate.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double x() {
		return this.x;
	}

	/**
	 * Answer the y coordinate.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double y() {
		return this.y;
	}

	/**
	 * Answer the z coordinate.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double z() {
		return this.z;
	}
}
