package sra.jun.geometry.basic;

import java.io.*;
import sra.jun.geometry.abstracts.*;

/**
 * JunAngle class
 * 
 *  @author    nisinaka
 *  @created   1998/10/05 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunAngle.java,v 6.0 2003/10/28 10:52:04 nisinaka Exp $
 */
public class JunAngle extends JunGeometry {
	/** Some useful values. */
	static final double _Pi = Math.PI;
	static final double _RadiansPerDegree = _Pi / 180.0d;
	private double rad;
	private Double deg = null;
	private Double sin = null;
	private Double cos = null;
	private Double tan = null;

	/**
	 * Convert the value from degrees to radians.
	 * 
	 * @param degrees double
	 * 
	 * @return double
	 */
	public static double _DegreesToRadians(double degrees) {
		return degrees * _RadiansPerDegree;
	}

	/**
	 * Convert the value from radians to degrees.
	 * 
	 * @param radians double
	 * 
	 * @return double
	 */
	public static double _RadiansToDegrees(double radians) {
		return radians / _RadiansPerDegree;
	}

	/**
	 * Create a new JunAngle speficied with degrees.
	 * 
	 * @param degrees double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category Instance creation
	 */
	public static final JunAngle FromDeg_(double degrees) {
		return FromDeg_(JunAngle.class, degrees);
	}

	/**
	 * Create a new JunAngle speficied with degrees.
	 * 
	 * @param aClass java.lang.Class
	 * @param degrees double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category Instance creation
	 */
	public static JunAngle FromDeg_(Class aClass, double degrees) {
		JunAngle anAngle = (JunAngle) _New(aClass);
		anAngle.deg_(degrees);

		return anAngle;
	}

	/**
	 * Create a new JunAngle speficied with radians.
	 * 
	 * @param radians double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category Instance creation
	 */
	public static final JunAngle FromRad_(double radians) {
		return FromRad_(JunAngle.class, radians);
	}

	/**
	 * Create a new JunAngle speficied with radians.
	 * 
	 * @param aClass java.lang.Class
	 * @param radians double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category Instance creation
	 */
	public static JunAngle FromRad_(Class aClass, double radians) {
		JunAngle anAngle = (JunAngle) _New(aClass);
		anAngle.rad_(radians);

		return anAngle;
	}

	/**
	 * Answer a constant, zero, angle.
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category Constants access
	 */
	public static JunAngle Zero() {
		return JunAngle.FromRad_(0);
	}

	/**
	 * Answer as cos.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double cos() {
		if (cos == null) {
			cos = new Double(Math.cos(rad));
		}

		return cos.doubleValue();
	}

	/**
	 * Answer as degrees.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double deg() {
		if (deg == null) {
			deg = new Double(_RadiansToDegrees(rad));
		}

		return deg.doubleValue();
	}

	/**
	 * Set this angle from degrees.
	 * 
	 * @param degrees double
	 * 
	 * @category accessing
	 */
	public void deg_(double degrees) {
		this.setRad_(_DegreesToRadians(degrees));
	}

	/**
	 * Divide the receiver by aNumber.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle div_(double aNumber) {
		return JunAngle.FromRad_(rad / aNumber);
	}

	/**
	 * Answer the result of dividing the receiver by delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle dividedBy_(double delta) {
		return JunAngle.FromRad_(rad / delta);
	}

	/**
	 * Answer the result of dividing the receiver by the JunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle dividedBy_(JunAngle anAngle) {
		return this.dividedBy_(anAngle.rad());
	}

	/**
	 * Answer true if the Object is equal to the receiver, otherwise false.
	 * 
	 * @param anObject
	 * 
	 * @return boolean
	 * 
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof JunAngle == false) {
			return false;
		}

		return this.rad == ((JunAngle) anObject).rad;
	}

	/**
	 * Answer the number representing the ordering of the receiver in the
	 * generality hierarchy.
	 * 
	 * @return int
	 * 
	 * @category coercing
	 */
	public int generality() {
		return 120;
	}

	/**
	 * Answer the difference between the receiver and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle minus_(double delta) {
		return JunAngle.FromRad_(rad - delta);
	}

	/**
	 * Answer the difference between the receiver and the JunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle minus_(JunAngle anAngle) {
		return this.minus_(anAngle.rad());
	}

	/**
	 * Answer an instance of JunAngle which angle is multiplied by the
	 * specified number.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle mul_(double aNumber) {
		return JunAngle.FromRad_(this.getClass(), rad * aNumber);
	}

	/**
	 * Answer the result of multiplying the receiver by delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle multipliedBy_(double delta) {
		return JunAngle.FromRad_(rad * delta);
	}

	/**
	 * Answer the result of multiplying the receiver by the JunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle multipliedBy_(JunAngle anAngle) {
		return this.multipliedBy_(anAngle.rad());
	}

	/**
	 * Answer the negation of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle negated() {
		return JunAngle.FromRad_(this.getClass(), -rad);
	}

	/**
	 * Answer the sum of the receiver and delta.
	 * 
	 * @param delta double
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle plus_(double delta) {
		return JunAngle.FromRad_(rad + delta);
	}

	/**
	 * Answer the sum of the receiver and the JunAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * 
	 * @category arithmetic
	 */
	public JunAngle plus_(JunAngle anAngle) {
		return this.plus_(anAngle.rad());
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 * 
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("Angle (");
		aWriter.write(String.valueOf(this.deg()));
		aWriter.write(")");
	}

	/**
	 * Answer as radians.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double rad() {
		return rad;
	}

	/**
	 * Set this angle from radians.
	 * 
	 * @param radians double
	 * 
	 * @category accessing
	 */
	public void rad_(double radians) {
		this.setRad_(radians);
	}

	/**
	 * Answer as sin.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double sin() {
		if (sin == null) {
			sin = new Double(Math.sin(rad));
		}

		return sin.doubleValue();
	}

	/**
	 * Answer as tan.
	 * 
	 * @return double
	 * 
	 * @category accessing
	 */
	public double tan() {
		if (tan == null) {
			tan = new Double(Math.tan(rad));
		}

		return tan.doubleValue();
	}

	/**
	 * Set the new value of 'rad' and reset other cached values.
	 * 
	 * @param radians double
	 * 
	 * @category private
	 */
	private void setRad_(double radians) {
		rad = radians;
		deg = null;
		sin = null;
		cos = null;
		tan = null;
	}
}
