package sra.jun.geometry.curves;

import java.io.*;
import sra.smalltalk.*;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.transformations.Jun2dTransformation;

/**
 * Jun2dLine class
 * 
 *  @author    nisinaka
 *  @created   1998/10/09 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dLine.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class Jun2dLine extends JunLine {
	protected double x0;
	protected double y0;
	protected double f;
	protected double g;

	/**
	 * Construct a new instance of Jun2dLine with two points.
	 * 
	 * @param fromPoint Jun2dPoint
	 * @param toPoint Jun2dPoint
	 * @category Instance creation
	 */
	public Jun2dLine(Jun2dPoint fromPoint, Jun2dPoint toPoint) {
		super();
		this.from_to_(fromPoint, toPoint);
	}

	/**
	 * Answer a new Jun2dLine with unit values.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @category Constants access
	 */
	public static final Jun2dLine Unity() {
		Jun2dLine aLine = new Jun2dLine(Jun2dPoint.Zero(), Jun2dPoint.Unity());
		return aLine.normalizedLine();
	}

	/**
	 * Answer the angle between this line and aLine.
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * @param aLine sra.jun.geometry.curves.Jun2dLine
	 * @category accessing
	 */
	public JunAngle angleWithLine_(Jun2dLine aLine) {
		double denominator = (Math.pow(this.f(), 2) + Math.pow(this.g(), 2)) * (Math.pow(aLine.f(), 2) + Math.pow(aLine.g(), 2));
		denominator = Math.sqrt(denominator);
		if (denominator < Jun2dPoint.Accuracy()) {
			SmalltalkException.Error("unexpected line parameters");
		}
		double numerator = this.f() * aLine.f() + this.g() * aLine.g();
		double gamma = Math.min(Math.max(numerator / denominator, -1.0d), 1.0d);
		return JunAngle.FromRad_(Math.acos(gamma));
	}

	/**
	 * Answer the angle between this line and aLine.
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * @category accessing
	 */
	public JunAngle angleWithLine_(Jun3dLine aLine) {
		Jun2dLine theLine = new Jun2dLine(new Jun2dPoint(aLine.first().x(), aLine.first().y()), new Jun2dPoint(aLine.last().x(), aLine.last().y()));
		;
		return this.angleWithLine_(theLine);
	}

	/**
	 * Answer a regularized point of this curve at a specified number.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param t double
	 * @category accessing
	 */
	public Jun2dPoint atT_(double t) {
		return new Jun2dPoint(x0 + (f * t), y0 + (g * t));
	}

	/**
	 * Answer the center point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint center() {
		return this.atT_(0.5d);
	}

	/**
	 * Answer true if the receiver contains aPoint.
	 * 
	 * @return boolean
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category testing
	 */
	public boolean containsPoint_(Jun2dPoint aPoint) {
		return this.distanceFromPoint_(aPoint) < ACCURACY;
	}

	/**
	 * Answer the distance between the Jun2dPoint and the receiver.
	 * 
	 * @return double
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category functions
	 */
	public double distanceFromPoint_(Jun2dPoint aPoint) {
		Jun2dPoint pointOnMe = this.nearestPointFromPoint_(aPoint);
		return aPoint.distance_(pointOnMe);
	}

	/**
	 * Answer true if the receiver equals to anObject.
	 * 
	 * @return boolean
	 * @param anObject java.lang.Object
	 * @category comparing
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun2dLine == false) {
			return false;
		}
		Jun2dLine aLine = (Jun2dLine) anObject;
		return this.from().equals(aLine.from()) && this.to().equals(aLine.to());
	}

	/**
	 * Answer the f parameter.
	 * 
	 * @return double
	 * @category parameters
	 */
	public double f() {
		return f;
	}

	/**
	 * Answer the start point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint first() {
		return this.from();
	}

	/**
	 * Answer the start point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint from() {
		return this.atT_(0.0d);
	}

	/**
	 * Initialize this line with two points.
	 * 
	 * @param fromPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param toPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category initialize-release
	 */
	public void from_to_(Jun2dPoint fromPoint, Jun2dPoint toPoint) {
		x0 = fromPoint.x();
		y0 = fromPoint.y();
		f = toPoint.x() - x0;
		g = toPoint.y() - y0;
	}

	/**
	 * Answer the g parameter.
	 * 
	 * @return double
	 * @category parameters
	 */
	public double g() {
		return g;
	}

	/**
	 * Answer the intersecting point of the Jun2dLine and the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param aLine sra.jun.geometry.curves.Jun2dLine
	 * @category functions
	 */
	public Jun2dPoint intersectingPointWithLine_(Jun2dLine aLine) {
		double t = this.intersectingTWithLine_(aLine);
		if (Double.isNaN(t)) {
			return null;
		}
		return (Jun2dPoint) this.atT_(t);
	}

	/**
	 * Answer the T parameter of the intersecting point of the Jun2dLine and
	 * the receiver.
	 * 
	 * @return double
	 * @param aLine sra.jun.geometry.curves.Jun2dLine
	 * @category functions
	 */
	public double intersectingTWithLine_(Jun2dLine aLine) {
		double yourF = aLine.f;
		double yourG = aLine.g;
		double denominator = (yourF * g) - (f * yourG);

		if (Math.abs(denominator) < ACCURACY) {
			// These are parallel lines.
			return Double.NaN;
		}

		return (yourF * (aLine.y0 - y0) - yourG * (aLine.x0 - x0)) / denominator;
	}

	/**
	 * Answer true if this line and aLine is parallel, otherwise false.
	 * 
	 * @return boolean
	 * @param aLine sra.jun.geometry.curves.Jun2dLine
	 * @category accessing
	 */
	public boolean isParallelWithLine_(Jun2dLine aLine) {
		double degrees = this.angleWithLine_(aLine).deg();
		return (Math.abs(0.0d - degrees) < Jun2dPoint.Accuracy() || Math.abs(180.0d - degrees) < Jun2dPoint.Accuracy());
	}

	/**
	 * Answer the end point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint last() {
		return this.to();
	}

	/**
	 * Answer true if the receiver contains the Jun2dPoint, otherwise false.
	 * 
	 * @return boolean
	 * @param aJun2dPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category functions
	 */
	public boolean lineSegmentContainsPoint_(Jun2dPoint aJun2dPoint) {
		if (!this.containsPoint_(aJun2dPoint)) {
			return false;
		}
		double t = this.tAtPoint_(aJun2dPoint);
		return (-ACCURACY <= t) && (t <= (1 + ACCURACY));
	}

	/**
	 * Answer the distance between the Jun2dPoint and the receiver.
	 * 
	 * @return double
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category functions
	 */
	public double lineSegmentDistanceFromPoint_(Jun2dPoint aPoint) {
		Jun2dPoint pointOnMe = this.lineSegmentNearestPointFromPoint_(aPoint);
		return aPoint.distance_(pointOnMe);
	}

	/**
	 * Answer the intersecting point of the Jun2dLine and the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param aLine sra.jun.geometry.curves.Jun2dLine
	 * @category functions
	 */
	public Jun2dPoint lineSegmentIntersectingPointWithLineSegment_(Jun2dLine aLine) {
		Jun2dPoint intersectingPoint = this.intersectingPointWithLine_(aLine);
		if (intersectingPoint == null) {
			return null;
		}
		if (!this.lineSegmentContainsPoint_(intersectingPoint)) {
			return null;
		}
		if (!aLine.lineSegmentContainsPoint_(intersectingPoint)) {
			return null;
		}
		return intersectingPoint;
	}

	/**
	 * Answer the nearest point on the line segment from the Jun2dPoint.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category functions
	 */
	public Jun2dPoint lineSegmentNearestPointFromPoint_(Jun2dPoint aPoint) {
		double denominator = (f * f) + (g * g);
		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}
		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		double t = (f * x + g * y) / denominator;
		if (t < 0) {
			t = 0;
		} else if (t > 1) {
			t = 1;
		}
		return (Jun2dPoint) this.atT_(t);
	}

	/**
	 * Answer a coefficient for my normalization.
	 * 
	 * @return double
	 * @category accessing
	 */
	public double n() {
		double denominator = (f * f) + (g * g);
		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}
		return 1.0d / Math.sqrt(denominator);
	}

	/**
	 * Answer the nearest point on the receiver from the Jun2dPoint.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category functions
	 */
	public Jun2dPoint nearestPointFromPoint_(Jun2dPoint aPoint) {
		double denominator = (f * f) + (g * g);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;

		return (Jun2dPoint) this.atT_((f * x + g * y) / denominator);
	}

	/**
	 * Create a normalized line of this.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @category parameters
	 */
	public Jun2dLine normalizedLine() {
		Jun2dLine aLine = (Jun2dLine) this.copy();
		double n = this.n();
		aLine.f = f * n;
		aLine.g = g * n;

		return aLine;
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws IOException DOCUMENT ME!
	 * @category printing
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("2dLine (");
		this.from().printOn_(aWriter);
		aWriter.write(" , ");
		this.to().printOn_(aWriter);
		aWriter.write(")");
	}

	/**
	 * Answer the reversed line of the receiver.
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 * @category converting
	 */
	public Jun2dLine reversed() {
		return new Jun2dLine(this.to(), this.from());
	}

	/**
	 * Answer the new Jun2dLine which is rotated the receiver by anAngle.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * @category transforming
	 */
	public Jun2dLine rotatedBy_(JunAngle anAngle) {
		return this.transform_(Jun2dTransformation.Rotate_(anAngle));
	}

	/**
	 * Answer the new Jun2dLine which is scaled by aPoint.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category transforming
	 */
	public Jun2dLine scaledBy_(Jun2dPoint aPoint) {
		return this.transform_(Jun2dTransformation.Scale_(aPoint));
	}

	/**
	 * Answer the 't' value of the Jun2dPoint.
	 * 
	 * @return double
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public double tAtPoint_(Jun2dPoint aPoint) {
		double denominator = (f * f) + (g * g);
		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}
		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		return (f * x + g * y) / denominator;
	}

	/**
	 * Answer the end point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 * @category accessing
	 */
	public Jun2dPoint to() {
		return this.atT_(1.0d);
	}

	/**
	 * Apply aJunTransformation to the receiver.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.Jun2dTransformation
	 * @category transforming
	 */
	public Jun2dLine transform_(Jun2dTransformation aTransformation) {
		return new Jun2dLine(this.from().transform_(aTransformation), this.to().transform_(aTransformation));
	}

	/**
	 * Answer the new Jun2dLine which is translated by anPoint.
	 * 
	 * @return sra.jun.geometry.curves.Jun2dLine
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @category transforming
	 */
	public Jun2dLine translatedBy_(Jun2dPoint aPoint) {
		return this.transform_(Jun2dTransformation.Translate_(aPoint));
	}

	/**
	 * Answer the x0 parameter.
	 * 
	 * @return double
	 * @category parameters
	 */
	public double x0() {
		return x0;
	}

	/**
	 * Answer the g parameter.
	 * 
	 * @return double
	 * @category parameters
	 */
	public double y0() {
		return y0;
	}
}
