package sra.jun.geometry.curves;

import java.io.*;
import sra.smalltalk.*;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.transformations.Jun3dTransformation;
import sra.jun.opengl.objects.*;

/**
 * Jun3dLine class
 * 
 *  @author    nisinaka
 *  @created   1998/10/09 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dLine.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class Jun3dLine extends JunLine {
	protected double x0;
	protected double y0;
	protected double z0;
	protected double f;
	protected double g;
	protected double h;

	/** The cache of the bounding box. */
	protected Jun3dBoundingBox boundingBox = null;

	/**
	 * Construct a new instance of Jun3dLine with two points.
	 * 
	 * @param fromPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param toPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dLine(Jun3dPoint fromPoint, Jun3dPoint toPoint) {
		super();
		this.from_to_(fromPoint, toPoint);
	}

	/**
	 * Answer a new Jun3dLine with unit values.
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public static Jun3dLine Unity() {
		Jun3dLine aLine = new Jun3dLine(Jun3dPoint.Zero(), Jun3dPoint.Unity());

		return aLine.normalizedLine();
	}

	/**
	 * Answer the angle between the receiver and aLine.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 */
	public JunAngle angleWithLine_(Jun3dLine aLine) {
		double denominator = ((f * f) + (g * g) + h * h) * ((aLine.f * aLine.f) + (aLine.g * aLine.g) + aLine.h * aLine.h);
		denominator = Math.sqrt(denominator);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("unexpected line parameters");
		}

		double numerator = (f * aLine.f) + (g * aLine.g) + (h * aLine.h);
		double gamma = Math.min(Math.max(numerator / denominator, -1.0d), 1.0d);

		return JunAngle.FromRad_(Math.acos(gamma));
	}

	/**
	 * Convert the receiver as a JunOpenGL3dObject.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject asJunOpenGL3dObject() {
		Jun3dPoint[] vertexes = {(Jun3dPoint) this.from(), (Jun3dPoint) this.to()};

		return new JunOpenGL3dPolyline(vertexes);
	}

	/**
	 * Answer a regularized point of this curve at a specified number.
	 * 
	 * @param t double
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint atT_(double t) {
		return new Jun3dPoint(x0 + (f * t), y0 + (g * t), z0 + (h * t));
	}

	/**
	 * Answer the bounding box of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 */
	public Jun3dBoundingBox boundingBox() {
		if (boundingBox == null) {
			boundingBox = this.preferredBoundingBox();
		}

		return boundingBox;
	}

	/**
	 * Answer the center point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint center() {
		return this.atT_(0.5d);
	}

	/**
	 * Answer true if the receiver contains aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 */
	public boolean containsPoint_(Jun3dPoint aPoint) {
		return this.distanceFromPoint_(aPoint) < ACCURACY;
	}

	/**
	 * Answer the distance between the Jun3dPoint and the receiver.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 */
	public double distanceFromPoint_(Jun3dPoint aPoint) {
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		double z = aPoint.z() - z0;
		double t = ((f * x) + (g * y) + h * z) / denominator;
		Jun3dPoint pointOnMe = this.atT_(t);

		return aPoint.distance_(pointOnMe);
	}

	/**
	 * Answer true if the receiver equals to anObject.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return boolean
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun3dLine == false) {
			return false;
		}

		Jun3dLine aLine = (Jun3dLine) anObject;

		return this.from().equals(aLine.from()) && this.to().equals(aLine.to());
	}

	/**
	 * Answer the parameter f.
	 * 
	 * @return double
	 */
	public double f() {
		return f;
	}

	/**
	 * Answer the start point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint first() {
		return this.from();
	}

	/**
	 * Answer the start point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint from() {
		return this.atT_(0.0d);
	}

	/**
	 * Initialize this line with two points.
	 * 
	 * @param fromPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param toPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public void from_to_(Jun3dPoint fromPoint, Jun3dPoint toPoint) {
		x0 = fromPoint.x();
		y0 = fromPoint.y();
		z0 = fromPoint.z();
		f = toPoint.x() - x0;
		g = toPoint.y() - y0;
		h = toPoint.z() - z0;
	}

	/**
	 * Answer the parameter g.
	 * 
	 * @return double
	 */
	public double g() {
		return g;
	}

	/**
	 * Answer the parameter h.
	 * 
	 * @return double
	 */
	public double h() {
		return h;
	}

	/**
	 * Answer the intersecting point with the Jun3dLine.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint intersectingPointWithLine_(Jun3dLine aLine) {
		double yourF = aLine.f;
		double yourG = aLine.g;
		double yourH = aLine.h;
		double yourX0 = aLine.x0;
		double yourY0 = aLine.y0;
		double yourZ0 = aLine.z0;
		double denominatorXY = (yourF * g) - (f * yourG);
		double denominatorYZ = (yourG * h) - (g * yourH);
		double denominatorZX = (yourH * f) - (h * yourF);
		double denominatorXYabs = Math.abs(denominatorXY);
		double denominatorYZabs = Math.abs(denominatorYZ);
		double denominatorZXabs = Math.abs(denominatorZX);
		double t;
		double yourT;

		if (Math.max(Math.max(denominatorXYabs, denominatorYZabs), denominatorZXabs) < ACCURACY) {
			// These are parallel lines.
			return null;
		}

		if ((denominatorXYabs >= denominatorYZabs) && (denominatorXYabs >= denominatorZXabs)) {
			t = (yourF * (yourY0 - y0) - yourG * (yourX0 - x0)) / denominatorXY;

			if (Math.abs(yourF) > Math.abs(yourG)) {
				yourT = ((f * t) + x0 - yourX0) / yourF;
			} else {
				yourT = ((g * t) + y0 - yourY0) / yourG;
			}
		} else if (denominatorYZabs >= denominatorZXabs) {
			t = (yourG * (yourZ0 - z0) - yourH * (yourY0 - y0)) / denominatorYZ;

			if (Math.abs(yourG) > Math.abs(yourH)) {
				yourT = ((g * t) + y0 - yourY0) / yourG;
			} else {
				yourT = ((h * t) + z0 - yourZ0) / yourH;
			}
		} else {
			t = (yourH * (yourX0 - x0) - yourF * (yourZ0 - z0)) / denominatorZX;

			if (Math.abs(yourH) > Math.abs(yourF)) {
				yourT = ((h * t) + z0 - yourZ0) / yourH;
			} else {
				yourT = ((f * t) + x0 - yourX0) / yourF;
			}
		}

		double x = (f * t) + x0;
		double y = (g * t) + y0;
		double z = (h * t) + z0;
		double yourX = yourF * yourT * yourX0;
		double yourY = yourG * yourT * yourY0;
		double yourZ = yourH * yourT * yourZ0;

		if ((Math.abs(x - yourX) > ACCURACY) || (Math.abs(y - yourY) > ACCURACY) || (Math.abs(z - yourZ) > ACCURACY)) {
			return null;
		}

		return new Jun3dPoint(x, y, z);
	}

	/**
	 * Answer the end point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint last() {
		return this.to();
	}

	/**
	 * Answer true if the receiver contains the Jun3dPoint, otherwise false.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 */
	public boolean lineSegmentContainsPoint_(Jun3dPoint aPoint) {
		return this.lineSegmentDistanceFromPoint_(aPoint) < ACCURACY;
	}

	/**
	 * Answer the distance between the Jun3dPoint and the receiver.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 */
	public double lineSegmentDistanceFromPoint_(Jun3dPoint aPoint) {
		Jun3dPoint thePoint = (Jun3dPoint) aPoint;
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = thePoint.x() - x0;
		double y = thePoint.y() - y0;
		double z = thePoint.z() - z0;
		double t = ((f * x) + (g * y) + h * z) / denominator;
		Jun3dPoint pointOnMe;

		if (t < 0) {
			pointOnMe = (Jun3dPoint) this.atT_(0);
		} else if (t > 1) {
			pointOnMe = (Jun3dPoint) this.atT_(1);
		} else {
			pointOnMe = (Jun3dPoint) this.atT_(t);
		}

		return thePoint.distance_(pointOnMe);
	}

	/**
	 * Answer the nearest point on the line segment from the Jun3dPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint lineSegmentNearestPointFromPoint_(Jun3dPoint aPoint) {
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		double z = aPoint.z() - z0;
		double t = ((f * x) + (g * y) + h * z) / denominator;

		if (t < 0) {
			t = 0;
		} else if (t > 1) {
			t = 1;
		}

		return (Jun3dPoint) this.atT_(t);
	}

	/**
	 * Answer a coefficient for my normalization.
	 * 
	 * @return double
	 */
	public double n() {
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		return 1.0d / Math.sqrt(denominator);
	}

	/**
	 * Answer the nearest point on the receiver from the Jun3dPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint nearestPointFromPoint_(Jun3dPoint aPoint) {
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		double z = aPoint.z() - z0;

		return (Jun3dPoint) this.atT_(((f * x) + (g * y) + h * z) / denominator);
	}

	/**
	 * Create a normalized line of this.
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine normalizedLine() {
		Jun3dLine aLine = (Jun3dLine) this.copy();
		double n = this.n();
		aLine.f = f * n;
		aLine.g = g * n;
		aLine.h = h * n;

		return aLine;
	}

	/**
	 * Answer the preferred bounding box of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 */
	public Jun3dBoundingBox preferredBoundingBox() {
		return Jun3dBoundingBox.Vertex_vertex_((Jun3dPoint) this.from(), (Jun3dPoint) this.to());
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("3dLine (");
		this.from().printOn_(aWriter);
		aWriter.write(" , ");
		this.to().printOn_(aWriter);
		aWriter.write(")");
	}

	/**
	 * Answer the reversed line of the receiver.
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine reversed() {
		return new Jun3dLine(this.to(), this.from());
	}

	/**
	 * Answer the new Jun3dLine which is rotated by anAngle.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine rotatedBy_(JunAngle anAngle) {
		return this.transform_(Jun3dTransformation.Rotate_(anAngle));
	}

	/**
	 * Answer the new Jun3dLine which is scaled by aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine scaledBy_(Jun3dPoint aPoint) {
		return this.transform_(Jun3dTransformation.Scale_(aPoint));
	}

	/**
	 * Answer the 't' value of the Jun3dPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 */
	public double tAtPoint_(Jun3dPoint aPoint) {
		double denominator = (f * f) + (g * g) + (h * h);

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("can not define a line");
		}

		double x = aPoint.x() - x0;
		double y = aPoint.y() - y0;
		double z = aPoint.z() - z0;

		return ((f * x) + (g * y) + h * z) / denominator;
	}

	/**
	 * Answer the end point of this curve.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint to() {
		return this.atT_(1.0d);
	}

	/**
	 * Apply aJunTransformation to the receiver.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.Jun3dTransformation
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine transform_(Jun3dTransformation aTransformation) {
		return new Jun3dLine(this.from().transform_(aTransformation), this.to().transform_(aTransformation));
	}

	/**
	 * Answer the new Jun3dLine which is translated by aPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.abstracts.Jun3dLine
	 */
	public Jun3dLine translatedBy_(Jun3dPoint aPoint) {
		return this.transform_(Jun3dTransformation.Translate_(aPoint));
	}

	/**
	 * Answer the parameter x0.
	 * 
	 * @return double
	 */
	public double x0() {
		return x0;
	}

	/**
	 * Answer the parameter y0.
	 * 
	 * @return double
	 */
	public double y0() {
		return y0;
	}

	/**
	 * Answer the parameter z0.
	 * 
	 * @return double
	 */
	public double z0() {
		return z0;
	}
}
