package sra.jun.geometry.surfaces;

import java.util.Vector;
import sra.jun.geometry.abstracts.JunGeometry;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.curves.Jun2dLine;
import sra.jun.opengl.objects.*;
import sra.smalltalk.*;

/**
 * Jun2dPolygon class
 * 
 *  @author    nisinaka
 *  @created   1999/07/28 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun2dPolygon.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class Jun2dPolygon extends JunGeometry {
	/** A collection of Jun2dPoint. */
	protected Jun2dPoint[] points;

	/** A number to represent my signed area. */
	protected double signedArea = Double.NaN;

	/** My bounding box. */
	protected Jun2dBoundingBox bounds;

	/**
	 * Create a new instance of Jun2dPolygon and initialize it with the array
	 * of points.
	 * 
	 * @param anArrayOfPoint sra.jun.geometry.basic.Jun2dPoint[]
	 */
	public Jun2dPolygon(Jun2dPoint[] anArrayOfPoint) {
		this.setPoints_(anArrayOfPoint);
	}

	/**
	 * Create a new instance of Jun2dPolygon and initialize it with the
	 * collection of points.
	 * 
	 * @param aCollectionOfPoint java.util.Vector
	 */
	public Jun2dPolygon(Vector aCollectionOfPoint) {
		Jun2dPoint[] points = new Jun2dPoint[aCollectionOfPoint.size()];
		aCollectionOfPoint.copyInto(points);
		this.setPoints_(points);
	}

	/**
	 * Answer a number which represents the signed area for the array of
	 * points.
	 * 
	 * @param anArrayOfPoint sra.jun.geometry.basic.Jun2dPoint[]
	 * 
	 * @return double
	 */
	public static double SignedAreaForPoints_(Jun2dPoint[] anArrayOfPoint) {
		int size = anArrayOfPoint.length;
		double area = 0.0d;

		for (int i = 0; i < size; i++) {
			Jun2dPoint p1 = anArrayOfPoint[i];
			Jun2dPoint p2 = anArrayOfPoint[(i + 1) % size];
			area += ((p1.y() + p2.y()) * (p1.x() - p2.x()));
		}

		return area / 2;
	}

	/**
	 * Answer a number which represents the signed area for the array of
	 * points.
	 * 
	 * @param anArrayOfPoint sra.jun.geometry.basic.Jun3dPoint[]
	 * 
	 * @return double
	 */
	public static double SignedAreaForPoints_(Jun3dPoint[] anArrayOfPoint) {
		int size = anArrayOfPoint.length;
		double area = 0.0d;

		for (int i = 0; i < size; i++) {
			Jun3dPoint p1 = anArrayOfPoint[i];
			Jun3dPoint p2 = anArrayOfPoint[(i + 1) % size];
			area += ((p1.y() + p2.y()) * (p1.x() - p2.x()));
		}

		return area / 2;
	}

	/**
	 * Answer the signed area of the triangle.
	 * 
	 * @param aPoint1 sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint2 sra.jun.geometry.basic.Jun2dPoint
	 * @param aPoint3 sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 */
	public static double SignedAreaForTriangle_and_and_(Jun2dPoint aPoint1, Jun2dPoint aPoint2, Jun2dPoint aPoint3) {
		return (((aPoint1.y() + aPoint2.y()) * (aPoint1.x() - aPoint2.x())) + ((aPoint2.y() + aPoint3.y()) * (aPoint2.x() - aPoint3.x())) + (aPoint3.y() + aPoint1.y()) * (aPoint3.x() - aPoint1.x())) / 2;
	}

	/**
	 * Answer true if the array of points make a consistent polygon.
	 * 
	 * @param anArrayOfJun2dPoint sra.jun.geometry.basic.Jun2dPoint[]
	 * 
	 * @return boolean
	 */
	protected static boolean IsConsistentPolygonFromPoints_(Jun2dPoint[] anArrayOfJun2dPoint) {
		final Jun2dPoint[] thePoints = (new Jun2dPolygon(anArrayOfJun2dPoint)).asPositivePolygon().points();
		final int size = thePoints.length;
		StBlockClosure isConcaveBlock = new StBlockClosure() {
			public Object value_(Object anObject) {
				int index = ((Number) anObject).intValue();
				Jun2dPoint p1 = thePoints[(index - 1) % size];
				Jun2dPoint p2 = thePoints[index];
				Jun2dPoint p3 = thePoints[(index + 1) % size];

				return new Boolean((SignedAreaForTriangle_and_and_(p1, p2, p3) + ACCURACY) <= 0);
			}
		};

		StBlockClosure isConvexBlock = new StBlockClosure() {
			public Object value_(Object anObject) {
				int index = ((Number) anObject).intValue();
				Jun2dPoint p1 = thePoints[(index - 1) % size];
				Jun2dPoint p2 = thePoints[index];
				Jun2dPoint p3 = thePoints[(index + 1) % size];

				return new Boolean((SignedAreaForTriangle_and_and_(p1, p2, p3) + ACCURACY) >= 0);
			}
		};

		StBlockClosure containsBlock = new StBlockClosure() {
			public Object value_(Object anObject) {
				int index = ((Number) anObject).intValue();
				Jun2dPoint point = thePoints[index];
				Vector v = new Vector();

				for (int i = 0; i < thePoints.length; i++) {
					if (point.equals(thePoints[i]) == false) {
						v.addElement(thePoints[i]);
					}
				}

				Jun2dPoint[] ps = new Jun2dPoint[v.size()];
				v.copyInto(ps);

				return new Boolean(PolygonFromPoints_containsPoint_(ps, point));
			}
		};

		//
		for (int i = 0; i < size; i++) {
			Integer index = new Integer(i);
			boolean contains = ((Boolean) containsBlock.value_(index)).booleanValue();

			if (((Boolean) isConcaveBlock.value_(index)).booleanValue() && !contains) {
				return false;
			}

			if (((Boolean) isConvexBlock.value_(index)).booleanValue() && contains) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Answer true if the polygon contains the point.
	 * 
	 * @param anArrayOfPoint sra.jun.geometry.basic.Jun2dPoint[]
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 */
	protected static boolean PolygonFromPoints_containsPoint_(Jun2dPoint[] anArrayOfPoint, Jun2dPoint aPoint) {
		int size = anArrayOfPoint.length;
		double around = 0.0d;

		for (int i = 0; i < size; i++) {
			Jun2dPoint p1 = anArrayOfPoint[i];
			Jun2dPoint p2 = anArrayOfPoint[(i + 1) % size];
			around += aPoint.angleBetween_and_(p1, p2).rad();
		}

		return (Math.abs(around) > Math.PI / 4.0d);
	}

	/**
	 * Answer true if the receiver contains the line segment.
	 * 
	 * @param from sra.jun.geometry.basic.Jun2dPoint
	 * @param to sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 */
	public boolean absContainsLineSegmentFrom_to_(final Jun2dPoint from, final Jun2dPoint to) {
		final double dx = to.x() - from.x();
		final double dy = to.y() - from.y();
		final StBlockClosure tBlock = new StBlockClosure() {
			public Object value_value_(Object anObject1, Object anObject2) {
				Jun2dPoint point1 = (Jun2dPoint) anObject1;
				Jun2dPoint point2 = (Jun2dPoint) anObject2;
				double pdx = point2.x() - point1.x();
				double pdy = point2.y() - point1.y();
				double d = (pdx * dy) - (dx * pdy);

				if (d < ACCURACY) {
					return null;
				} else {
					return new Double((pdx * (point1.y() - from.y()) - pdy * (point1.x() - from.x())) / d);
				}
			}
		};

		final StSortedCollection intersectingTs = new StSortedCollection(new StBlockClosure() {
			public Object value_value_(Object e1, Object e2) {
				return new Boolean(((Number) e1).doubleValue() <= ((Number) e2).doubleValue());
			}
		});
		intersectingTs.add_(new Double(0));
		intersectingTs.add_(new Double(1));
		this.edgesDo_(new StBlockClosure() {
			public Object value_value_(Object p1, Object p2) {
				Double t = (Double) tBlock.value_value_(p1, p2);

				if ((t != null) && (0.0d < t.doubleValue()) && (t.doubleValue() < 1.0d)) {
					intersectingTs.add_(t);
				}

				return null;
			}
		});

		if (intersectingTs.size() == 2) {
			return this.absContainsPoint_((Jun2dPoint) from.plus_(to).dividedBy_(2));
		}

		Object[] ts = intersectingTs._asArray();

		for (int i = 0; i < (ts.length - 1); i++) {
			double t1 = ((Number) ts[i]).doubleValue();
			double t2 = ((Number) ts[i + 1]).doubleValue();
			double t = (t1 + t2) / 2.0d;
			Jun2dPoint intersectingPoint = (Jun2dPoint) to.minus_(from).multipliedBy_(t).plus_(from);

			if (this.absContainsPoint_(intersectingPoint) == false) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Answer true if the receiver contains the point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 */
	public boolean absContainsPoint_(Jun2dPoint aPoint) {
		if (this.boundingBox().containsOrTouchesPoint_(aPoint) == false) {
			return false;
		}

		final double x = aPoint.x();
		final double y = aPoint.y();
		final StValueHolder count = new StValueHolder(0);
		Object result = this.edgesDo_(new StBlockClosure() {
			public Object value_value_(Object anObject1, Object anObject2) {
				Jun2dPoint p1 = (Jun2dPoint) anObject1;
				Jun2dPoint p2 = (Jun2dPoint) anObject2;

				if ((Math.abs(p1.x() - x) < ACCURACY) && (Math.abs(p1.y() - y) < ACCURACY)) {
					return Boolean.TRUE;
				}

				Jun2dPoint point1;
				Jun2dPoint point2;

				if (p1.x() <= p2.x()) {
					point1 = p1;
					point2 = p2;
				} else {
					point1 = p2;
					point2 = p1;
				}

				if ((point1.x() <= x) && (x < point2.x())) {
					double y0 = ((x - point1.x()) / (point2.x() - point1.x()) * (point2.y() - point1.y())) + point1.y();

					if (Math.abs(y0 - y) < ACCURACY) {
						return Boolean.TRUE;
					}

					if (y0 >= y) {
						count.value_(count._intValue() + 1);
					}
				}

				return null;
			}
		});

		if (result != null) {
			return true;
		}

		return (count._intValue() % 2 == 1);
	}

	/**
	 * Answer true if the receiver is a convex polygon.
	 * 
	 * @return boolean
	 */
	public boolean absIsConvex() {
		return this.asPositivePolygon().isConvex();
	}

	/**
	 * Answer a number which represents my area.
	 * 
	 * @return double
	 */
	public double area() {
		return Math.abs(this.signedArea());
	}

	/**
	 * Enumerate every convex polygons and evaluate the block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return java.lang.Object
	 */
	public Object asConvexPolygonsDo_(final StBlockClosure aBlock) {
		if (this.isNegative()) {
			return this.asPositivePolygon().asConvexPolygonsDo_(new StBlockClosure() {
				public Object value_(Object anObject) {
					Jun2dPolygon polygon = (Jun2dPolygon) anObject;

					return aBlock.value_(polygon.asNegativePolygon());
				}
			});
		}

		if (this.isConvex()) {
			return aBlock.value_(this);
		}

		int size = this.pointsSize();
		int concaveIndex = this.indexOfFirstConcaveVertex();
		Jun2dPoint concavePoint = this.pointAt_(concaveIndex);
		Jun2dPoint divisionVector = (Jun2dPoint) concavePoint.minus_(this.pointAt_((concaveIndex - 1 + size) % size)).unitVector();
		divisionVector = (Jun2dPoint) divisionVector.plus_(concavePoint.minus_(this.pointAt_((concaveIndex + 1) % size)).unitVector());

		int divisionIndex;
		Jun2dPoint divisionPoint;

		if (Math.abs(divisionVector.x()) > Math.abs(divisionVector.y())) {
			if (divisionVector.x() > 0) {
				Jun2dPoint minPoint = null;
				int minIndex = -1;

				for (int i = concaveIndex; i <= ((concaveIndex + size) - 3); i++) {
					Jun2dPoint p1 = this.pointAt_((i + 1) % size);
					Jun2dPoint p2 = this.pointAt_((i + 2) % size);

					if ((p1.y() < p2.y()) && (p1.y() < concavePoint.y()) && (p2.y() >= concavePoint.y())) {
						double x = (((concavePoint.y() - p1.y()) * (p2.x() - p1.x())) / (p2.y() - p1.y())) + p1.x();

						if ((concavePoint.x() < x) && (minPoint == null || x <= minPoint.x())) {
							minPoint = new Jun2dPoint(x, concavePoint.y());
							minIndex = (i + 1) % size;
						}
					}
				}

				if (minIndex < 0) {
					throw SmalltalkException.Error("internal error");
				}

				divisionIndex = minIndex;
				divisionPoint = minPoint;
			} else {
				Jun2dPoint maxPoint = null;
				int maxIndex = -1;

				for (int i = concaveIndex; i <= ((concaveIndex + size) - 3); i++) {
					Jun2dPoint p1 = this.pointAt_((i + 1) % size);
					Jun2dPoint p2 = this.pointAt_((i + 2) % size);

					if ((p1.y() > p2.y()) && (p1.y() > concavePoint.y()) && (p2.y() <= concavePoint.y())) {
						double x = (((concavePoint.y() - p1.y()) * (p2.x() - p1.x())) / (p2.y() - p1.y())) + p1.x();

						if ((concavePoint.x() > x) && (maxPoint == null || x >= maxPoint.x())) {
							maxPoint = new Jun2dPoint(x, concavePoint.y());
							maxIndex = (i + 1) % size;
						}
					}
				}

				if (maxIndex < 0) {
					throw SmalltalkException.Error("internal error");
				}

				divisionIndex = maxIndex;
				divisionPoint = maxPoint;
			}
		} else {
			if (divisionVector.y() > 0) {
				Jun2dPoint minPoint = null;
				int minIndex = -1;

				for (int i = concaveIndex; i <= ((concaveIndex + size) - 3); i++) {
					Jun2dPoint p1 = this.pointAt_((i + 1) % size);
					Jun2dPoint p2 = this.pointAt_((i + 2) % size);

					if ((p1.x() > p2.x()) && (p1.x() > concavePoint.x()) && (p2.x() <= concavePoint.x())) {
						double y = (((concavePoint.x() - p1.x()) * (p2.y() - p1.y())) / (p2.x() - p1.x())) + p1.y();

						if ((concavePoint.y() < y) && (minPoint == null || y <= minPoint.y())) {
							minPoint = new Jun2dPoint(concavePoint.x(), y);
							minIndex = (i + 1) % size;
						}
					}
				}

				if (minIndex < 0) {
					throw SmalltalkException.Error("internal error");
				}

				divisionIndex = minIndex;
				divisionPoint = minPoint;
			} else {
				Jun2dPoint maxPoint = null;
				int maxIndex = -1;

				for (int i = concaveIndex; i <= ((concaveIndex + size) - 3); i++) {
					Jun2dPoint p1 = this.pointAt_((i + 1) % size);
					Jun2dPoint p2 = this.pointAt_((i + 2) % size);

					if ((p1.x() < p2.x()) && (p1.x() < concavePoint.x()) && (p2.x() >= concavePoint.x())) {
						double y = (((concavePoint.x() - p1.x()) * (p2.y() - p1.y())) / (p2.x() - p1.x())) + p1.y();

						if ((concavePoint.y() > y) && (maxPoint == null || y >= maxPoint.y())) {
							maxPoint = new Jun2dPoint(concavePoint.x(), y);
							maxIndex = (i + 1) % size;
						}
					}
				}

				if (maxIndex < 0) {
					throw SmalltalkException.Error("internal error");
				}

				divisionIndex = maxIndex;
				divisionPoint = maxPoint;
			}
		}

		Vector newPoints1 = new Vector();
		Vector newPoints2 = new Vector();

		if (divisionIndex < concaveIndex) {
			for (int i = 0; i <= divisionIndex; i++) {
				newPoints1.addElement(this.pointAt_(i));
			}

			if (divisionPoint.equals(this.pointAt_(divisionIndex)) == false) {
				newPoints1.addElement(divisionPoint);
			}

			for (int i = concaveIndex; i < size; i++) {
				newPoints1.addElement(this.pointAt_(i));
			}

			if (divisionPoint.equals(this.pointAt_((divisionIndex + 1) % size)) == false) {
				newPoints2.addElement(divisionPoint);
			}

			for (int i = divisionIndex + 1; i <= concaveIndex; i++) {
				newPoints2.addElement(this.pointAt_(i));
			}
		} else {
			for (int i = 0; i <= concaveIndex; i++) {
				newPoints1.addElement(this.pointAt_(i));
			}

			if (divisionPoint.equals(this.pointAt_((divisionIndex + 1) % size)) == false) {
				newPoints1.addElement(divisionPoint);
			}

			for (int i = divisionIndex + 1; i < size; i++) {
				newPoints1.addElement(this.pointAt_(i));
			}

			for (int i = concaveIndex; i <= divisionIndex; i++) {
				newPoints2.addElement(this.pointAt_(i));
			}

			if (divisionPoint.equals(this.pointAt_(divisionIndex)) == false) {
				newPoints2.addElement(divisionPoint);
			}
		}

		Object result = (new Jun2dPolygon(newPoints1)).asConvexPolygonsDo_(aBlock);

		if (result != null) {
			return result;
		}

		result = (new Jun2dPolygon(newPoints2)).asConvexPolygonsDo_(aBlock);

		if (result != null) {
			return result;
		}

		return null;
	}

	/**
	 * Convert the receiver to a JunOpenGL3dObject.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject asJunOpenGL3dObject() {
		final Vector polygons = new Vector();
		this.asJunOpenGL3dObjectsDo_(new StBlockClosure() {
			public Object value_(Object polygon) {
				polygons.add(polygon);

				return null;
			}
		});

		return (polygons.size() == 1) ? (JunOpenGL3dObject) polygons.firstElement() : JunOpenGL3dCompoundObject.Components_(polygons);
	}

	/**
	 * Enumerate the convex polygons as JunOpenGL3dObjects and evaluate the
	 * block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return java.lang.Object
	 */
	public Object asJunOpenGL3dObjectsDo_(final StBlockClosure aBlock) {
		return this.asConvexPolygonsDo_(new StBlockClosure() {
			public Object value_(Object anObject) {
				Jun2dPolygon polygon = (Jun2dPolygon) anObject;
				Jun2dPoint[] points = polygon.points();
				Jun3dPoint[] vertexes = new Jun3dPoint[points.length];

				for (int i = 0; i < points.length; i++) {
					vertexes[i] = new Jun3dPoint(points[i], 0);
				}

				return aBlock.value_(new JunOpenGL3dPolygon(vertexes));
			}
		});
	}

	/**
	 * Convert the receiver as a negative polygon.
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon asNegativePolygon() {
		if (this.signedArea() <= 0) {
			return this;
		} else {
			return this.reversed();
		}
	}

	/**
	 * Convert the receiver as a positive polygon.
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon asPositivePolygon() {
		if (this.signedArea() >= 0) {
			return this;
		} else {
			return this.reversed();
		}
	}

	/**
	 * Answer my bounding box.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 */
	public Jun2dBoundingBox boundingBox() {
		if (bounds == null) {
			Jun2dPoint min = points[0];
			Jun2dPoint max = points[0];

			for (int i = 1; i < points.length; i++) {
				min = (Jun2dPoint) min.min_(points[i]);
				max = (Jun2dPoint) max.max_(points[i]);
			}

			bounds = Jun2dBoundingBox.Origin_corner_(min, max);
		}

		return bounds;
	}

	/**
	 * Answer true if the receiver contains the point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return boolean
	 */
	public boolean containsPoint_(Jun2dPoint aPoint) {
		return (this.absContainsPoint_(aPoint) != this.isNegative());
	}

	/**
	 * Enumerate the edges and evaluate the block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return java.lang.Object
	 */
	public Object edgesDo_(StBlockClosure aBlock) {
		for (int i = 0; i < points.length; i++) {
			Object result = aBlock.value_value_(points[i], points[(i + 1) % points.length]);

			if (result != null) {
				return result;
			}
		}

		return null;
	}

	/**
	 * Answer the extent of my bounding box.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint extent() {
		return (Jun2dPoint) this.boundingBox().extent();
	}

	/**
	 * Answer the height of my bounding box.
	 * 
	 * @return double
	 */
	public double height() {
		return this.extent().y();
	}

	/**
	 * Answer true if the receiver is consistent.
	 * 
	 * @return boolean
	 */
	public boolean isConsistent() {
		return IsConsistentPolygonFromPoints_(this.points());
	}

	/**
	 * Answer true if the receiver is a convex polygon.
	 * 
	 * @return boolean
	 */
	public boolean isConvex() {
		if (this.isPositive()) {
			return (this.indexOfFirstConcaveVertex() < 0);
		} else {
			return (this.asPositivePolygon().indexOfFirstConcaveVertex() < 0);
		}
	}

	/**
	 * Answer true if the receiver is a negative polygon.
	 * 
	 * @return boolean
	 */
	public boolean isNegative() {
		return (this.isPositive() == false);
	}

	/**
	 * Answer true if the receiver is a positive polygon.
	 * 
	 * @return boolean
	 */
	public boolean isPositive() {
		return (this.signedArea() + this.ACCURACY >= 0);
	}

	/**
	 * Answer the nearest point on the receiver from the point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint nearestPointFromPoint_(Jun2dPoint aPoint) {
		if (this.containsPoint_(aPoint)) {
			return aPoint;
		} else {
			return this.nearestPointOnEdgeFromPoint_(aPoint);
		}
	}

	/**
	 * Answer the nearest point on the receiver's edge from the point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint nearestPointOnEdgeFromPoint_(Jun2dPoint aPoint) {
		final Jun2dPoint thePoint = new Jun2dPoint(aPoint.x(), aPoint.y());
		final StValueHolder nearestPoint = new StValueHolder(null);
		final StValueHolder minDistance = new StValueHolder(Double.POSITIVE_INFINITY);
		this.edgesDo_(new StBlockClosure() {
			public Object value_value_(Object anObject1, Object anObject2) {
				Jun2dPoint point1 = (Jun2dPoint) anObject1;
				Jun2dPoint point2 = (Jun2dPoint) anObject2;

				if (point1.equals(point2) == false) {
					Jun2dLine edge = new Jun2dLine(point1, point2);
					Jun2dPoint point = edge.lineSegmentNearestPointFromPoint_(thePoint);
					double distance = thePoint.distance_(point);

					if (minDistance._doubleValue() > distance) {
						minDistance.value_(distance);
						nearestPoint.value_(point);
					}
				}

				return null;
			}
		});

		return (Jun2dPoint) nearestPoint.value();
	}

	/**
	 * Answer the point at the specified index.
	 * 
	 * @param index int
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint pointAt_(int index) {
		return points[index];
	}

	/**
	 * Answer the array of my points.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint[]
	 */
	public Jun2dPoint[] points() {
		return points;
	}

	/**
	 * Answer the number of my points.
	 * 
	 * @return int
	 */
	public int pointsSize() {
		return points.length;
	}

	/**
	 * Create a reversed polygon.
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon reversed() {
		Jun2dPoint[] reversedPoints = new Jun2dPoint[points.length];

		for (int i = 0; i < points.length; i++) {
			reversedPoints[i] = points[points.length - i - 1];
		}

		return new Jun2dPolygon(reversedPoints);
	}

	/**
	 * Answer a copy of the receiver scaled by the specified extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon scaledBy_(Jun2dPoint aPoint) {
		Jun2dPoint[] scaledPoints = new Jun2dPoint[points.length];

		for (int i = 0; i < points.length; i++) {
			scaledPoints[i] = points[i].scaledBy_(aPoint);
		}

		return new Jun2dPolygon(scaledPoints);
	}

	/**
	 * Answer a number which represents my signed area.
	 * 
	 * @return double
	 */
	public double signedArea() {
		if (Double.isNaN(signedArea)) {
			signedArea = SignedAreaForPoints_(points);
		}

		return signedArea;
	}

	/**
	 * Answer a copy of the receiver translated by the specified extent.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon translatedBy_(Jun2dPoint aPoint) {
		Jun2dPoint[] scaledPoints = new Jun2dPoint[points.length];

		for (int i = 0; i < points.length; i++) {
			scaledPoints[i] = points[i].translatedBy_(aPoint);
		}

		return new Jun2dPolygon(scaledPoints);
	}

	/**
	 * Answer the trasposed polygon of the recevier.
	 * 
	 * @return sra.jun.geometry.surfaces.Jun2dPolygon
	 */
	public Jun2dPolygon transposed() {
		Jun2dPoint[] transposedPoints = new Jun2dPoint[points.length];

		for (int i = 0; i < points.length; i++) {
			transposedPoints[i] = new Jun2dPoint(points[i].y(), points[i].x());
		}

		return new Jun2dPolygon(transposedPoints);
	}

	/**
	 * Answer the width of my bounding box.
	 * 
	 * @return double
	 */
	public double width() {
		return this.extent().x();
	}

	/**
	 * Answer the index of the first concave vertex. If not found, return -1.
	 * 
	 * @return int
	 */
	protected int indexOfFirstConcaveVertex() {
		int pointsSize = this.pointsSize();

		for (int i = 0; i < pointsSize; i++) {
			Jun2dPoint p1 = this.pointAt_((i - 1 + pointsSize) % pointsSize);
			Jun2dPoint p2 = this.pointAt_(i);
			Jun2dPoint p3 = this.pointAt_((i + 1) % pointsSize);

			if ((SignedAreaForTriangle_and_and_(p1, p2, p3) + ACCURACY) < 0) {
				return i;
			}
		}

		return -1;
	}

	/**
	 * Set the points.
	 * 
	 * @param anArrayOfPoint sra.jun.geometry.basic.Jun2dPoint[]
	 */
	protected void setPoints_(Jun2dPoint[] anArrayOfPoint) {
		int size = anArrayOfPoint.length;
		points = new Jun2dPoint[size];
		System.arraycopy(anArrayOfPoint, 0, points, 0, size);
		signedArea = Double.NaN;
	}
}
