package sra.jun.geometry.surfaces;

import java.awt.Point;
import sra.jun.collections.sequences.*;
import sra.jun.geometry.abstracts.JunSurface;
import sra.jun.geometry.basic.Jun3dPoint;
import sra.jun.opengl.objects.JunOpenGL3dNurbsSurface;
import sra.jun.opengl.objects.JunOpenGL3dObject;
import sra.jun.opengl.support.JunOpenGLRenderingContext;

/**
 * JunNurbsSurface class
 * 
 *  @author    nisinaka
 *  @created   1998/12/18 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunNurbsSurface.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class JunNurbsSurface extends JunSurface {
	/** A JunMatrix of control points. */
	protected JunMatrix controlPoints = null;

	/** A JunDoubleMatrix of weights. */
	protected JunDoubleMatrix weights = null;

	/** A knot vector of u-parameter. */
	protected double[] uKnotVector = null;

	/** A knot vector of v-parameter. */
	protected double[] vKnotVector = null;

	/**
	 * Create a new instance of JunNurbsSurface and initialize it with the
	 * bezier control points.
	 * 
	 * @param aMatrix sra.jun.collections.sequences.JunMatrix
	 * 
	 * @return sra.jun.geometry.surfaces.JunNurbsSurface
	 */
	public static final JunNurbsSurface BezierControlPoints_(JunMatrix aMatrix) {
		int columnSize;
		int rowSize;
		double[] uKnotVector;
		double[] vKnotVector;

		columnSize = aMatrix.columnSize();
		rowSize = aMatrix.rowSize();
		uKnotVector = new double[rowSize * 2];

		for (int index = 0; index < rowSize; index++) {
			uKnotVector[index] = 0.0;
			uKnotVector[index + rowSize] = 1.0;
		}

		vKnotVector = new double[columnSize * 2];

		for (int index = 0; index < columnSize; index++) {
			vKnotVector[index] = 0.0;
			vKnotVector[index + columnSize] = 1.0;
		}

		return BSplineControlPoints_uKnotVector_vKnotVector_(aMatrix, uKnotVector, vKnotVector);
	}

	/**
	 * Create an instance of JunNurbsSurface and initialize it with the
	 * arguments.
	 * 
	 * @param aMatrix sra.jun.collections.sequences.JunMatrix
	 * @param anArray1 double[]
	 * @param anArray2 double[]
	 * 
	 * @return sra.jun.geometry.surfaces.JunNurbsSurface
	 */
	public static final JunNurbsSurface BSplineControlPoints_uKnotVector_vKnotVector_(JunMatrix aMatrix, double[] anArray1, double[] anArray2) {
		int rowSize;
		int columnSize;
		double[] weights;

		rowSize = aMatrix.rowSize();
		columnSize = aMatrix.columnSize();
		weights = new double[rowSize * columnSize];

		for (int i = 0; i < weights.length; i++) {
			weights[i] = 1;
		}

		return ControlPoints_weights_uKnotVector_vKnotVector_(aMatrix, new JunDoubleMatrix(rowSize, columnSize, weights), anArray1, anArray2);
	}

	/**
	 * Create a new instance of JunNurbsSurface and initialize it.
	 * 
	 * @param aMatrix1 sra.jun.collections.sequences.JunMatrix
	 * @param aMatrix2 sra.jun.collections.sequences.JunDoubleMatrix
	 * @param anArray1 double[]
	 * @param anArray2 double[]
	 * 
	 * @return sra.jun.geometry.surfaces.JunNurbsSurface
	 */
	public static final JunNurbsSurface ControlPoints_weights_uKnotVector_vKnotVector_(JunMatrix aMatrix1, JunDoubleMatrix aMatrix2, double[] anArray1, double[] anArray2) {
		JunNurbsSurface nurbsSurface = new JunNurbsSurface();
		nurbsSurface.controlPoints_(aMatrix1);
		nurbsSurface.weights_(aMatrix2);
		nurbsSurface.uKnotVector_(anArray1);
		nurbsSurface.vKnotVector_(anArray2);

		return nurbsSurface;
	}

	/**
	 * Convert the receiver as a JunOpenGL3dNurbsSurface.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dNurbsSurface
	 */
	public JunOpenGL3dNurbsSurface asJunOpenGL3dNurbsSurface() {
		return JunOpenGL3dNurbsSurface.ControlPoints_weights_uKnotVector_vKnotVector_(controlPoints, weights, uKnotVector, vKnotVector);
	}

	/**
	 * Convert the receiver as a JunOpenGL3dObject.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject asJunOpenGL3dObject() {
		return this.asJunOpenGL3dNurbsSurface();
	}

	/**
	 * Answer the control point at the specified point.
	 * 
	 * @param aPoint java.awt.Point
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint controlPointAt_(Point aPoint) {
		return (Jun3dPoint) controlPoints._at(aPoint.x, aPoint.y);
	}

	/**
	 * Set the control point at the specified point.
	 * 
	 * @param aPoint java.awt.Point
	 * @param a3dPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public void controlPointAt_put_(Point aPoint, Jun3dPoint a3dPoint) {
		controlPoints._put(aPoint.x, aPoint.y, a3dPoint);
	}

	/**
	 * Answer the control point at the specified point.
	 * 
	 * @param u int
	 * @param v int
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint controlPointAtU_v_(int u, int v) {
		return (Jun3dPoint) controlPoints._at(u, v);
	}

	/**
	 * Answer the JunMatrix of the contol points.
	 * 
	 * @return sra.jun.collections.sequences.JunMatrix
	 */
	public JunMatrix controlPoints() {
		return controlPoints;
	}

	/**
	 * Set the JunMatrix as the control points of the receiver.
	 * 
	 * @param aMatrix sra.jun.collections.sequences.JunMatrix
	 */
	public void controlPoints_(JunMatrix aMatrix) {
		controlPoints = aMatrix;
	}

	/**
	 * Answer the size of the control points.
	 * 
	 * @return java.awt.Point
	 */
	public Point controlPointSize() {
		return controlPoints.matrixSize();
	}

	/**
	 * Render the receiver on the JunOpenGLRenderingContext.
	 * 
	 * @param aRenderingContext
	 *        sra.jun.opengl.support.JunOpenGLRenderingContext
	 */
	public void renderOn_(JunOpenGLRenderingContext aRenderingContext) {
	}

	/**
	 * Enumerate the trim curves, but do nothing.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return java.lang.Object
	 */
	public Object trimCurvesDo_(Object anObject) {
		return null;
	}

	/**
	 * Answer the knot vector of the u-parameter.
	 * 
	 * @return double[]
	 */
	public double[] uKnotVector() {
		return uKnotVector;
	}

	/**
	 * Set the knot vector of the u-parameter.
	 * 
	 * @param anArray double[]
	 */
	public void uKnotVector_(double[] anArray) {
		uKnotVector = anArray;
	}

	/**
	 * Answer the order number in a row.
	 * 
	 * @return int
	 */
	public int uOrder() {
		return uKnotVector.length - this.uSize();
	}

	/**
	 * Answer the number of the control points in a row.
	 * 
	 * @return int
	 */
	public int uSize() {
		return controlPoints.rowSize();
	}

	/**
	 * Answer the knot vector of the v-parameter.
	 * 
	 * @return double[]
	 */
	public double[] vKnotVector() {
		return vKnotVector;
	}

	/**
	 * Set the knot vector of the v-parameter.
	 * 
	 * @param anArray double[]
	 */
	public void vKnotVector_(double[] anArray) {
		vKnotVector = anArray;
	}

	/**
	 * Answer the order number in a column.
	 * 
	 * @return int
	 */
	public int vOrder() {
		return vKnotVector.length - this.vSize();
	}

	/**
	 * Answer the number of the control points in a column.
	 * 
	 * @return int
	 */
	public int vSize() {
		return controlPoints.columnSize();
	}

	/**
	 * Answer the weight at the specified point.
	 * 
	 * @param u int
	 * @param v int
	 * 
	 * @return double
	 */
	public double weightAtU_v_(int u, int v) {
		return weights._doubleValueAt(u, v);
	}

	/**
	 * Answer the weights of the receiver.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix weights() {
		return weights;
	}

	/**
	 * Set the weights of the receiver.
	 * 
	 * @param aMatrix sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public void weights_(JunDoubleMatrix aMatrix) {
		weights = aMatrix;
	}
}
