package sra.jun.geometry.surfaces;

import java.io.*;
import sra.jun.geometry.abstracts.*;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.curves.Jun3dLine;
import sra.jun.geometry.transformations.*;
import sra.smalltalk.SmalltalkException;

/**
 * JunPlane class
 * 
 *  @author    nisinaka
 *  @created   1998/11/09 (by nisinaka)
 *  @updated   1999/08/04 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunPlane.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class JunPlane extends JunSurface {
	/** The parameters. */
	protected double a;

	/** The parameters. */
	protected double b;

	/** The parameters. */
	protected double c;

	/** The parameters. */
	protected double d;

	/** Three points to determine the plane. */
	protected Jun3dPoint p1;

	/** Three points to determine the plane. */
	protected Jun3dPoint p2;

	/** Three points to determine the plane. */
	protected Jun3dPoint p3;

	/**
	 * Create a new instance of JunPlane with three Jun3dPoints. This
	 * corresponds to the instance creation method "on:on:on:".
	 * 
	 * @param aPoint1 sra.jun.geometry.basic.Jun3dPoint
	 * @param aPoint2 sra.jun.geometry.basic.Jun3dPoint
	 * @param aPoint3 sra.jun.geometry.basic.Jun3dPoint
	 */
	public JunPlane(Jun3dPoint aPoint1, Jun3dPoint aPoint2, Jun3dPoint aPoint3) {
		this._initialize(aPoint1, aPoint2, aPoint3);
	}

	/**
	 * The default constructor.
	 */
	protected JunPlane() {
	}

	/**
	 * Create a new instance of the JunPlane with a point and a line.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.surfaces.JunPlane
	 */
	public static final JunPlane On_vertical_(Jun3dPoint aPoint, Jun3dLine aLine) {
		JunPlane aPlane = new JunPlane();
		aPlane.p1 = aPoint;
		aPlane.a = aLine.f();
		aPlane.b = aLine.g();
		aPlane.c = aLine.h();
		aPlane.d = - ((aLine.f() * aPoint.x()) + (aLine.g() * aPoint.y()) + aLine.h() * aPoint.z());

		Jun3dTransformation aT;

		while ((aPoint == null) || aPoint.equals(aPlane.p1)) {
			aT = Jun3dTransformation.Translate_(new Jun3dPoint(0, 1, 0));
			aLine = (Jun3dLine) aLine.transform_(aT);
			aPoint = aPlane.intersectingPointWithLine_(aLine);
		}

		aPlane.p2 = aPoint;

		while ((aPoint == null) || aPoint.equals(aPlane.p2)) {
			aT = Jun3dTransformation.Translate_(new Jun3dPoint(1, 0, 0));
			aLine = (Jun3dLine) aLine.transform_(aT);
			aPoint = aPlane.intersectingPointWithLine_(aLine);
		}

		aPlane.p3 = aPoint;

		return aPlane;
	}

	/**
	 * Answer the parameter a.
	 * 
	 * @return double
	 */
	public double a() {
		return a;
	}

	/**
	 * Answer the angle with the Jun3dLine.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.JunAngle
	 */
	public JunAngle angleWithLine_(Jun3dLine aLine) {
		double f = aLine.f();
		double g = aLine.g();
		double h = aLine.h();
		double denominator = Math.sqrt(((a * a) + (b * b) + c * c) * ((f * f) + (g * g) + h * h));

		if (denominator < ACCURACY) {
			throw SmalltalkException.Error("unexpected line or plane parameters");
		}

		double numerator = (a * f) + (b * g) + (c * h);
		double gamma = Math.min(Math.max(numerator / denominator, -1), 1);

		return JunAngle.FromRad_((Math.PI / 2.0d) - Math.acos(gamma));
	}

	/**
	 * Answer the parameter b.
	 * 
	 * @return double
	 */
	public double b() {
		return b;
	}

	/**
	 * Answer the parameter c.
	 * 
	 * @return double
	 */
	public double c() {
		return c;
	}

	/**
	 * Answer true if the receiver contains the Jun3dPoint, otherwise false.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return boolean
	 */
	public boolean containsPoint_(Jun3dPoint aPoint) {
		return Math.abs(this.valueF_(aPoint)) < ACCURACY;
	}

	/**
	 * Answer the parameter d.
	 * 
	 * @return double
	 */
	public double d() {
		return d;
	}

	/**
	 * Answer the distance from the Point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 */
	public double distanceFromPoint_(Jun3dPoint aPoint) {
		double denominator = (a * a) + (b * b) + (c * c);

		if ((denominator - ACCURACY) < 0) {
			throw SmalltalkException.Error("unexpect plane parameters");
		}

		double numerator = (a * aPoint.x()) + (b * aPoint.y()) + (c * aPoint.z()) + d;
		numerator = numerator * numerator;

		return Math.sqrt(numerator / denominator);
	}

	/**
	 * Answer the Point which intersects with the Line.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint intersectingPointWithLine_(Jun3dLine aLine) {
		double f = aLine.f();
		double g = aLine.g();
		double h = aLine.h();
		double denominator = (a * f) + (b * g) + (c * h);

		if (Math.abs(denominator) < aLine.ACCURACY) {
			// The receiver parallels the Line.
			return null;
		}

		double x0 = aLine.x0();
		double y0 = aLine.y0();
		double z0 = aLine.z0();
		double numerator = - ((a * x0) + (b * y0) + (c * z0) + d);
		Jun3dPoint thePoint = new Jun3dPoint(x0 + ((f * numerator) / denominator), y0 + ((g * numerator) / denominator), z0 + ((h * numerator) / denominator));

		return thePoint;
	}

	/**
	 * Answer the intersecting point with the line segment Jun3dLine.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint intersectingPointWithLineSegment_(Jun3dLine aLine) {
		Jun3dPoint thePoint = this.intersectingPointWithLine_(aLine);

		if ((thePoint == null) || !aLine.lineSegmentContainsPoint_(thePoint)) {
			return null;
		}

		return thePoint;
	}

	/**
	 * Answer the normal unit vector of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint normalUnitVector() {
		double dd = Math.sqrt((a * a) + (b * b) + (c * c));

		return new Jun3dPoint(a / dd, b / dd, c / dd);
	}

	/**
	 * Answer the normal vector of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint normalVector() {
		return new Jun3dPoint(a, b, c);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("3dPlane (");
		p1.printOn_(aWriter);
		aWriter.write(", ");
		p2.printOn_(aWriter);
		aWriter.write(", ");
		p3.printOn_(aWriter);
		aWriter.write(")");
	}

	/**
	 * Answer the projection of the Jun3dLine on the receiver.
	 * 
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.curves.Jun3dLine
	 */
	public Jun3dLine projectionOfLine_(Jun3dLine aLine) {
		Jun3dPoint from = (Jun3dPoint) aLine.from();
		Jun3dPoint to = (Jun3dPoint) aLine.to();

		return new Jun3dLine(this.projectionOfPoint_(from), this.projectionOfPoint_(to));
	}

	/**
	 * Answer the projection of the Jun3dPoint on the receiver.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint projectionOfPoint_(Jun3dPoint aPoint) {
		double denominator = (a * a) + (b * b) + (c * c);

		if (Math.abs(denominator) < ACCURACY) {
			this.normalizeParameters();
			denominator = (a * a) + (b * b) + (c * c);

			if (Math.abs(denominator) < ACCURACY) {
				throw SmalltalkException.Error("can not define a plane");
			}
		}

		double numerator = - ((a * aPoint.x()) + (b * aPoint.y()) + (c * aPoint.z()) + d);
		double x = aPoint.x() + ((a * numerator) / denominator);
		double y = aPoint.y() + ((b * numerator) / denominator);
		double z = aPoint.z() + ((c * numerator) / denominator);

		return new Jun3dPoint(x, y, z);
	}

	/**
	 * Apply the JunTransformation to this JunGeometry. This method is not
	 * defined in Smalltalk version.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.JunTransformation
	 * 
	 * @return sra.jun.geometry.abstracts.JunGeometry
	 */
	public JunGeometry transform_(JunTransformation aTransformation) {
		Object[] arguments = { p1.transform_(aTransformation), p2.transform_(aTransformation), p3.transform_(aTransformation)};

		return (JunGeometry) _New(this.getClass(), arguments);
	}

	/**
	 * Answer the 'f' value of the receiver with the Jun3dPoint.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return double
	 */
	public double valueF_(Jun3dPoint aPoint) {
		return (a * aPoint.x()) + (b * aPoint.y()) + (c * aPoint.z()) + d;
	}

	/**
	 * Initialize the receiver's parameters with three Jun3dPoints.
	 * 
	 * @param aPoint1 sra.jun.geometry.basic.Jun3dPoint
	 * @param aPoint2 sra.jun.geometry.basic.Jun3dPoint
	 * @param aPoint3 sra.jun.geometry.basic.Jun3dPoint
	 */
	protected final void _initialize(Jun3dPoint aPoint1, Jun3dPoint aPoint2, Jun3dPoint aPoint3) {
		this.p1 = aPoint1;
		this.p2 = aPoint2;
		this.p3 = aPoint3;

		Jun3dPoint thePoint2 = (Jun3dPoint) aPoint2.minus_(aPoint1);
		Jun3dPoint thePoint3 = (Jun3dPoint) aPoint3.minus_(aPoint1);
		this.a = (thePoint2.y() * thePoint3.z()) - (thePoint2.z() * thePoint3.y());
		this.b = (thePoint2.z() * thePoint3.x()) - (thePoint2.x() * thePoint3.z());
		this.c = (thePoint2.x() * thePoint3.y()) - (thePoint2.y() * thePoint3.x());
		this.d = - ((aPoint1.x() * this.a) + (aPoint1.y() * this.b) + aPoint1.z() * this.c);
	}

	/**
	 * Normalize the parameters.
	 */
	protected void normalizeParameters() {
		double max = Math.max(Math.max(Math.abs(a), Math.abs(b)), Math.abs(c));
		a = a / max;
		b = b / max;
		c = c / max;
		d = d / max;
	}
}
