package sra.jun.geometry.transformations;

import java.io.*;
import sra.jun.collections.sequences.*;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.curves.Jun3dLine;
import sra.jun.opengl.objects.JunOpenGL3dObject;

/**
 * Jun3dTransformation class
 * 
 *  @author    nisinaka
 *  @created   1998/10/09 (by nisinaka)
 *  @updated   2000/01/06 (by nisinaka)
 *  @updated   2003/07/01 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: Jun3dTransformation.java,v 6.0 2003/10/28 10:52:06 nisinaka Exp $
 */
public class Jun3dTransformation extends JunTransformation {

	public double a;
	public double b;
	public double c;
	public double p;
	public double d;
	public double e;
	public double f;
	public double q;
	public double g;
	public double h;
	public double i;
	public double r;
	public double l;
	public double m;
	public double n;
	public double s;

	/**
	 * Create a new transformation which translate the alignmentPoint to the
	 * relativePoint.
	 * 
	 * @param alignmentPoint sra.jun.geometry.basic.JunPoint
	 * @param relativePoint sra.jun.geometry.basic.JunPoint
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Align_with_(Jun3dPoint alignmentPoint, Jun3dPoint relativePoint) {
		return Jun3dTransformation.Translate_(relativePoint.minus_(alignmentPoint));
	}

	/**
	 * Create a new Jun3dTransformation which aligns the specified points.
	 * 
	 * @param anArrayOfPointArray sra.jun.geometry.basic.Jun3dPoint[][]
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation AlignPoints_(Jun3dPoint[][] anArrayOfPointArray) {
		int size = anArrayOfPointArray.length;
		double[] xEquations = new double[size * 5];
		double[] yEquations = new double[size * 5];
		double[] zEquations = new double[size * 5];
		int index = 0;

		for (int i = 0; i < size; i++) {
			double[] points = anArrayOfPointArray[i][0].asArray();
			xEquations[index] = yEquations[index] = zEquations[index] = points[0];
			index++;
			xEquations[index] = yEquations[index] = zEquations[index] = points[1];
			index++;
			xEquations[index] = yEquations[index] = zEquations[index] = points[2];
			index++;
			xEquations[index] = yEquations[index] = zEquations[index] = 1.0d;
			index++;

			Jun3dPoint aPoint = anArrayOfPointArray[i][1];
			xEquations[index] = -aPoint.x();
			yEquations[index] = -aPoint.y();
			zEquations[index] = -aPoint.z();
			index++;
		}

		double[] xSolution = (new JunLinearEquations(size, 5, xEquations)).solution();
		double[] ySolution = (new JunLinearEquations(size, 5, yEquations)).solution();
		double[] zSolution = (new JunLinearEquations(size, 5, zEquations)).solution();
		Jun3dTransformation aTransformation = new Jun3dTransformation();
		aTransformation.a = xSolution[0];
		aTransformation.b = ySolution[0];
		aTransformation.c = zSolution[0];
		aTransformation.d = xSolution[1];
		aTransformation.e = ySolution[1];
		aTransformation.f = zSolution[1];
		aTransformation.g = xSolution[2];
		aTransformation.h = ySolution[2];
		aTransformation.i = zSolution[2];
		aTransformation.l = xSolution[3];
		aTransformation.m = ySolution[3];
		aTransformation.n = zSolution[3];
		aTransformation.p = 0.0d;
		aTransformation.q = 0.0d;
		aTransformation.r = 0.0d;
		aTransformation.s = 1.0d;

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun3dTransformation and initialize it with an
	 * array.
	 * 
	 * @param anArray double[]
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation FromArray_(double[] anArray) {
		Jun3dTransformation aTransformation = new Jun3dTransformation();
		aTransformation.a = anArray[0];
		aTransformation.b = anArray[1];
		aTransformation.c = anArray[2];
		aTransformation.p = anArray[3];
		aTransformation.d = anArray[4];
		aTransformation.e = anArray[5];
		aTransformation.f = anArray[6];
		aTransformation.q = anArray[7];
		aTransformation.g = anArray[8];
		aTransformation.h = anArray[9];
		aTransformation.i = anArray[10];
		aTransformation.r = anArray[11];
		aTransformation.l = anArray[12];
		aTransformation.m = anArray[13];
		aTransformation.n = anArray[14];
		aTransformation.s = anArray[15];

		return aTransformation;
	}

	/**
	 * Typical transformation - Perspective.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Perspective_(Jun3dPoint aPoint) {
		Jun3dTransformation aTransformation = Unity();
		aTransformation.a = -aPoint.z();
		aTransformation.e = -aPoint.z();
		aTransformation.g = aPoint.x();
		aTransformation.h = aPoint.y();
		aTransformation.i = -1;
		aTransformation.r = 1;
		aTransformation.s = -aPoint.z();

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the z axis.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Rotate_(JunAngle anAngle) {
		return RotateZ_(anAngle);
	}

	/**
	 * Typical transformation - Rotate around a line.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * @param aLine sra.jun.geometry.curves.Jun3dLine
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Rotate_around_(JunAngle anAngle, Jun3dLine aLine) {
		JunAngle theAngle = (JunAngle) anAngle.copy();
		Jun3dPoint firstPoint = (Jun3dPoint) aLine.first().copy();
		Jun3dPoint lastPoint = (Jun3dPoint) aLine.last().copy();
		Jun3dPoint aVector = (Jun3dPoint) lastPoint.minus_(firstPoint);
		double vx = aVector.x();
		double vy = aVector.y();
		double vz = aVector.z();
		double aV = Math.sqrt((vy * vy) + (vz * vz));
		double aL = Math.sqrt((vx * vx) + (vy * vy) + (vz * vz));
		Jun3dTransformation aT = Translate_(firstPoint.negated());
		Jun3dTransformation aRx = Unity();

		if (aV != 0.0d) {
			aRx.e = vz / aV;
			aRx.f = vy / aV;
			aRx.h = -vy / aV;
			aRx.i = vz / aV;
		}

		Jun3dTransformation aRy = Unity();

		if (aL != 0.0d) {
			aRy.a = aV / aL;
			aRy.c = vx / aL;
			aRy.g = -vx / aL;
			aRy.i = aV / aL;
		}

		Jun3dTransformation aRz = Rotate_(theAngle);
		Jun3dTransformation rRy = Unity();

		if (aL != 0.0d) {
			rRy.a = aV / aL;
			rRy.c = -vx / aL;
			rRy.g = vx / aL;
			rRy.i = aV / aL;
		}

		Jun3dTransformation rRx = Unity();

		if (aV != 0.0d) {
			rRx.e = vz / aV;
			rRx.f = -vy / aV;
			rRx.h = vy / aV;
			rRx.i = vz / aV;
		}

		Jun3dTransformation rT = Translate_(firstPoint);
		Jun3dTransformation aTransformation = aT.product_(aRx);
		aTransformation = aTransformation.product_(aRy);
		aTransformation = aTransformation.product_(aRz);
		aTransformation = aTransformation.product_(rRy);
		aTransformation = aTransformation.product_(rRx);
		aTransformation = aTransformation.product_(rT);

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the x axis.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation RotateX_(JunAngle anAngle) {
		double cos = anAngle.cos();
		double sin = anAngle.sin();
		Jun3dTransformation aTransformation = Unity();
		aTransformation.e = cos;
		aTransformation.f = sin;
		aTransformation.h = -sin;
		aTransformation.i = cos;

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the y axis.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation RotateY_(JunAngle anAngle) {
		double cos = anAngle.cos();
		double sin = anAngle.sin();
		Jun3dTransformation aTransformation = Unity();
		aTransformation.a = cos;
		aTransformation.c = -sin;
		aTransformation.g = sin;
		aTransformation.i = cos;

		return aTransformation;
	}

	/**
	 * Typical transformation - Rotate around the z axis.
	 * 
	 * @param anAngle sra.jun.geometry.basic.JunAngle
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation RotateZ_(JunAngle anAngle) {
		double cos = anAngle.cos();
		double sin = anAngle.sin();
		Jun3dTransformation aTransformation = Unity();
		aTransformation.a = cos;
		aTransformation.b = sin;
		aTransformation.d = -sin;
		aTransformation.e = cos;

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun3dTransformation with the specified scale.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Scale_(double aNumber) {
		Jun3dTransformation aTransformation = Unity();
		aTransformation.a = aNumber;
		aTransformation.e = aNumber;
		aTransformation.i = aNumber;

		return aTransformation;
	}

	/**
	 * Create a new instance of JunTransformation with the specified scale.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Scale_(Jun3dPoint aPoint) {
		Jun3dTransformation aTransformation = Unity();
		aTransformation.a = aPoint.x();
		aTransformation.e = aPoint.y();
		aTransformation.i = aPoint.z();

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun3dTransformation with the specified
	 * translation.
	 * 
	 * @param aNumber double
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Translate_(double aNumber) {
		Jun3dTransformation aTransformation = Unity();
		aTransformation.l = aNumber;
		aTransformation.m = aNumber;
		aTransformation.n = aNumber;

		return aTransformation;
	}

	/**
	 * Create a new instance of Jun3dTransformation with the specified
	 * translation.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Translate_(Jun3dPoint aPoint) {
		Jun3dTransformation aTransformation = Unity();
		aTransformation.l = aPoint.x();
		aTransformation.m = aPoint.y();
		aTransformation.n = aPoint.z();

		return aTransformation;
	}

	/**
	 * Answer a new Jun3dTransformation with unit values.
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Unity() {
		Jun3dTransformation aTransformation = new Jun3dTransformation();
		aTransformation.a = 1;
		aTransformation.b = 0;
		aTransformation.c = 0;
		aTransformation.p = 0;
		aTransformation.d = 0;
		aTransformation.e = 1;
		aTransformation.f = 0;
		aTransformation.q = 0;
		aTransformation.g = 0;
		aTransformation.h = 0;
		aTransformation.i = 1;
		aTransformation.r = 0;
		aTransformation.l = 0;
		aTransformation.m = 0;
		aTransformation.n = 0;
		aTransformation.s = 1;

		return aTransformation;
	}

	/**
	 * Answer a new Jun3dTransformation with zero values.
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public static Jun3dTransformation Zero() {
		Jun3dTransformation aTransformation = new Jun3dTransformation();
		aTransformation.a = 0;
		aTransformation.b = 0;
		aTransformation.c = 0;
		aTransformation.p = 0;
		aTransformation.d = 0;
		aTransformation.e = 0;
		aTransformation.f = 0;
		aTransformation.q = 0;
		aTransformation.g = 0;
		aTransformation.h = 0;
		aTransformation.i = 0;
		aTransformation.r = 0;
		aTransformation.l = 0;
		aTransformation.m = 0;
		aTransformation.n = 0;
		aTransformation.s = 0;

		return aTransformation;
	}

	/**
	 * Apply this transformation to aPoint and return the result.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint applyTo_(Jun3dPoint aPoint) {
		return aPoint.transform_(this);
	}

	/**
	 * Apply this transformation to anObject and return the result.
	 * 
	 * @param anObject sra.jun.opengl.objects.JunOpenGL3dObject
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject applyTo_(JunOpenGL3dObject anObject) {
		return anObject.transform_(this);
	}

	/**
	 * Apply this transformation to aPoint and return the result.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint applyToPoint_(Jun3dPoint aPoint) {
		double nx;
		double ny;
		double nz;
		double scale;
		nx = (aPoint.x() * a) + (aPoint.y() * d) + (aPoint.z() * g) + l;
		ny = (aPoint.x() * b) + (aPoint.y() * e) + (aPoint.z() * h) + m;
		nz = (aPoint.x() * c) + (aPoint.y() * f) + (aPoint.z() * i) + n;
		scale = (aPoint.x() * p) + (aPoint.y() * q) + (aPoint.z() * r) + s;

		if (scale == 0.0d) {
			scale = ACCURACY;
		}

		return new Jun3dPoint(nx / scale, ny / scale, nz / scale);
	}

	/**
	 * Convert this transformation as an array.
	 * 
	 * @return double[]
	 */
	public double[] asArray() {
		double[] array = new double[16];
		array[0] = a;
		array[1] = b;
		array[2] = c;
		array[3] = p;
		array[4] = d;
		array[5] = e;
		array[6] = f;
		array[7] = q;
		array[8] = g;
		array[9] = h;
		array[10] = i;
		array[11] = r;
		array[12] = l;
		array[13] = m;
		array[14] = n;
		array[15] = s;

		return array;
	}

	/**
	 * Convert this transformation as a matrix.
	 * 
	 * @return sra.jun.collections.sequences.JunDoubleMatrix
	 */
	public JunDoubleMatrix asMatrix() {
		return new JunDoubleMatrix(4, this.asArray());
	}

	/**
	 * Answer true if aTransformation is equal to this point, otherwise false.
	 * 
	 * @param anObject java.lang.Object
	 * 
	 * @return boolean
	 */
	public boolean equals(Object anObject) {
		if (anObject instanceof Jun3dTransformation == false) {
			return false;
		}

		Jun3dTransformation aTransformation = (Jun3dTransformation) anObject;

		return (this.a == aTransformation.a)
			&& (this.b == aTransformation.b)
			&& (this.c == aTransformation.c)
			&& (this.p == aTransformation.p)
			&& (this.d == aTransformation.d)
			&& (this.e == aTransformation.e)
			&& (this.f == aTransformation.f)
			&& (this.q == aTransformation.q)
			&& (this.g == aTransformation.g)
			&& (this.h == aTransformation.h)
			&& (this.i == aTransformation.i)
			&& (this.r == aTransformation.r)
			&& (this.l == aTransformation.l)
			&& (this.m == aTransformation.m)
			&& (this.n == aTransformation.n)
			&& (this.s == aTransformation.s);
	}

	/**
	 * Answer the inverse of this transformation.
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public Jun3dTransformation inverse() {
		return Jun3dTransformation.FromArray_(this.asMatrix().inverse().asArrayOfDouble());
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write("3dTransformation (");
		aWriter.write(a + " " + b + " " + c + " " + p);
		aWriter.write(" ,  ");
		aWriter.write(d + " " + e + " " + f + " " + q);
		aWriter.write(" ,  ");
		aWriter.write(g + " " + h + " " + i + " " + r);
		aWriter.write(" ,  ");
		aWriter.write(l + " " + m + " " + n + " " + s);
		aWriter.write(")");
	}

	/**
	 * Answer a new transformation producted by another instance of
	 * Jun3dTransformation.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.Jun3dTransformation
	 * 
	 * @return sra.jun.geometry.transformations.Jun3dTransformation
	 */
	public Jun3dTransformation product_(Jun3dTransformation aTransformation) {
		Jun3dTransformation newTransformation = new Jun3dTransformation();
		newTransformation.a = (a * aTransformation.a) + (b * aTransformation.d) + (c * aTransformation.g) + (p * aTransformation.l);
		newTransformation.b = (a * aTransformation.b) + (b * aTransformation.e) + (c * aTransformation.h) + (p * aTransformation.m);
		newTransformation.c = (a * aTransformation.c) + (b * aTransformation.f) + (c * aTransformation.i) + (p * aTransformation.n);
		newTransformation.p = (a * aTransformation.p) + (b * aTransformation.q) + (c * aTransformation.r) + (p * aTransformation.s);
		newTransformation.d = (d * aTransformation.a) + (e * aTransformation.d) + (f * aTransformation.g) + (q * aTransformation.l);
		newTransformation.e = (d * aTransformation.b) + (e * aTransformation.e) + (f * aTransformation.h) + (q * aTransformation.m);
		newTransformation.f = (d * aTransformation.c) + (e * aTransformation.f) + (f * aTransformation.i) + (q * aTransformation.n);
		newTransformation.q = (d * aTransformation.p) + (e * aTransformation.q) + (f * aTransformation.r) + (q * aTransformation.s);
		newTransformation.g = (g * aTransformation.a) + (h * aTransformation.d) + (i * aTransformation.g) + (r * aTransformation.l);
		newTransformation.h = (g * aTransformation.b) + (h * aTransformation.e) + (i * aTransformation.h) + (r * aTransformation.m);
		newTransformation.i = (g * aTransformation.c) + (h * aTransformation.f) + (i * aTransformation.i) + (r * aTransformation.n);
		newTransformation.r = (g * aTransformation.p) + (h * aTransformation.q) + (i * aTransformation.r) + (r * aTransformation.s);
		newTransformation.l = (l * aTransformation.a) + (m * aTransformation.d) + (n * aTransformation.g) + (s * aTransformation.l);
		newTransformation.m = (l * aTransformation.b) + (m * aTransformation.e) + (n * aTransformation.h) + (s * aTransformation.m);
		newTransformation.n = (l * aTransformation.c) + (m * aTransformation.f) + (n * aTransformation.i) + (s * aTransformation.n);
		newTransformation.s = (l * aTransformation.p) + (m * aTransformation.q) + (n * aTransformation.r) + (s * aTransformation.s);

		return newTransformation;
	}

	/**
	 * Answer the rotation angle of the transformation.
	 *
	 * @return sra.jun.geometry.basic.JunAngle
	 */
	public JunAngle rotationAngle() {
		return JunAngle.FromRad_(Math.acos((a + e + i - 1) / 2));
	}

	/**
	 * Answer the rotation vector of the transformation.
	 *
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint rotationVector() {
		double denominator = Math.sqrt((f - h) * (f - h) + (c - g) * (c - g) + (b - d) * (b - d));
		return new Jun3dPoint((h - f) / denominator, (c - g) / denominator, (d - b) / denominator);
	}
}
