package sra.jun.goodies.files;

import java.awt.Cursor;
import java.io.File;
import java.util.*;
import sra.smalltalk.*;
import sra.jun.goodies.cursors.JunCursors;
import sra.jun.system.framework.JunApplicationModel;

/**
 * JunFileModel class
 * 
 *  @author    nisinaka
 *  @created   2003/03/05 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on Jun433 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunFileModel.java,v 6.0 2003/10/28 10:52:09 nisinaka Exp $
 */
public class JunFileModel extends JunApplicationModel implements StBlockValued {

	/**
	 * The FileType class defines a file type name with its extensions.
	 */
	public static class FileType {
		protected String name;
		protected String[] extensions;

		public FileType(String nameString, String[] extensionStrings) {
			name = nameString;
			extensions = extensionStrings;
		}
		public static FileType All(String nameString) {
			return new FileType(nameString, null);
		}
		public String name() {
			return name;
		}
		public String[] extensions() {
			return extensions;
		}
	}

	protected final static String DefaultDirectoryItem = "Default directory";

	protected StValueHolder fileName;
	protected StValueHolder fileList;
	protected StValueHolder fileType;
	protected JunFileModel.FileType[] fileTypes;
	protected File currentDirectory;
	protected StValueHolder currentDirectoryLabel;
	protected int currentDirectoryLevel;
	protected StValueHolder directoryList;
	protected StValueHolder currentDrive;
	protected File[] drives;

	/**
	 * Answer the default directory.
	 *
	 * @return java.io.File
	 */
	public static File DefaultDirectory() {
		return new File(System.getProperty("user.dir"));
	}

	/**
	 * Answer a StBlockValue that computes aBlock with the receiver's value as the argument.
	 *
	 * @return sra.smalltalk.StBlockValue
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public StBlockValue compute_(StBlockClosure aBlock) {
		return new StBlockValue(aBlock, this);
	}

	/**
	 * Answer the current directory.
	 *
	 * @return java.io.File
	 */
	public File currentDirectory() {
		if (currentDirectory == null) {
			this.currentDirectory_(DefaultDirectory());
		}
		return currentDirectory;
	}

	/**
	 * Set the new directory as a current directory.
	 *
	 * @param aFile java.io.File
	 */
	public void currentDirectory_(File aFile) {
		File newDirectory = null;

		if (aFile == null) {
			newDirectory = DefaultDirectory();
		} else {
			newDirectory = aFile;
			if (newDirectory.exists() && newDirectory.isDirectory()) {
				if (newDirectory.isAbsolute() == false) {
					newDirectory = newDirectory.getAbsoluteFile();
				}
			} else {
				newDirectory = DefaultDirectory();
			}
		}

		currentDirectory = newDirectory;
		currentDirectoryLevel = -1;
		this.updateCurrentDrive();
		this.updateCurrentDirectoryLabel();
		this.updateDirectoryAndFileList();
	}

	/**
	 * Answer the value holder for the current directory label.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder currentDirectoryLabel() {
		if (currentDirectoryLabel == null) {
			currentDirectoryLabel = new StValueHolder(new String());
		}
		return currentDirectoryLabel;
	}

	/**
	 * Answer the current directory level.
	 *
	 * @return int
	 */
	public int currentDirectoryLevel() {
		if (currentDirectoryLevel < 0) {
			currentDirectoryLevel = 0;
			File aFile = this.currentDirectory();
			while (aFile != null) {
				currentDirectoryLevel++;
				aFile = aFile.getParentFile();
			}
		}
		return currentDirectoryLevel;
	}

	/**
	 * Answer the value holder for the current drive.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder currentDrive() {
		if (currentDrive == null) {
			currentDrive = new StValueHolder(null);
			this.updateCurrentDrive();
		}
		return currentDrive;
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunFileViewAwt(this);
		} else {
			return new JunFileViewSwing(this);
		}
	}

	/**
	 * Answer the value holder for the directory list.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder directoryList() {
		if (directoryList == null) {
			directoryList = new StValueHolder(new String[0]);
		}
		return directoryList;
	}

	/**
	 * Answer the array of all drives.
	 *
	 * @return java.io.File[]
	 */
	public File[] drives() {
		return File.listRoots();
	}

	/**
	 * Answer the value holder for the file list.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder fileList() {
		if (fileList == null) {
			fileList = new StValueHolder(new String[0]);
		}
		return fileList;
	}

	/**
	 * Answer the value holder for the file name.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder fileName() {
		if (fileName == null) {
			fileName = new StValueHolder(new String());
		}
		return fileName;
	}

	/**
	 * Answer the value holder for the file type.
	 *
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder fileType() {
		if (fileType == null) {
			fileType = new StValueHolder(this.fileTypes()[0]);
		}
		return fileType;
	}

	/**
	 * Answer the current file types.
	 *
	 * @return sra.jun.goodies.files.JunFileModel.FileType[]
	 */
	public FileType[] fileTypes() {
		if (fileTypes == null) {
			fileTypes = this.defaultFileTypes();
		}
		return fileTypes;
	}

	/**
	 * Set the new file types.
	 *
	 * @param newFileTypes sra.jun.goodies.files.JunFileModel.FileType[]
	 */
	public void fileTypes_(FileType[] newFileTypes) {
		fileTypes = newFileTypes;
	}

	/**
	 * Answer the current file.
	 *
	 * @return java.lang.Object
	 */
	public Object value() {
		return new File(this.currentDirectory(), (String) this.fileName().value());
	}

	/**
	 * Set the file as an initial value.
	 *
	 * @param aFile java.io.File
	 */
	public void value_(File aFile) {
		if (aFile == null) {
			return;
		}

		if (aFile.exists()) {
			if (aFile.isDirectory()) {
				this.currentDirectory_(aFile);
			} else {
				this.currentDirectory_(aFile.getParentFile());
				this.fileName().value_(aFile.getName());
			}
		} else {
			this.fileName().value_(aFile.getName());
			File directory = aFile.getParentFile();
			while (directory != null && directory.exists() == false) {
				directory = directory.getParentFile();
			}
			this.currentDirectory_(directory);
		}
	}

	/**
	 * Answer true if the filename string matches the current file type, otherwise false.
	 *
	 * @return boolean
	 * @param filenameString java.lang.String
	 */
	protected boolean checkFileTypeMatch_(String filenameString) {
		FileType aFileType = (FileType) this.fileType().value();
		if (aFileType == null) {
			return true;
		}

		String[] extensions = aFileType.extensions();
		if (extensions == null) {
			return true;
		}

		for (int i = 0; i < extensions.length; i++) {
			if (filenameString.endsWith(extensions[i])) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Called when the current drive is changed.
	 *
	 * @param directoryName java.lang.String
	 */
	protected void currentDriveChanged(String directoryName) {
		if (DefaultDirectoryItem.equals(directoryName)) {
			this.currentDirectory_(DefaultDirectory());
		} else {
			this.currentDirectory_(new File(directoryName));
		}
		this.changed_($("value"));
	}

	/**
	 * Answer the current file types.
	 *
	 * @return sra.jun.goodies.files.JunFileModel.FileType[]
	 */
	protected JunFileModel.FileType[] defaultFileTypes() {
		return new JunFileModel.FileType[] { JunFileModel.FileType.All("All files (*)")};
	}

	/**
	 * Called when the item on the directory list is double clicked.
	 *
	 * @param index int
	 * @param aString java.lang.String
	 */
	protected void directoryListDoubleClicked(int index, String aString) {
		String name = aString.trim();
		File directory = this.currentDirectory();
		if (index >= this.currentDirectoryLevel()) {
			this.currentDirectory_(new File(directory, name));
		} else {
			for (int i = this.currentDirectoryLevel() - 1; i > index; i--) {
				directory = directory.getParentFile();
			}
			this.currentDirectory_(directory);
		}
		this.changed_($("value"));
	}

	/**
	 * Called when a file is selected on a file list.
	 *
	 * @param fileName java.lang.String
	 */
	protected void fileListChanged(String fileName) {
		this.fileName().value_(fileName);
		this.changed_($("value"));
	}

	/**
	 * Called when the file name is changed.
	 *
	 * @param fileName java.lang.String
	 */
	protected void fileNameChanged(String fileName) {
		this.fileName().value_(fileName);
		this.changed_($("value"));
	}

	/**
	 * Called when a file type is changed.
	 *
	 * @param index int
	 */
	protected void fileTypeChanged(int index) {
		this.fileType().value_(this.fileTypes()[index]);
		this.updateDirectoryAndFileList();
		this.changed_($("value"));
	}

	/**
	 * Initialize the ApplicationModel when created.
	 */
	protected void initialize() {
		super.initialize();

		fileName = null;
		fileList = null;
		fileType = null;
		fileTypes = null;
		currentDirectory = null;
		currentDirectoryLabel = null;
		currentDirectoryLevel = -1;
		directoryList = null;
		fileList = null;
		currentDrive = null;
		drives = null;
	}

	/**
	 * Update the current directory label.
	 */
	protected void updateCurrentDirectoryLabel() {
		int maxSize = 32;
		String directoryString = this.currentDirectory().getPath();
		if (directoryString.length() > maxSize) {
			directoryString = directoryString.substring(0, 3) + "..." + directoryString.substring(directoryString.length() - maxSize + 5, directoryString.length());
		}
		this.currentDirectoryLabel().value_(directoryString);
	}

	/**
	 * Update the current drive.
	 */
	protected void updateCurrentDrive() {
		File directory = this.currentDirectory();
		while (directory.getParent() != null) {
			directory = directory.getParentFile();
		}
		this.currentDrive().value_(directory);
	}

	/**
	 * Update the directory list and the file list.
	 */
	protected void updateDirectoryAndFileList() {
		JunCursors cursor = new JunCursors(new Cursor(Cursor.WAIT_CURSOR));
		try {
			cursor._show();

			String listingName;
			LinkedList dirs1 = new LinkedList();
			File directory = this.currentDirectory();
			while (directory != null) {
				listingName = directory.getName();
				if (listingName.length() == 0) {
					listingName = directory.getPath();
				}
				dirs1.addFirst(listingName);
				directory = directory.getParentFile();
			}

			LinkedList dirs2 = new LinkedList();
			LinkedList files = new LinkedList();
			File[] directoryContents = this.currentDirectory().listFiles();
			for (int i = 0; i < directoryContents.length; i++) {
				File aFile = directoryContents[i];
				listingName = aFile.getName();
				if (aFile.canRead()) {
					if (aFile.isDirectory()) {
						dirs2.add(listingName);
					} else {
						if (this.checkFileTypeMatch_(listingName)) {
							files.add(listingName);
						}
					}
				}
			}

			Collections.sort(dirs2);
			dirs1.addAll(dirs2);
			this.directoryList().value_(dirs1.toArray());

			Collections.sort(files);
			this.fileList().value_(files.toArray());

		} finally {
			cursor._restore();
		}
	}
}
