package sra.jun.goodies.font;

import java.awt.*;
import java.io.*;
import sra.smalltalk.*;
import sra.jun.goodies.display.JunDisplayModel;
import sra.jun.system.framework.JunApplicationModel;

/**
 * JunFontModel class
 * 
 *  @author    nisinaka
 *  @created   2003/02/28 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on Jun433 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunFontModel.java,v 6.0 2003/10/28 10:52:09 nisinaka Exp $
 */
public class JunFontModel extends JunApplicationModel {

	protected JunDisplayModel displayModel;
	protected String[] fontNames;
	protected String fontName;
	protected int[] fontSizes;
	protected int fontSize;

	/**
	 * JunFontModel constructor comment.
	 */
	public JunFontModel() {
		super();
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunFontViewAwt(this);
		} else {
			return new JunFontViewSwing(this);
		}
	}

	/**
	 * Answer my current display model.
	 *
	 * @return sra.jun.goodies.display.JunDisplayModel
	 */
	public JunDisplayModel displayModel() {
		if (displayModel == null) {
			displayModel = new JunDisplayModel();
		}
		return displayModel;
	}

	/**
	 * Answer the current font name.
	 *
	 * @return java.lang.String
	 */
	public String fontName() {
		return fontName;
	}

	/**
	 * Set the new font name.
	 *
	 * @param newFontName java.lang.String
	 */
	public void fontName_(String newFontName) {
		fontName = newFontName;
	}

	/**
	 * Answer the current font names.
	 *
	 * @return java.lang.String[]
	 */
	public String[] fontNames() {
		if (fontNames == null) {
			fontNames = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();
		}
		return fontNames;
	}

	/**
	 * Answer the current font size.
	 *
	 * @return int
	 */
	public int fontSize() {
		return fontSize;
	}

	/**
	 * Set the new font size.
	 *
	 * @param newFontSize int
	 */
	public void fontSize_(int newFontSize) {
		fontSize = newFontSize;
	}

	/**
	 * Answer the current font sizes.
	 *
	 * @return int[]
	 */
	public int[] fontSizes() {
		if (fontSizes == null) {
			fontSizes = this.defaultFontSizes();
		}
		return fontSizes;
	}

	/**
	 * Answer the current font.
	 *
	 * @return java.awt.Font
	 */
	public Font getFont() {
		return new Font(this.fontName(), Font.PLAIN, this.fontSize());
	}

	/**
	 * Update the visual object when opened.
	 *
	 * @return java.awt.Frame
	 * @param aView sra.smalltalk.StView
	 */
	protected Frame allButOpenView_(StView aView) {
		Frame aFrame = super.allButOpenView_(aView);
		this.updateVisualObject();
		return aFrame;
	}

	/**
	 * Answer the default font name.
	 *
	 * @return java.lang.String
	 */
	protected String defaultFontName() {
		return "Dialog";
	}

	/**
	 * Answer the default font size.
	 *
	 * @return int
	 */
	protected int defaultFontSize() {
		return 12;
	}

	/**
	 * Answer the default font sizes.
	 *
	 * @return int[]
	 */
	protected int[] defaultFontSizes() {
		return new int[] { 9, 10, 12, 14, 18, 24, 36, 48, 72 };
	}

	/**
	 * Answer the default text.
	 *
	 * @return java.lang.String
	 */
	protected String defaultText() {
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		try {
			pw.print("abcdefghijklmnopqrstuvwxyz 0123456789_~!\"'|\\,");
			pw.println();
			pw.print("ABCDEFGHIJKLMNOPQRSTUVWXYZ +-*/=<>{}()#$%&@^.");
		} finally {
			pw.flush();
			pw.close();
		}
		return sw.toString();
	}

	/**
	 * Answer the index of the current font name.
	 *
	 * @return int
	 */
	protected int getFontNameIndex() {
		int index = -1;
		String fontName = this.fontName();
		String[] fontNames = this.fontNames();
		for (int i = 0; i < fontNames.length; i++) {
			if (fontName.equals(fontNames[i])) {
				index = i;
				break;
			}
		}
		return index;
	}

	/**
	 * Answer the index of the current font size.
	 *
	 * @return int
	 */
	protected int getFontSizeIndex() {
		int index = -1;
		int fontSize = this.fontSize();
		int[] fontSizes = this.fontSizes();
		for (int i = 0; i < fontSizes.length; i++) {
			if (fontSize == fontSizes[i]) {
				index = i;
				break;
			}
		}
		return index;
	}

	/**
	 * Initialize this object.
	 */
	protected void initialize() {
		super.initialize();

		fontName = this.defaultFontName();
		fontSize = this.defaultFontSize();
	}

	/**
	 * Set the new index of the font name.
	 *
	 * @param newFontNameIndex int
	 */
	protected void setFontNameIndex(int newFontNameIndex) {
		fontName = this.fontNames()[newFontNameIndex];
		this.updateVisualObject();
	}

	/**
	 * Set the new index of the font size.
	 *
	 * @param newFontSizeIndex int
	 */
	protected void setFontSizeIndex(int newFontSizeIndex) {
		fontSize = this.fontSizes()[newFontSizeIndex];
		this.updateVisualObject();
	}

	/**
	 * Update the visual object.
	 */
	protected void updateVisualObject() {
		this.displayModel().visual_(new StComposedText(this.defaultText(), this.getFont()));
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "Font Model";
	}
}
