package sra.jun.goodies.progress;

import java.awt.*;
import sra.smalltalk.*;
import sra.jun.goodies.button.JunButtonModel;
import sra.jun.system.framework.JunApplicationModel;

/**
 * JunProgress class
 * 
 *  @author    nisinaka
 *  @created   1998/11/06 (by nisinaka)
 *  @updated   1999/08/05 (by nisinaka)
 *  @updated   2003/01/07 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunProgress.java,v 6.0 2003/10/28 10:52:13 nisinaka Exp $
 */
public class JunProgress extends JunApplicationModel {

	/** The progress value. */
	protected float progressValue;

	/** The progress message. */
	protected String progressMessage;

	/** The stop condition. */
	protected boolean stopCondition;

	/** The raising flag. */
	protected boolean raisingFlag;

	protected Object _result;
	protected Dialog _dialog;
	protected JunButtonModel _stopButton;

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		return this.defaultViewWithNoButton();
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultViewWithNoButton() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return JunProgressViewAwt.WithNoButton_(this);
		} else {
			return JunProgressViewSwing.WithNoButton_(this);
		}
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultViewWithStopButton() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return JunProgressViewAwt.WithStopButton_(this);
		} else {
			return JunProgressViewSwing.WithStopButton_(this);
		}
	}

	/**
	 * Evaluate the BlockClosure.
	 * 
	 * @return java.lang.Object
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public Object do_(final StBlockClosure aBlock) {
		(new Thread() {
			public void run() {
				_result = run_(aBlock);
			}
		}).start();
		this.openWithStopButton_(false);
		return _result;
	}

	/**
	 * Evaluate the BlockClosure.
	 * 
	 * @return java.lang.Object
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public Object doWithStopButton_(final StBlockClosure aBlock) {
		(new Thread() {
			public void run() {
				_result = run_(aBlock);
			}
		}).start();
		this.openWithStopButton_(true);
		return _result;
	}

	/**
	 * Answer true if the receiver has a stop button, otherwise false.
	 *
	 * @return boolean
	 */
	public boolean hasStopButton() {
		return (_stopButton != null);
	}

	/**
	 * Answer true if one of the progress view is open, otherwise false.
	 *
	 * @return boolean
	 */
	public boolean isOpen() {
		if (this.builder() == null) {
			return false;
		}

		Window[] windows = this.builder().windows();
		if (windows.length == 0) {
			return false;
		}

		for (int i = 0; i < windows.length; i++) {
			if (windows[i].isShowing()) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Answer the current progress message.
	 * 
	 * @return java.lang.String
	 */
	public String message() {
		return progressMessage;
	}

	/**
	 * Set the new progress message.
	 * 
	 * @param aString java.lang.String
	 */
	public void message_(String aString) {
		String oldString = progressMessage;
		progressMessage = aString;
		if (oldString.equals(progressMessage) == false) {
			this.changed_($("message"));
		}

		Thread.yield();
	}

	/**
	 * Open a view with a stop button if a boolean is true, otherwise a view with no button.
	 * Use modal dialog to avoid blocking EventDispatchThread.
	 *
	 * @param aBoolean boolean
	 */
	public void openWithStopButton_(boolean aBoolean) {
		stopCondition = false;
		StView aView = aBoolean ? this.defaultViewWithStopButton() : this.defaultViewWithNoButton();

		_dialog = new Dialog(new Frame(), this.windowTitle(), true);
		_dialog.add(aView.toComponent());
		_dialog.pack();
		_ShowAtMousePoint(_dialog);
	}

	/**
	 * Answer true if the stop button is pressed, otherwise false.
	 *
	 * @return boolean
	 */
	public boolean stopButtonPressed() {
		return stopCondition;
	}

	/**
	 * Answer the current progress value.
	 * 
	 * @return float
	 */
	public float value() {
		return progressValue;
	}

	/**
	 * Set the new progress value.
	 * 
	 * @param normalizedNumber float
	 */
	public void value_(float normalizedNumber) {
		if ((0 <= normalizedNumber) && (normalizedNumber <= 1)) {
			float oldValue = progressValue;
			progressValue = normalizedNumber;
			if (oldValue != progressValue) {
				this.changed_($("value"));
			}
		}

		Thread.yield();

		if (this.stopButtonPressed()) {
			throw new RuntimeException("stop");
		}
	}

	/**
	 * Close the progress.
	 */
	protected void close() {
		if (_dialog != null) {
			_dialog.dispose();
			_dialog = null;
		}

		stopCondition = true;
	}

	/**
	 * Answer one of the progress views.
	 * 
	 * @return sra.jun.goodies.progress.JunProgressView
	 */
	protected JunProgressView getProgressView() {
		Object[] dependents = this.dependents();
		for (int i = 0; i < dependents.length; i++) {
			if (dependents[i] instanceof JunProgressView) {
				return (JunProgressView) dependents[i];
			}
		}

		return null;
	}

	/**
	 * Initialize this object.
	 */
	protected void initialize() {
		super.initialize();

		progressValue = 0;
		progressMessage = new String();
		raisingFlag = false;
		_result = null;
		_dialog = null;
		_stopButton = null;
	}

	/**
	 * Answer the current raising flag.
	 * 
	 * @return boolean
	 */
	protected boolean raisingFlag() {
		return raisingFlag;
	}

	/**
	 * Evaluate the block while in progress.
	 * 
	 * @return java.lang.Object
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	protected Object run_(StBlockClosure aBlock) {
		Object result = null;

		try {
			if (aBlock.numArgs() == 0) {
				result = aBlock.value();
			} else {
				result = aBlock.value_(this);
			}
		} catch (RuntimeException e) {
			if (e.getMessage().equals("stop") == false) {
				throw e;
			}
		} finally {
			this.close();
		}

		return result;
	}

	/**
	 * Answer my stop button.
	 * If there is no stop button, return null.
	 *
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	protected JunButtonModel stopButton() {
		if (_stopButton == null) {
			_stopButton = new JunButtonModel(false, "Stop", new StBlockClosure() {
				public Object value() {
					close();
					return null;
				}
			});
		}
		return _stopButton;
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "Progress";
	}
}
