package sra.jun.goodies.track;

import java.io.*;
import sra.smalltalk.*;

/**
 * JunTrackSliderModel class
 * 
 *  @author    MATSUDA Ryouichi
 *  @created   1998/11/19 (by MATSUDA Ryouichi)
 *  @updated   1999/12/10 (by MATSUDA Ryouichi)
 *  @updated   2002/10/28 (by nisinaka)
 *  @updated   2003/03/24 (by nisinaka)
 *  @version   406b (with StPL6.0) based on Jun475 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunTrackSliderModel.java,v 6.0 2003/10/28 10:52:14 nisinaka Exp $
 */
public class JunTrackSliderModel extends StApplicationModel implements StBlockValued {

	protected StValueHolder valueHolder;
	protected StValueHolder intervalHolder;
	protected double[] intervalBounds;
	protected boolean fixFirstMarker;
	protected boolean fixLastMarker;
	protected boolean useMarkers;
	protected StMenu enabledPopupMenu;
	protected StMenu disabledPopupMenu;
	protected JunTrackerModel parentTracker;

	/**
	 * Create a new instance of JunTrackSliderModel.
	 */
	public JunTrackSliderModel() {
		this.initialize();
	}

	/**
	 * Answer true if the markers are in active, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean areMarkersActive() {
		return useMarkers;
	}

	/**
	 * Answer a child tracker if exist.
	 * 
	 * @return JunTrackSliderModel or null
	 */
	public JunTrackSliderModel childTracker() {
		return null;
	}

	/**
	 * Answer a BlockValue.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return sra.smalltalk.StBlockValue
	 */
	public StBlockValue compute_(StBlockClosure aBlock) {
		return new StBlockValue(aBlock, this);
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunTrackSliderViewAwt(this);
		} else {
			return new JunTrackSliderViewSwing(this);
		}
	}

	/**
	 * Menu message: Disable markers.
	 */
	public void disableMarkers() {
		useMarkers = false;
		if (this.parentTracker() == null) {
			this.changed_($("useMarkers"));
		}
	}

	/**
	 * Answer my value as double.
	 * 
	 * @return java.lang.Object
	 */
	public double doubleValue() {
		return ((Number) this.value()).doubleValue();
	}

	/**
	 * Menu message: Enable markers.
	 */
	public void enableMarkers() {
		useMarkers = true;
		if (this.parentTracker() == null) {
			this.changed_($("useMarkers"));
		}
	}

	/**
	 * Answer my current first marker.
	 * 
	 * @return double
	 */
	public double firstMarker() {
		return this.interval()[0];
	}

	/**
	 * Set the first marker.
	 * 
	 * @param normalizedValue
	 */
	public void firstMarker_(double normalizedValue) {
		if (this.fixFirstMarker() == false) {
			double first = Math.min(normalizedValue, this.lastMarker());
			first = Math.max(first, this.intervalBounds()[0]);
			first = Math.min(first, this.intervalBounds()[1]);
			this.interval_(new double[] { first, this.lastMarker()});
		}
	}

	/**
	 * Answer true if the first marker is needed to be fixed, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean fixFirstMarker() {
		return fixFirstMarker;
	}

	/**
	 * Set whether to fix the first marker or not.
	 *
	 * @param aBoolean boolean
	 */
	public void fixFirstMarker_(boolean aBoolean) {
		fixFirstMarker = aBoolean;
	}

	/**
	 * Answer true if the last marker is needed to be fixed, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean fixLastMarker() {
		return fixLastMarker;
	}

	/**
	 * Set whether to fix the last marker or not.
	 *
	 * @param aBoolean boolean
	 */
	public void fixLastMarker_(boolean aBoolean) {
		fixLastMarker = aBoolean;
	}

	/**
	 * Menu message: Flush my markers.
	 */
	public void flushMarkers() {
		this.lastMarker_(0);
	}

	/**
	 * Answer my current interval as an array of double.
	 * 
	 * @return double[]
	 */
	public double[] interval() {
		return (double[]) this.intervalHolder().value();
	}

	/**
	 * Set the new interval.
	 * 
	 * @param anArray
	 */
	public void interval_(double[] anArray) {
		this.intervalHolder().value_(this.adjustInterval_(anArray));
		this.changed_($("interval"));
	}

	/**
	 * Answer my current interval holder.
	 * 
	 * @return StValueHolder
	 */
	public StValueHolder intervalHolder() {
		if (intervalHolder == null) {
			intervalHolder = new StValueHolder(new double[] { 0, 0 });
		}
		return intervalHolder;
	}

	/**
	 * Answer true if the current interval is not empty, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean isIntervalNotEmpty() {
		return this.areMarkersActive() && (this.firstMarker() < this.lastMarker());
	}

	/**
	 * Answer my current last marker.
	 * 
	 * @return double
	 */
	public double lastMarker() {
		return this.interval()[1];
	}

	/**
	 * Set the last marker.
	 * 
	 * @param normalizedValue
	 */
	public void lastMarker_(double normalizedValue) {
		if (this.fixLastMarker() == false) {
			double last = Math.max(normalizedValue, this.firstMarker());
			last = Math.max(last, this.intervalBounds()[0]);
			last = Math.min(last, this.intervalBounds()[1]);
			this.interval_(new double[] { this.firstMarker(), last });
		}
	}

	/**
	 * Answer my current parent tracker.
	 * 
	 * @return sra.jun.goodies.track.JunTrackerModel
	 */
	public JunTrackerModel parentTracker() {
		return parentTracker;
	}

	/**
	 * Set my new parent tracker.
	 * 
	 * @param aTracker
	 */
	public void parentTracker_(JunTrackerModel aTracker) {
		parentTracker = aTracker;
		aTracker.childTracker_(this);
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * @throws IOException
	 */
	public void printOn_(Writer aWriter) throws IOException {
		super.printOn_(aWriter);
		aWriter.write(" (");
		aWriter.write(String.valueOf(this.value()));
		aWriter.write(')');
	}

	/**
	 * Menu message: Select entire movie with markers.
	 */
	public void selectAll() {
		this.lastMarker_(1);
		this.firstMarker_(0);
	}

	/**
	 * Menu message: Set the first marker.
	 */
	public void setFirstMarker() {
		this.firstMarker_(this.doubleValue());
	}

	/**
	 * Menu message: Set the last marker.
	 */
	public void setLastMarker() {
		this.lastMarker_(this.doubleValue());
	}

	/**
	 * Get the value of valueHolder.
	 * 
	 * @return java.lang.Object
	 */
	public Object value() {
		return valueHolder.value();
	}

	/**
	 * Set the value of valueHolder.
	 * 
	 * @param normalizedNumber double
	 */
	public void value_(double normalizedNumber) {
		valueHolder.value_(Math.max(0, Math.min(normalizedNumber, 1)));
		this.changed_($("value"));
	}

	/**
	 * Get the valueHolder with double value.
	 * 
	 * @return sra.smalltalk.StValueHolder
	 */
	public StValueHolder valueHolder() {
		if (valueHolder == null) {
			valueHolder = new StValueHolder(0.0d);
		}
		return valueHolder;
	}

	/**
	 * Answer the default StMenu for the disabled popup menu.
	 * 
	 * @return StMenu
	 */
	protected StMenu _disabledPopupMenu() {
		StMenuModel rootMenu = new StMenuModel();
		StMenu popupMenu = new StMenu("DisabledPopupMenu");
		popupMenu.add(this.createMenuItem("Enable markers", "enableMarkers"));
		rootMenu.add(popupMenu);
		return popupMenu;
	}

	/**
	 * Answer the default StMenu for the enabled popup menu.
	 * 
	 * @return StMenu
	 */
	protected StMenu _enabledPopupMenu() {
		StMenuModel rootMenu = new StMenuModel();
		StMenu popupMenu = new StMenu("EnabledPopupMenu");
		popupMenu.add(this.createMenuItem("Select all", "selectAll"));
		popupMenu.addSeparator();
		popupMenu.add(this.createMenuItem("Set first marker", "setFirstMarker"));
		popupMenu.add(this.createMenuItem("Set last marker", "setLastMarker"));
		popupMenu.add(this.createMenuItem("Flush markers", "flushMarkers"));
		popupMenu.addSeparator();
		popupMenu.add(this.createMenuItem("Disable markers", "disableMarkers"));
		rootMenu.add(popupMenu);
		return popupMenu;
	}

	/**
	 * Adjust the interval.
	 * 
	 * @param anArray double[]
	 * @return double[]
	 */
	protected double[] adjustInterval_(double[] anArray) {
		double first = Math.min(anArray[0], anArray[1]);
		double last = Math.max(anArray[0], anArray[1]);
		first = Math.max(0, Math.min(first, 1));
		last = Math.max(first, Math.min(last, 1));
		return new double[] { first, last };
	}

	/**
	 * Initialize the JunTrackSliderModel.
	 */
	protected void initialize() {
		super.initialize();

		valueHolder = null;
		intervalHolder = null;
		intervalBounds = null;
		fixFirstMarker = false;
		fixLastMarker = false;
		useMarkers = false;
		enabledPopupMenu = null;
		disabledPopupMenu = null;
		parentTracker = null;
		this.valueHolder();
		this.intervalHolder();
	}

	/**
	 * Answer my current interval bounds.
	 * 
	 * @return double[]
	 */
	protected double[] intervalBounds() {
		if (intervalBounds == null) {
			intervalBounds = new double[] { 0, 1 };
		}
		return intervalBounds;
	}

	/**
	 * Set my new interval bounds.
	 *
	 * @param anArray double[]
	 */
	protected void intervalBounds_(double[] anArray) {
		intervalBounds = this.adjustInterval_(anArray);
		this.changed_($("intervalBounds"));
	}

	/**
	 * Answer the StMenu for the popup menu.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu popupMenu() {
		if (this.parentTracker() != null) {
			return this.parentTracker().popupMenu();
		}
		if (this.areMarkersActive()) {
			if (enabledPopupMenu == null) {
				enabledPopupMenu = this._enabledPopupMenu();
			}
			StMenuItem menuItem = enabledPopupMenu.getItemWithValue("flushMarkers");
			if (menuItem != null) {
				menuItem.setEnabled(this.isIntervalNotEmpty());
			}
			return enabledPopupMenu;
		} else {
			if (disabledPopupMenu == null) {
				disabledPopupMenu = this._disabledPopupMenu();
			}
			return disabledPopupMenu;
		}
	}
}
