package sra.jun.opengl.chart;

import java.awt.*;
import java.awt.event.MouseEvent;
import java.io.*;
import java.util.*;
import sra.jun.geometry.basic.*;
import sra.jun.opengl.display.JunOpenGLDisplayModel;
import sra.jun.opengl.objects.*;
import sra.jun.opengl.texture.JunOpenGLTexture;
import sra.smalltalk.*;

/**
 * JunChartAbstract class
 * 
 *  @author    nisinaka
 *  @created   1998/11/19 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunChartAbstract.java,v 6.0 2003/10/28 10:52:17 nisinaka Exp $
 */
public abstract class JunChartAbstract extends JunOpenGLDisplayModel {
	/** The chart data. */
	protected JunChartData data = null;

	/** The hashtable which keeps all visual objects. */
	protected Hashtable visualObjectDictionary = null;

	/** The JunOpenGL3dObject for x-axis. */
	protected JunOpenGL3dObject xAxis = null;

	/** The JunOpenGL3dObject for y-axis. */
	protected JunOpenGL3dObject yAxis = null;

	/** The JunOpenGL3dObject for z-axis. */
	protected JunOpenGL3dObject zAxis = null;

	/** The color for axis. */
	protected Color axisColor = null;

	/** The foreground color for a label. */
	protected Color labelForegroundColor = null;

	/** The background color for a label. */
	protected Color labelBackgroundColor = null;

	/** The interval between charts. */
	protected double intervalBetweenCharts = 0.3;

	/** The thickness of each chart element. */
	protected double thickness = 0.1;

	/** The colors to paint each chart element. */
	protected Color[] colors = null;

	/** The index to point the current color. */
	protected int colorIndex = 0;

	/** The extra 3d objects. */
	protected Vector extra3dObjects = null;

	protected StMenu _axesMenu;

	/**
	 * The default constructor.
	 */
	protected JunChartAbstract() {
		this.initialize();
	}

	/**
	 * Create a new instance of a Chart and initialize it with the collection
	 * of sample. This method corresponds to the class method "samples:" of
	 * JunChartAbstract in Smalltalk.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 */
	protected JunChartAbstract(Vector aCollectionOfSample) {
		this();
		this.data_(new JunChartData(aCollectionOfSample));
	}

	/**
	 * Create a new instance of a Chart and initialize it with the collection
	 * of sample and a number of keys. This method corresponds to the class
	 * method "samples:numberOfKeys:" of JunChartAbstract in Smalltalk.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 * @param numberOfKeys int
	 */
	protected JunChartAbstract(Vector aCollectionOfSample, int numberOfKeys) {
		this();
		this.data_(new JunChartData(aCollectionOfSample, numberOfKeys));
	}

	/**
	 * Create a new instance of a Chart and initialize it with the
	 * JunChartData. This method corresponds to the class method "data:" of
	 * JunChartAbstract in Smalltalk.
	 * 
	 * @param aChartData sra.jun.opengl.chart.JunChartData
	 */
	protected JunChartAbstract(JunChartData aChartData) {
		this();
		this.data_(aChartData);
	}

	/**
	 * Answer the class for the chart data of this chart.
	 * 
	 * @return java.lang.Class
	 */
	public static Class ChartDataClass() {
		return JunChartData.class;
	}

	/**
	 * Convert the receiver as a JunOpenGL3dObject.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject as3dObject() {
		JunOpenGL3dCompoundObject chart3dObject = this.createChart3dObject();

		if (xAxis != null) {
			chart3dObject.add_(xAxis);
		}

		if (yAxis != null) {
			chart3dObject.add_(yAxis);
		}

		if (zAxis != null) {
			chart3dObject.add_(zAxis);
		}

		return chart3dObject;
	}

	/**
	 * Answer the color for axis.
	 * 
	 * @return java.awt.Color
	 */
	public Color axisColor() {
		if (axisColor == null) {
			axisColor = this.defaultAxisColor();
		}

		return axisColor;
	}

	/**
	 * Set the color for axis.
	 * 
	 * @param aColor java.awt.Color
	 */
	public void axisColor_(Color aColor) {
		axisColor = aColor;
	}

	/**
	 * Answer the array of color. These colors are used to paint each chart
	 * elements.
	 * 
	 * @return java.awt.Color[]
	 */
	public Color[] colors() {
		if (colors == null) {
			colors = this.defaultColors();
		}

		return colors;
	}

	/**
	 * Set the colors.
	 * 
	 * @param anArray java.awt.Color[]
	 */
	public void colors_(Color[] anArray) {
		colors = anArray;
	}

	/**
	 * Answer the chart data of the receiver.
	 * 
	 * @return sra.jun.opengl.chart.JunChartData
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 */
	public JunChartData data() {
		try {
			if (data == null) {
				Class aChartDataClass = (Class) _Perform(this.getClass(), "ChartDataClass");
				data = (JunChartData) _New(aChartDataClass);
			}

			return data;
		} catch (Exception e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Set the chart data of the receiver.
	 * 
	 * @param aChartData sra.jun.opengl.chart.JunChartData
	 */
	public void data_(JunChartData aChartData) {
		data = aChartData;
	}

	/**
	 * Answer the JunOpenGL3dObject to show.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject displayObject() {
		if (super.displayObject() == null) {
			super.displayObject_(this.as3dObject());
		}

		return super.displayObject();
	}

	/**
	 * Answer the collection of extra 3d objects.
	 * 
	 * @return java.util.Vector
	 */
	public Vector extra3dObjects() {
		return extra3dObjects;
	}

	/**
	 * Set the collection of extra 3d objects.
	 * 
	 * @param aCollectionOf3dObject java.util.Vector
	 */
	public void extra3dObjects_(Vector aCollectionOf3dObject) {
		extra3dObjects = aCollectionOf3dObject;
	}

	/**
	 * Answer true if the receiver has an x-axis, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean hasXAxis() {
		return (this.xAxis() != null);
	}

	/**
	 * Answer true if the receiver has a y-axis, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean hasYAxis() {
		return (this.yAxis() != null);
	}

	/**
	 * Answer true if the receiver has a z-axis, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean hasZAxis() {
		return (this.zAxis() != null);
	}

	/**
	 * Hide the x-axis.
	 */
	public void hideXAxis() {
		xAxis = null;
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Hide the y-axis.
	 */
	public void hideYAxis() {
		yAxis = null;
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Hide the z-axis.
	 */
	public void hideZAxis() {
		zAxis = null;
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Initialize the receiver.
	 */
	public void initialize() {
		super.initialize();
		this.initializeProjector();

		//this.initializeLights();
		//
		data = null;
		this.flushDisplayObject();
		this.flushAxes();
		axisColor = null;
		labelForegroundColor = null;
		labelBackgroundColor = null;
		colors = null;
		thickness = 0.1;
		intervalBetweenCharts = 0.3;
		extra3dObjects = null;
	}

	/**
	 * Initialize the lights to be off.
	 */
	public void initializeLights() {
		for (int i = 0; i < this.displayLights().length; i++) {
			this.displayLights()[i].beOff();
		}
	}

	/**
	 * Initialize the default projector.
	 */
	public void initializeProjector() {
		this.defaultEyePoint_(new Jun3dPoint(0.5, 0.5, 100.0));
		this.defaultSightPoint_(new Jun3dPoint(0.5, 0.5, 0.0));
		this.defaultUpVector_(new Jun3dPoint(0, 1, 0));
	}

	/**
	 * Answer the interval between charts of the receiver.
	 * 
	 * @return double
	 */
	public double intervalBetweenCharts() {
		return intervalBetweenCharts;
	}

	/**
	 * Set the inteval between charts of the receiver.
	 * 
	 * @param aNumber double
	 */
	public void intervalBetweenCharts_(double aNumber) {
		intervalBetweenCharts = aNumber;
	}

	/**
	 * Answer the background color for a label.
	 * 
	 * @return java.awt.Color
	 */
	public Color labelBackgroundColor() {
		if (labelBackgroundColor == null) {
			labelBackgroundColor = this.defaultLabelBackgroundColor();
		}

		return labelBackgroundColor;
	}

	/**
	 * Set the background color for a labe.
	 * 
	 * @param aColor java.awt.Color
	 */
	public void labelBackgroundColor_(Color aColor) {
		labelBackgroundColor = aColor;
	}

	/**
	 * Answer the foreground color for a label.
	 * 
	 * @return java.awt.Color
	 */
	public Color labelForegroundColor() {
		if (labelForegroundColor == null) {
			labelForegroundColor = this.defaultLabelForegroundColor();
		}

		return labelForegroundColor;
	}

	/**
	 * Set the foreground color for a labe.
	 * 
	 * @param aColor java.awt.Color
	 */
	public void labelForegroundColor_(Color aColor) {
		labelForegroundColor = aColor;
	}

	/**
	 * Answer another color.
	 * 
	 * @return java.awt.Color
	 */
	public Color nextColor() {
		Color[] colors = this.colors();

		if ((colors == null) || (colors.length == 0)) {
			return null;
		}

		if (colorIndex >= colors.length) {
			this.resetColorIndex();
		}

		return this.colors()[colorIndex++];
	}

	/**
	 * Answer the JunOpenGL3dObject at the specified point on the view.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param anEvent java.awt.event.MouseEvent
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject pick_with_(Jun2dPoint aPoint, MouseEvent anEvent) {
		JunOpenGL3dObject pickedObject = super.pick_with_(aPoint, anEvent);

		if (pickedObject != null) {
			this.pickedObject_(pickedObject);
		}

		return pickedObject;
	}

	/**
	 * Print my string representation on the Writer.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 */
	public void printOn_(Writer aWriter) throws IOException {
		super.printOn_(aWriter);

		//
		BufferedWriter bw = (aWriter instanceof BufferedWriter) ? (BufferedWriter) aWriter : new BufferedWriter(aWriter);
		bw.write('(');
		bw.newLine();
		bw.write("     data :  ");

		if (data == null) {
			bw.write("null");
		} else {
			data.printOn_(bw);
		}

		bw.newLine();
		bw.write(')');
		bw.flush();
	}

	/**
	 * Register the JunOpenGL3dObject as a visual object of the receiver.
	 * 
	 * @param a3dObject sra.jun.opengl.objects.JunOpenGL3dObject
	 * @param aValue java.lang.Object
	 */
	public void registerVisualObject_withValue_(JunOpenGL3dObject a3dObject, Object aValue) {
		a3dObject.name_(JunOpenGL3dObject.UniqueId());
		this.visualObjectDictionary().put(a3dObject.name(), aValue);
	}

	/**
	 * Answer the current samples of the chart.
	 * 
	 * @return java.lang.Object[][][]
	 */
	public Object[][][] samples() {
		return this.data().samples();
	}

	/**
	 * Set the samples of the chart.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 */
	public void samples_(Vector aCollectionOfSample) {
		this.data().samples_(aCollectionOfSample);
		this.flushDisplayObject();
	}

	/**
	 * Show the x-axis.
	 */
	public void showXAxis() {
		xAxis = this.createXAxis();
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Show the x-axis with labels.
	 */
	public void showXAxisWithLabels() {
		xAxis = JunOpenGL3dCompoundObject.With_with_(this.createXAxis(), this.createXAxisLabels());
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Show the y-axis.
	 */
	public void showYAxis() {
		yAxis = this.createYAxis();
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Show the y-axis with labels.
	 */
	public void showYAxisWithLabels() {
		yAxis = JunOpenGL3dCompoundObject.With_with_(this.createYAxis(), this.createYAxisLabels());
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Show the z-axis.
	 */
	public void showZAxis() {
		zAxis = this.createZAxis();
		this.updateAxesMenuIndication();
		this.flushDisplayObject();
		this.changed_($("presentation"));
	}

	/**
	 * Action for a menu : Spawn a band chart.
	 */
	public void spawnBandChart() {
		JunChartBand aChart = new JunChartBand();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Action for a menu : Spawn a bar chart.
	 */
	public void spawnBarChart() {
		JunChartBar aChart = new JunChartBar();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Action for a menu : Spawn a line chart.
	 */
	public void spawnLineChart() {
		JunChartLine aChart = new JunChartLine();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Action for a menu : Spawn a pie chart.
	 */
	public void spawnPieChart() {
		JunChartPie aChart = new JunChartPie();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Action for a menu : Spawn a plot chart.
	 */
	public void spawnPlotChart() {
		JunChartPlot aChart = new JunChartPlot();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Action for a menu : Spawn a radar chart.
	 */
	public void spawnRadarChart() {
		JunChartRadar aChart = new JunChartRadar();
		this._copyAttributes(aChart);
		aChart.open();
	}

	/**
	 * Answer the thickness of the chart element of the receiver.
	 * 
	 * @return double
	 */
	public double thickness() {
		return thickness;
	}

	/**
	 * Set the thickness of the chart element of the receiver.
	 * 
	 * @param aNumber double
	 */
	public void thickness_(double aNumber) {
		thickness = aNumber;
	}

	/**
	 * Toggle show or hide x-axis.
	 */
	public void toggleXAxis() {
		if (this.hasXAxis()) {
			this.hideXAxis();
		} else {
			this.showXAxisWithLabels();
		}
	}

	/**
	 * Toggle show or hide y-axis.
	 */
	public void toggleYAxis() {
		if (this.hasYAxis()) {
			this.hideYAxis();
		} else {
			this.showYAxisWithLabels();
		}
	}

	/**
	 * Toggle show or hide z-axis.
	 */
	public void toggleZAxis() {
		if (this.hasZAxis()) {
			this.hideZAxis();
		} else {
			this.showZAxis();
		}
	}

	/**
	 * Update the check marks of the axes menu.
	 */
	public void updateAxesMenuIndication() {
		StMenu axesMenu = this._axesMenu();
		if (axesMenu == null) {
			return;
		}

		((StCheckBoxMenuItem) axesMenu.getItem("Show X-Axis")).setSelected(hasXAxis());
		((StCheckBoxMenuItem) axesMenu.getItem("Show Y-Axis")).setSelected(hasYAxis());
		((StCheckBoxMenuItem) axesMenu.getItem("Show Z-Axis")).setSelected(hasZAxis());
	}

	/**
	 * Update the menu indications.
	 */
	public void updateMenuIndication() {
		super.updateMenuIndication();

		this.updateAxesMenuIndication();
	}

	/**
	 * Update the misc menu indication.
	 */
	public void updateMiscMenuIndication() {
		super.updateMiscMenuIndication();

		StMenu miscMenu = this._miscMenu();
		if (miscMenu == null) {
			return;
		}

		StMenuItem menuItem;
		boolean displayObjectIsNotEmpty = !this.isEmpty();

		// Spawn
		menuItem = miscMenu.getItem("Spawn");
		if (menuItem != null) {
			menuItem.setEnabled(displayObjectIsNotEmpty);
		}
	}

	/**
	 * Answer the Object which corresponds to the JunOpenGL3dObject.
	 * 
	 * @param a3dObject sra.jun.opengl.objects.JunOpenGL3dObject
	 * 
	 * @return java.lang.Object
	 */
	public Object valueForVisualObject_(JunOpenGL3dObject a3dObject) {
		if (a3dObject == null) {
			return null;
		}

		return this.visualObjectDictionary.get(a3dObject);
	}

	/**
	 * Set the information about the value range.
	 * 
	 * @param minimumNumber double
	 * @param maximumNumber double
	 */
	public void valueRangeFrom_to_(double minimumNumber, double maximumNumber) {
		this.data().valueRangeFrom_to_(minimumNumber, maximumNumber);
	}

	/**
	 * Set the information about the value range.
	 * 
	 * @param minimumNumber double
	 * @param maximumNumber double
	 * @param interval double
	 */
	public void valueRangeFrom_to_(double minimumNumber, double maximumNumber, double interval) {
		this.data().valueRangeFrom_to_by_(minimumNumber, maximumNumber, interval);
	}

	/**
	 * Answer the hastable for all visual objects.
	 * 
	 * @return java.util.Hashtable
	 */
	public Hashtable visualObjectDictionary() {
		if (visualObjectDictionary == null) {
			visualObjectDictionary = new Hashtable();
		}

		return visualObjectDictionary;
	}

	/**
	 * Answer the title string.
	 * 
	 * @return java.lang.String
	 */
	public String windowTitle() {
		return "Chart";
	}

	/**
	 * Answer the x-axis of the receiver.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject xAxis() {
		return xAxis;
	}

	/**
	 * Answer the array of x-axis labels.
	 * 
	 * @return java.lang.String[]
	 */
	public String[] xAxisLabels() {
		return this.data().keyLabels();
	}

	/**
	 * Set the array of x-axis labels.
	 * 
	 * @param anArray java.lang.String[]
	 */
	public void xAxisLabels_(String[] anArray) {
		this.data().keyLabels_(anArray);
	}

	/**
	 * Answer the y-axis of the receiver.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject yAxis() {
		return yAxis;
	}

	/**
	 * Answer the z-axis of the receiver.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject zAxis() {
		return zAxis;
	}

	/**
	 * Answer my axes menu.
	 *
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _axesMenu() {
		if (_axesMenu == null) {
			_axesMenu = this._defaultAxesMenu();
		}
		return _axesMenu;
	}

	/**
	 * Initialize the receiver with the attributes of the chart.
	 * 
	 * @param aChart sra.jun.opengl.chart.JunChartAbstract
	 */
	protected void _copyAttributes(JunChartAbstract aChart) {
		aChart.initialize();

		//
		JunChartData aData = aChart.data();
		aData.samples_(this.data().samples());
		aData.numberOfKeys_(this.data().numberOfKeys());

		if (axisColor != null) {
			aChart.axisColor_(axisColor);
		}

		if (labelForegroundColor != null) {
			aChart.labelForegroundColor_(labelForegroundColor);
		}

		if (labelBackgroundColor != null) {
			aChart.labelBackgroundColor_(labelBackgroundColor);
		}

		aChart.intervalBetweenCharts_(intervalBetweenCharts);
		aChart.thickness_(thickness);

		if (colors != null) {
			aChart.colors_(colors);
		}

		if (this.hasXAxis()) {
			aChart.showXAxisWithLabels();
		}

		if (this.hasYAxis()) {
			aChart.showYAxisWithLabels();
		}

		if (this.hasZAxis()) {
			aChart.showZAxis();
		}
	}

	/**
	 * Answer the default axes menu for JunChartAbstract.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultAxesMenu() {
		StMenu defaultAxesMenu = new StMenu("Axes");

		// Show X-Axies
		defaultAxesMenu.add(createCheckBoxMenuItem("Show X-Axis", "toggleXAxis"));

		// Show Y-Axis
		defaultAxesMenu.add(createCheckBoxMenuItem("Show Y-Axis", "toggleYAxis"));

		// Show Z-Axis
		defaultAxesMenu.add(createCheckBoxMenuItem("Show Z-Axis", "toggleZAxis"));

		return defaultAxesMenu;
	}

	/**
	 * Answer the default misc menu for JunChartAbstract.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultMiscMenu() {
		StMenu defaultMiscMenu = new StMenu("Misc");

		// Texture
		StMenu textureMenu = new StMenu("Texture");
		textureMenu.add(createMenuItem("From user...", "textureFromUser"));
		textureMenu.add(createMenuItem("From file...", "textureFromFile"));
		textureMenu.add(createMenuItem("Take away", "textureTakeAway"));
		textureMenu.addSeparator();
		textureMenu.add(createCheckBoxMenuItem("Linear approx.", "textureLinear"));
		textureMenu.add(createCheckBoxMenuItem("Clamp", "textureClamp"));
		textureMenu.add(createCheckBoxMenuItem("Modulate", "textureModulate"));
		textureMenu.add(createCheckBoxMenuItem("Mipmap", "textureMipmap"));
		defaultMiscMenu.add(textureMenu);

		// Spawn
		StMenu spawnMenu = new StMenu("Spawn");
		spawnMenu.add(createMenuItem("Band Chart", "spawnBandChart"));
		spawnMenu.add(createMenuItem("Bar Chart", "spawnBarChart"));
		spawnMenu.add(createMenuItem("Pie Chart", "spawnPieChart"));
		spawnMenu.add(createMenuItem("Plot Chart", "spawnPlotChart"));
		spawnMenu.add(createMenuItem("Line Chart", "spawnLineChart"));
		spawnMenu.add(createMenuItem("Radar Chart", "spawnRadarChart"));
		defaultMiscMenu.add(spawnMenu);

		// Viewport
		defaultMiscMenu.add(createMenuItem("Viewport", "spawnViewport"));

		// Bounds
		defaultMiscMenu.add(createMenuItem("Bounds", "showBounds"));

		return defaultMiscMenu;
	}

	/**
	 * Create a JunOpenGL3dObject which represents the chart.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dCompoundObject
	 */
	protected abstract JunOpenGL3dCompoundObject createChart3dObject();

	/**
	 * Create a JunOpenGL3dObject of x-axis.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject createXAxis() {
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();
		compoundObject.name_("x axis");

		Jun3dPoint[] points1 = { new Jun3dPoint(0, 0, 0), new Jun3dPoint(1, 0, 0)};
		compoundObject.add_(JunOpenGL3dPolyline.Vertexes_paint_(points1, this.axisColor()));

		Jun3dPoint[] points2 = { new Jun3dPoint(1, 0, 0), new Jun3dPoint(0.95, -0.02, 0), new Jun3dPoint(0.95, 0, 0), new Jun3dPoint(1, 0, 0)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points2, this.axisColor()));

		Jun3dPoint[] points3 = { new Jun3dPoint(1, 0, 0), new Jun3dPoint(0.95, 0, -0.02), new Jun3dPoint(0.95, 0, 0), new Jun3dPoint(1, 0, 0)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points3, this.axisColor()));
		compoundObject.paint_(this.axisColor());

		return compoundObject;
	}

	/**
	 * Create a JunOpenGL3dObject of x-axis labels.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected abstract JunOpenGL3dObject createXAxisLabels();

	/**
	 * Create a JunOpenGL3dObject of y-axis.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject createYAxis() {
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();
		compoundObject.name_("y axis");

		Jun3dPoint[] points1 = { new Jun3dPoint(0, 0, 0), new Jun3dPoint(0, 1, 0)};
		compoundObject.add_(JunOpenGL3dPolyline.Vertexes_paint_(points1, this.axisColor()));

		Jun3dPoint[] points2 = { new Jun3dPoint(0, 1, 0), new Jun3dPoint(-0.02, 0.95, 0), new Jun3dPoint(0, 0.95, 0), new Jun3dPoint(0, 1, 0)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points2, this.axisColor()));

		Jun3dPoint[] points3 = { new Jun3dPoint(0, 1, 0), new Jun3dPoint(0, 0.95, -0.02), new Jun3dPoint(0, 0.95, 0), new Jun3dPoint(0, 1, 0)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points3, this.axisColor()));
		compoundObject.paint_(this.axisColor());

		return compoundObject;
	}

	/**
	 * Create a JunOpenGL3dObject of y-axis labels.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject createYAxisLabels() {
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();
		JunChartData data = this.data();
		double minimumNumber = data.minimumOfValueRange();
		double maximumNumber = data.maximumOfValueRange();
		double interval = data.intervalOfValueRange();

		for (double aNumber = (Math.floor(minimumNumber / interval) + 1) * interval; aNumber <= (Math.floor(maximumNumber / interval) * interval); aNumber += interval) {
			if ((aNumber == minimumNumber) || (aNumber == maximumNumber)) {
				continue;
			}

			double y = data.normalizeValue_(aNumber);
			Jun3dPoint[] points = { new Jun3dPoint(0, y, 0), new Jun3dPoint(-0.05, y, 0)};
			compoundObject.add_(JunOpenGL3dPolyline.Vertexes_(points));
			compoundObject.add_(this.label_rightCenterAt_(String.valueOf(aNumber), new Jun3dPoint(-0.05, y, 0)));
		}

		return compoundObject;
	}

	/**
	 * Create a JunOpenGL3dObject of z-axis.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject createZAxis() {
		JunOpenGL3dCompoundObject compoundObject = new JunOpenGL3dCompoundObject();
		compoundObject.name_("z axis");

		Jun3dPoint[] points1 = { new Jun3dPoint(0, 0, 0), new Jun3dPoint(0, 0, 1)};
		compoundObject.add_(JunOpenGL3dPolyline.Vertexes_paint_(points1, this.axisColor()));

		Jun3dPoint[] points2 = { new Jun3dPoint(0, 0, 1), new Jun3dPoint(-0.02, 0, 0.95), new Jun3dPoint(0, 0, 0.95), new Jun3dPoint(0, 0, 1)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points2, this.axisColor()));

		Jun3dPoint[] points3 = { new Jun3dPoint(0, 0, 1), new Jun3dPoint(0, -0.02, 0.95), new Jun3dPoint(0, 0, 0.95), new Jun3dPoint(0, 0, 1)};
		compoundObject.add_(JunOpenGL3dPolygon.Vertexes_paint_(points3, this.axisColor()));
		compoundObject.paint_(this.axisColor());

		return compoundObject;
	}

	/**
	 * Answer the block closure to evaluate which an object is picked.
	 * 
	 * @return StBlockClosure
	 */
	protected StBlockClosure defaultActionForPickedObject() {
		return new StBlockClosure();
	}

	/**
	 * Answer the default color for axis.
	 * 
	 * @return java.awt.Color
	 */
	protected Color defaultAxisColor() {
		return Color.blue;
	}

	/**
	 * Answer the default colors.
	 * 
	 * @return java.awt.Color[]
	 */
	protected Color[] defaultColors() {
		if (this.samples() == null) {
			Color[] defaultColors = { Color.black };

			return defaultColors;
		}

		return this.defaultColors_(this.data().sheet().valueSize());
	}

	/**
	 * Answer the default colors.  The number of the colors are specified.
	 * 
	 * @param size int
	 * 
	 * @return java.awt.Color[]
	 */
	protected Color[] defaultColors_(int size) {
		Color[] defaultColors = new Color[size];

		for (int i = 0; i < size; i++) {
			defaultColors[i] = Color.getHSBColor((float) ((double) i / size), (float) 1, (float) 1);
		}

		return defaultColors;
	}

	/**
	 * Answer the default background color for a label.
	 * 
	 * @return java.awt.Color
	 */
	protected Color defaultLabelBackgroundColor() {
		return Color.white;
	}

	/**
	 * Answer the default foreground color for a label.
	 * 
	 * @return java.awt.Color
	 */
	protected Color defaultLabelForegroundColor() {
		return Color.black;
	}

	/**
	 * Answer the default height for a label.
	 * 
	 * @return double
	 */
	protected double defaultLabelHeight() {
		return 0.05;
	}

	/**
	 * Create the menubar for JunOpenGLDisplay.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();

		rootMenu.add(this._fileMenu());
		rootMenu.add(this._editMenu());
		rootMenu.add(this._viewMenu());
		rootMenu.add(this._lightMenu());
		rootMenu.add(this._axesMenu());
		rootMenu.add(this._miscMenu());

		return rootMenu;
	}

	/**
	 * Flush the cached JunOpenGL3dObjects for axes.
	 */
	protected void flushAxes() {
		xAxis = null;
		yAxis = null;
		zAxis = null;
	}

	/**
	 * Flush the cache for display objects.
	 */
	protected void flushDisplayObject() {
		super.flushObject();
		visualObjectDictionary = null;
		this.resetColorIndex();
	}

	/**
	 * Create a label with the String at the right-center of the Point.
	 * 
	 * @param aString java.lang.String
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject label_rightCenterAt_(String aString, Jun3dPoint aPoint) {
		JunOpenGLTexture texture = JunOpenGLTexture.TextureForString_foreColor_backColor_(aString, this.labelForegroundColor(), this.labelBackgroundColor());
		double height = this.defaultLabelHeight();
		double width = (height * texture.width()) / texture.height();
		double originX = aPoint.x() - width;
		double originY = aPoint.y() - (height / 2);
		double cornerX = aPoint.x();
		double cornerY = aPoint.y() + (height / 2);
		double z = aPoint.z();
		Jun3dPoint[] points = { new Jun3dPoint(originX, originY, z), new Jun3dPoint(cornerX, originY, z), new Jun3dPoint(cornerX, cornerY, z), new Jun3dPoint(originX, cornerY, z)};
		JunOpenGL3dObject label = JunOpenGL3dPolygon.Vertexes_(points);
		label.texture_(texture);

		return label;
	}

	/**
	 * Create a label with the String at the top-center of the Point.
	 * 
	 * @param aString java.lang.String
	 * @param aPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject label_topCenterAt_(String aString, Jun3dPoint aPoint) {
		JunOpenGLTexture texture = JunOpenGLTexture.TextureForString_foreColor_backColor_(aString, this.labelForegroundColor(), this.labelBackgroundColor());
		double height = this.defaultLabelHeight();
		double width = (height * texture.width()) / texture.height();
		double originX = aPoint.x() - (width / 2);
		double originY = aPoint.y() - height;
		double cornerX = aPoint.x() + (width / 2);
		double cornerY = aPoint.y();
		double z = aPoint.z();
		Jun3dPoint[] points = { new Jun3dPoint(originX, originY, z), new Jun3dPoint(cornerX, originY, z), new Jun3dPoint(cornerX, cornerY, z), new Jun3dPoint(originX, cornerY, z)};
		JunOpenGL3dObject label = JunOpenGL3dPolygon.Vertexes_(points);
		label.texture_(texture);

		return label;
	}

	/**
	 * The JunOpenGL3dObject is picked. Do something according to the
	 * definition.
	 * 
	 * @param anObject sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected void pickedObject_(JunOpenGL3dObject anObject) {
		this.defaultActionForPickedObject().value_(anObject);
	}

	/**
	 * Reset the index of the colors.
	 */
	private void resetColorIndex() {
		colorIndex = 0;
	}
}
