package sra.jun.opengl.chart;

import java.io.*;
import java.util.Vector;
import sra.smalltalk.*;

/**
 * JunChartDataWithNumericalKeys class
 * 
 *  @author    nisinaka
 *  @created   1999/01/13 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunChartDataWithNumericalKeys.java,v 6.0 2003/10/28 10:52:17 nisinaka Exp $
 */
public class JunChartDataWithNumericalKeys extends JunChartData {
	/** The key range. */
	protected double[] keyRange = null;

	/** The minimum key. */
	protected double minimumKey = Double.NaN;

	/** The maximum key. */
	protected double maximumKey = Double.NaN;

	/**
	 * The default constructor. Create a new instance and initialize it.
	 */
	public JunChartDataWithNumericalKeys() {
		super();
	}

	/**
	 * Create a new instance of a ChartData and initialize it with the
	 * collection of sample.. This method corresponds to the class method
	 * "samples:" of JunChartData in Smalltalk.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 */
	public JunChartDataWithNumericalKeys(java.util.Vector aCollectionOfSample) {
		super(aCollectionOfSample);
	}

	/**
	 * Create a new JunChartData with a collection of sample and a number of
	 * keys. This method corresponds to the class method
	 * "samples:numberOfKeys:" of JunChartData in Smalltalk.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 * @param numberOfKeys int
	 */
	public JunChartDataWithNumericalKeys(java.util.Vector aCollectionOfSample, int numberOfKeys) {
		super(aCollectionOfSample, numberOfKeys);
	}

	/**
	 * Denormalize the number in the key range.
	 * 
	 * @param aNumber double
	 * 
	 * @return double
	 */
	public double denormalizeKey_(double aNumber) {
		double[] range = { 0, 1 };

		return this.mapNumber_fromRange_toRange_(aNumber, range, this.keyRange());
	}

	/**
	 * Answer true if the receiver has a valid keys, otherwise false.
	 * 
	 * @return boolean
	 */
	public boolean hasValidKeys() {
		if (this.numberOfKeys() != 1) {
			return false;
		}

		JunChartDataSheet[] sheets = this.sheets();

		for (int i = 0; i < sheets.length; i++) {
			Object result = sheets[i].keysDo_(new StBlockClosure() {
				public Object value_(Object key) {
					if (key instanceof Number == false) {
						return Boolean.FALSE;
					}

					return null;
				}
			});

			if (result == Boolean.FALSE) {
				return false;
			}
		}

		return true;
	}

	/**
	 * Answer the interval of the key range.
	 * 
	 * @return double
	 */
	public double intervalOfKeyRange() {
		return this.keyRange()[2];
	}

	/**
	 * Answer the information about the key range.
	 * 
	 * @return double[]
	 */
	public double[] keyRange() {
		if (keyRange == null) {
			keyRange = this.defaultKeyRange();
		}

		return keyRange;
	}

	/**
	 * Set the information about the key range.
	 * 
	 * @param anArray double[]
	 */
	public void keyRange_(double[] anArray) {
		keyRange = anArray;
	}

	/**
	 * Set the information about the key range.
	 * 
	 * @param minimumNumber double
	 * @param maximumNumber double
	 */
	public void keyRangeFrom_to_(double minimumNumber, double maximumNumber) {
		this.keyRangeFrom_to_by_(minimumNumber, maximumNumber, this.defaultIntervalFrom_to_(minimumNumber, maximumNumber));
	}

	/**
	 * Set the information about the key range.
	 * 
	 * @param minimumNumber double
	 * @param maximumNumber double
	 * @param interval double
	 */
	public void keyRangeFrom_to_by_(double minimumNumber, double maximumNumber, double interval) {
		if (minimumNumber >= maximumNumber) {
			return;
		}

		double[] anArray = { minimumNumber, maximumNumber, interval };
		this.keyRange_(anArray);
	}

	/**
	 * Answer the maximum key of the chart data.
	 * 
	 * @return double
	 */
	public double maximumKey() {
		if (Double.isNaN(maximumKey)) {
			final StValueHolder maximumValueHolder = new StValueHolder(Double.NEGATIVE_INFINITY);
			JunChartDataSheet[] sheets = this.sheets();

			for (int i = 0; i < sheets.length; i++) {
				sheets[i].keysDo_(new StBlockClosure() {
					public Object value_(Object anObject) {
						double aNumber = ((Number) anObject).doubleValue();

						if (aNumber > maximumValueHolder._doubleValue()) {
							maximumValueHolder.value_(aNumber);
						}

						return null;
					}
				});
			}

			maximumKey = maximumValueHolder._doubleValue();
		}

		return maximumKey;
	}

	/**
	 * Answer the maximum of the key range.
	 * 
	 * @return double
	 */
	public double maximumOfKeyRange() {
		return this.keyRange()[1];
	}

	/**
	 * Answer the minimum key of the chart data.
	 * 
	 * @return double
	 */
	public double minimumKey() {
		if (Double.isNaN(minimumKey)) {
			final StValueHolder minimumValueHolder = new StValueHolder(Double.MAX_VALUE);
			JunChartDataSheet[] sheets = this.sheets();

			for (int i = 0; i < sheets.length; i++) {
				sheets[i].keysDo_(new StBlockClosure() {
					public Object value_(Object anObject) {
						double aNumber = ((Number) anObject).doubleValue();

						if (aNumber < minimumValueHolder._doubleValue()) {
							minimumValueHolder.value_(anObject);
						}

						return null;
					}
				});
			}

			minimumKey = minimumValueHolder._doubleValue();
		}

		return minimumKey;
	}

	/**
	 * Answer the minimum of the key range.
	 * 
	 * @return double
	 */
	public double minimumOfKeyRange() {
		return this.keyRange()[0];
	}

	/**
	 * Normalize the number in the key range.
	 * 
	 * @param aNumber double
	 * 
	 * @return double
	 */
	public double normalizeKey_(double aNumber) {
		double[] range = { 0, 1 };

		return this.mapNumber_fromRange_toRange_(aNumber, this.keyRange(), range);
	}

	/**
	 * Print my string representation on the Writer.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException DOCUMENT ME!
	 */
	public void printOn_(Writer aWriter) throws IOException {
		super.printOn_(aWriter);

		//
		BufferedWriter bw = (aWriter instanceof BufferedWriter) ? (BufferedWriter) aWriter : new BufferedWriter(aWriter);
		bw.write('(');
		bw.newLine();
		bw.write("     keyRange :  ");

		if (keyRange == null) {
			bw.write("null");
		} else {
			bw.write(" from ");
			bw.write(String.valueOf(keyRange[0]));
			bw.write(" to ");
			bw.write(String.valueOf(keyRange[1]));
			bw.write(" by ");
			bw.write(String.valueOf(keyRange[2]));
		}

		bw.newLine();
		bw.write("     minimumKey :  ");
		bw.write(String.valueOf(minimumKey));
		bw.newLine();
		bw.write("     maximumKey :  ");
		bw.write(String.valueOf(maximumKey));
		bw.newLine();
		bw.write(')');
		bw.flush();
	}

	/**
	 * Set the samples of the receiver.
	 * 
	 * @param aCollectionOfSample java.util.Vector
	 */
	public void samples_(Vector aCollectionOfSample) {
		Object[][][] savedSamples = this.samples();
		super.samples_(aCollectionOfSample);

		if (this.hasValidKeys() == false) {
			StDialog.Warn_("Invalid keys:  Could not change samples.");
			super.samples_(savedSamples);
		}
	}

	/**
	 * Answer the array of JunChartDataSheet.
	 * 
	 * @return sra.jun.opengl.chart.JunChartDataSheet[]
	 */
	public JunChartDataSheet[] sheets() {
		if (sheets == null) {
			if (samples == null) {
				return null;
			}

			if (!this.hasValidSamples()) {
				return null;
			}

			//
			StBlockClosure sortBlock = new StBlockClosure() {
				public Object value_value_(Object array1, Object array2) {
					return new Boolean(((Number) ((Object[]) array1)[0]).doubleValue() < ((Number) ((Object[]) array2)[0]).doubleValue());
				}
			};

			int size = samples.length;
			sheets = new JunChartDataSheet[size];

			for (int i = 0; i < size; i++) {
				StSortedCollection sortedSample = new StSortedCollection(sortBlock);

				for (int j = 0; j < samples[i].length; j++) {
					sortedSample.add_(samples[i][j]);
				}

				Object[] array = sortedSample._asArray();
				int length = array.length;
				Object[][] collectionOfArray = new Object[length][];

				for (int j = 0; j < length; j++) {
					collectionOfArray[j] = (Object[]) array[j];
				}

				sheets[i] = new JunChartDataSheet(collectionOfArray, this.numberOfKeys());
			}
		}

		return sheets;
	}

	/**
	 * Answer the default key range.     0 : minimum value     1 : maximum
	 * value     2 : interval
	 * 
	 * @return double[]
	 */
	protected double[] defaultKeyRange() {
		double max = this.maximumKey();
		double min = this.minimumKey();
		double[] range = { min, max, this.defaultIntervalFrom_to_(min, max)};

		return range;
	}

	/**
	 * Answer the default number of keys.
	 * 
	 * @return int
	 */
	protected int defaultNumberOfKeys() {
		return 1;
	}

	/**
	 * Flush the cached values.
	 */
	protected void flushCaches() {
		super.flushCaches();
		minimumKey = Double.NaN;
		maximumKey = Double.NaN;
	}
}
