package sra.jun.opengl.display;

import java.awt.*;
import java.awt.event.*;
import sra.jun.geometry.basic.*;
import sra.jun.goodies.cursors.JunCursors;
import sra.smalltalk.*;

/**
 * JunOpenGLDisplayController class
 * 
 *  @author    He Weijie
 *  @created   1998/10/08 (by r-matuda)
 *  @updated   1998/10/30 (by He Weijie)
 *  @updated   1999/08/11 (by nisiaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLDisplayController.java,v 6.0 2003/10/28 10:52:17 nisinaka Exp $
 */
public class JunOpenGLDisplayController extends StController implements MouseListener, MouseMotionListener {
	protected static final long _InertiaTime = 100;

	/** The vector for a rotation. */
	public transient Jun2dPoint movementVector;
	protected transient Point _startPoint = null;
	protected transient Point _previousPoint = null;
	protected transient Thread _rotationThread = null;
	protected transient long _pressTime;
	protected transient Rectangle _boundingBox = null;
	protected Cursor _savedCursor = null;

	/**
	 * Flush the cached value of the movement vector.
	 */
	public void flushMovementVector() {
		movementVector = null;
		Thread.yield();
	}

	/**
	 * The General dragging action.
	 * 
	 * @param event java.awt.event.MouseEvent
	 */
	public void mouseDragged(MouseEvent event) {
		if (_startPoint != null) {
			Point currentPoint = event.getPoint();
			if (currentPoint.equals(_previousPoint) == false) {
				this.mouseDraggedFrom_to_with_(_previousPoint, currentPoint, event);
				_previousPoint = currentPoint;
				_pressTime = System.currentTimeMillis();
			}
		}
		Thread.yield();
	}

	/**
	 * Invoked when a mouse is entered on a component.
	 * 
	 * @param e java.awt.event.MouseEvent
	 */
	public void mouseEntered(MouseEvent e) {
		StSymbol state = this.getModel().buttonState();
		if (state == null) {
			return;
		}

		Component aComponent = this.getOpenGLView().toComponent();
		_savedCursor = aComponent.getCursor();

		if (state == $("pick")) {
			aComponent.setCursor(JunCursors.CrossCursor());
		} else if (state == $("grab")) {
			aComponent.setCursor(JunCursors.HandCursor());
		} else if (state == $("drag")) {
			aComponent.setCursor(JunCursors.QuartersCursor());
		} else if (state == $("focus")) {
			aComponent.setCursor(JunCursors.GlassCursor());
		}
	}

	/**
	 * Invoked when a mouse is exited on a component.
	 * 
	 * @param e java.awt.event.MouseEvent
	 */
	public void mouseExited(MouseEvent e) {
		this.getOpenGLView().toComponent().setCursor(_savedCursor);
	}

	/**
	 * The mouse button has been pressed.
	 * 
	 * @param event java.awt.event.MouseEvent
	 */
	public void mousePressed(MouseEvent event) {
		this.flushMovementVector();
		if (event.isMetaDown()) {
			// this.getOpenGLView()._showPopupMenu(event.getX(), event.getY());
		} else if (event.isAltDown()) {
		} else {
			_startPoint = event.getPoint();
			_previousPoint = _startPoint;
			_pressTime = System.currentTimeMillis();

			// Handle picking here.
			// Others, for dragging, are handled at mouseDragged().
			if (this.getModel().buttonState() == $("pick")) {
				this.pickActivityOn_with_(_startPoint, event);
			}
		}
	}

	/**
	 * The mouse button has been released.  The shape of the cursor reset to
	 * default.
	 * 
	 * @param event java.awt.event.MouseEvent
	 */
	public void mouseReleased(MouseEvent event) {
		if (_startPoint != null) {
			StSymbol state = this.getModel().buttonState();
			if (state == $("grab")) {
				if (System.currentTimeMillis() - _pressTime <= _InertiaTime) {
					if (this.movementVector() != null) {
						this.createMovementProcess();
					}
				}
			} else if (state == $("focus")) {
				Jun2dBoundingBox area1 = Jun2dBoundingBox.Origin_corner_(new Jun2dPoint(-1, -1), new Jun2dPoint(1, 1));
				Jun2dBoundingBox area2 = this.mouse2dRectangle();
				this._hideBoundingBox();

				if (event.isShiftDown()) {
					this.getModel().focus_to_(area2, area1);
				} else {
					this.getModel().focus_to_(area1, area2);
				}
			}

			_startPoint = null;
		}
	}

	/**
	 * Answer the current movement vector.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint movementVector() {
		return movementVector;
	}

	/**
	 * Set the current movement vector.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 */
	public void movementVector_(Jun2dPoint aPoint) {
		double x = Math.max(-1, Math.min(aPoint.x(), 1));
		double y = Math.max(-1, Math.min(aPoint.y(), 1));
		movementVector = new Jun2dPoint(x, y);
	}

	/**
	 * Answer the current bounding box for focusing.
	 * 
	 * @return java.awt.Rectangle
	 */
	protected Rectangle _getBoundingBox() {
		return _boundingBox;
	}

	/**
	 * Hide the bounding box if exist.
	 */
	protected void _hideBoundingBox() {
		if (_boundingBox != null) {
			Graphics aGraphics = ((Component) getOpenGLView()).getGraphics();

			try {
				aGraphics.setColor(Color.white);
				aGraphics.setXORMode(Color.black);
				aGraphics.drawRect(_boundingBox.x, _boundingBox.y, _boundingBox.width, _boundingBox.height);
			} finally {
				aGraphics.dispose();
			}

			_boundingBox = null;
		}
	}

	/**
	 * Show bounding box.
	 * 
	 * @param aRectangle java.awt.Rectangle
	 */
	protected void _showBoundingBox(Rectangle aRectangle) {
		if (_boundingBox != null) {
			this._hideBoundingBox();
		}

		_boundingBox = aRectangle;

		Graphics aGraphics = ((Component) getOpenGLView()).getGraphics();

		try {
			aGraphics.setColor(Color.white);
			aGraphics.setXORMode(Color.black);
			aGraphics.drawRect(_boundingBox.x, _boundingBox.y, _boundingBox.width, _boundingBox.height);
		} finally {
			aGraphics.dispose();
		}
	}

	/**
	 * add listeners.
	 * 
	 * @param newView sra.mvctest.StView
	 */
	protected void buildListener(StView newView) {
		Component glView = (Component) this.getOpenGLView();
		glView.addMouseListener(this);
		glView.addMouseMotionListener(this);
	}

	/**
	 * Initialize the rotation thread.
	 */
	protected void createMovementProcess() {
		_rotationThread = new Thread() {
			public void run() {
				Jun2dPoint aVector;
				while (view() != null && (aVector = movementVector()) != null) {
					getModel().grab_(aVector);
					Thread.yield();
				}
			}
		};

		_rotationThread.setPriority(Thread.NORM_PRIORITY);
		_rotationThread.start();
	}

	/**
	 * The drag activity.
	 * 
	 * @param fromPoint java.awt.Point
	 * @param toPoint java.awt.Point
	 * @param event DOCUMENT ME!
	 */
	protected void dragActivityFrom_to_with_(Point fromPoint, Point toPoint, MouseEvent event) {
		Jun2dPoint regularizeFromPoint = this.regularizePoint_(fromPoint);
		Jun2dPoint regularizeToPoint = this.regularizePoint_(toPoint);

		if (event.isShiftDown()) {
			this.getModel().look_xy_(regularizeFromPoint, regularizeToPoint);
		} else {
			this.getModel().slide_xy_(regularizeFromPoint, regularizeToPoint);
		}
	}

	/**
	 * Make sure to terminate the process when finalizing.
	 */
	protected void finalize() throws Throwable {
		super.finalize();
		view = null;
	}

	/**
	 * Activity for the focus.
	 * 
	 * @param fromPoint java.awt.Point
	 * @param toPoint java.awt.Point
	 */
	protected void focusActivityFrom_to_(Point fromPoint, Point toPoint) {
		int originX = Math.min(_startPoint.x, toPoint.x);
		int originY = Math.min(_startPoint.y, toPoint.y);
		int cornerX = Math.max(_startPoint.x, toPoint.x);
		int cornerY = Math.max(_startPoint.y, toPoint.y);
		Rectangle boundingBox = new Rectangle(originX, originY, cornerX - originX, cornerY - originY);
		this._showBoundingBox(boundingBox);
	}

	/**
	 * model accessing.
	 * 
	 * @return sra.jun.opengl.display.JunOpenGLDisplayModel
	 */
	protected JunOpenGLDisplayModel getModel() {
		return (JunOpenGLDisplayModel) this.model();
	}

	/**
	 * Answer my model as JunOpenGLDisplayModel.
	 * 
	 * @return sra.jun.opengl.display.JunOpenGLDisplayModel
	 */
	protected JunOpenGLDisplayModel getOpenGLDisplayModel() {
		return (JunOpenGLDisplayModel) this.model();
	}

	/**
	 * view accessing.
	 * 
	 * @return sra.jun.opengl.display.JunOpenGL3dView
	 */
	protected JunOpenGL3dView getOpenGLView() {
		StView view = this.view();

		if (view instanceof JunOpenGL3dView) {
			return (JunOpenGL3dView) view;
		} else if (view instanceof JunOpenGLDisplayView) {
			return ((JunOpenGLDisplayView) view).getOpenGLView();
		} else {
			return null;
		}
	}

	/**
	 * The grab activity.
	 * 
	 * @param fromPoint java.awt.Point
	 * @param toPoint java.awt.Point
	 */
	protected void grabActivityFrom_to_(Point fromPoint, Point toPoint) {
		this.flushMovementVector();
		Jun2dPoint regularizedFromPoint = this.regularizePoint_(fromPoint);
		Jun2dPoint regularizedToPoint = this.regularizePoint_(toPoint);
		this.getModel().grab_xy_(regularizedFromPoint, regularizedToPoint);
		this.movementVector_((Jun2dPoint) regularizedToPoint.minus_(regularizedFromPoint));
	}

	/**
	 * Answer the new focus area.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 */
	protected Jun2dBoundingBox mouse2dRectangle() {
		Rectangle boundingBox = this._getBoundingBox();

		if (boundingBox == null) {
			boundingBox = new Rectangle(_startPoint.x, _startPoint.y, 0, 0);
		}

		if (boundingBox.height <= 2) {
			int expandSize = Math.max(((Component) getOpenGLView()).getBounds().height / 4, 8);
			boundingBox.grow(expandSize, expandSize);
		}

		Jun2dPoint originPoint = this.regularizePoint_(new Point(boundingBox.x, boundingBox.y));
		Jun2dPoint cornerPoint = this.regularizePoint_(new Point(boundingBox.x + boundingBox.width, boundingBox.y + boundingBox.height));

		return Jun2dBoundingBox.Origin_corner_(originPoint, cornerPoint);
	}

	/**
	 * Action for the mouse drag.
	 * 
	 * @param fromPoint java.awt.Point
	 * @param toPoint java.awt.Point
	 * @param event java.awt.event.MouseEvent
	 */
	protected void mouseDraggedFrom_to_with_(Point fromPoint, Point toPoint, MouseEvent event) {
		StSymbol state = this.getModel().buttonState();
		if (state == $("grab")) {
			this.grabActivityFrom_to_(fromPoint, toPoint);
		} else if (state == $("drag")) {
			this.dragActivityFrom_to_with_(fromPoint, toPoint, event);
		} else if (state == $("focus")) {
			this.focusActivityFrom_to_(fromPoint, toPoint);
		}
	}

	/**
	 * define pick action.
	 * 
	 * @param aPoint
	 * @param event DOCUMENT ME!
	 */
	protected void pickActivityOn_with_(Point aPoint, MouseEvent event) {
		this.flushMovementVector();
		this.getModel().pick_with_(this.regularizePoint_(aPoint), event);
	}

	/**
	 * Regularize the point on the current view.
	 * 
	 * @param aPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	protected Jun2dPoint regularizePoint_(Point aPoint) {
		Dimension bounds = ((Component) getOpenGLView()).getSize();
		double x = (aPoint.x - bounds.width / 2.0) / (bounds.height / 2.0);
		double y = - (aPoint.y - bounds.height / 2.0) / (bounds.height / 2.0);

		return new Jun2dPoint(x, y);
	}
}
