package sra.jun.opengl.flux;

import java.awt.*;
import java.io.*;
import java.text.*;
import java.util.*;
import sra.smalltalk.*;
import sra.jun.geometry.basic.*;
import sra.jun.goodies.track.*;
import sra.jun.opengl.display.*;
import sra.jun.opengl.objects.*;
import sra.jun.opengl.projection.*;
import sra.jun.system.support.*;

/**
 * JunOpenGLFluxModel class
 * 
 *  @author    MATSUDA Ryouichi
 *  @created   1998/11/25 (by MATSUDA Ryouichi)
 *  @updated   1999/06/25 (by nisinaka)
 *  @updated   2003/03/24 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLFluxModel.java,v 6.0 2003/10/28 10:52:18 nisinaka Exp $
 */
public class JunOpenGLFluxModel extends JunOpenGLDisplayModel {

	protected JunOpenGLFluxObject fluxObject;
	protected JunTrackerModel trackerModel;

	/**
	 * Create a new instance of JunOpenGLFluxModel.
	 */
	public JunOpenGLFluxModel() {
		super();
		this.initialize();
	}

	/**
	 * Create a new instance of JunOpenGLFluxModel with flux object.
	 * 
	 * @param aFluxObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * 
	 * @return DOCUMENT ME!
	 */
	public static JunOpenGLFluxModel FluxObject_(JunOpenGLFluxObject aFluxObject) {
		JunOpenGLFluxModel fluxModel;

		fluxModel = new JunOpenGLFluxModel();
		fluxModel.fluxObject_(aFluxObject);

		return fluxModel;
	}

	/**
	 * Answer the bounding box of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dBoundingBox
	 */
	public Jun3dBoundingBox boundingBox() {
		if (this.fluxObject() == null) {
			return Jun3dBoundingBox.Origin_corner_(new Jun3dPoint(0, 0, 0), new Jun3dPoint(0, 0, 0));
		}

		return this.fluxObject().boundingBox();
	}

	/**
	 * Answer compound object.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dCompoundObject
	 */
	public JunOpenGL3dCompoundObject compoundObject() {
		return this.fluxObject().compoundObject();
	}

	/**
	 * Answer the stamp string for the LFT 1.0 file format.
	 * 
	 * @return java.lang.String
	 */
	public String defaultStampForLFT10() {
		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		pw.println("%LFT V1.0 List Flux Transmission (Lisp S Expression)");
		pw.println("% This file was created by " + JunSystem.System() + JunSystem.Version());
		pw.println("% " + DateFormat.getInstance().format(new Date()));
		pw.println();
		pw.flush();

		return sw.toString();
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunOpenGLFluxViewAwt(this);
		} else {
			return new JunOpenGLFluxViewSwing(this);
		}
	}

	/**
	 * End of TrackerModel.
	 */
	public void end() {
		this.trackerModel().end();
	}

	/**
	 * Answer fluxObject.
	 * 
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 */
	public JunOpenGLFluxObject fluxObject() {
		return fluxObject;
	}

	/**
	 * Set flux object.
	 * 
	 * @param aFluxObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 */
	public void fluxObject_(JunOpenGLFluxObject aFluxObject) {
		fluxObject = aFluxObject;

		if ((fluxObject.size() - 1) <= 0) {
			this.trackerModel().step_(0.0d);
		} else {
			this.trackerModel().step_(1.0d / (fluxObject.size() - 1));
		}

		this.trackerModel().first();
	}

	/**
	 * Initialize.
	 */
	public void initialize() {
		super.initialize();
		fluxObject = null;
		trackerModel = null;
	}

	/**
	 * Load a JunOpenGLFluxObject from the reader.
	 * 
	 * @param aReader java.io.BufferedReader
	 * 
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 * 
	 * @exception IOException
	 */
	public JunOpenGLFluxObject loadFromLFT10_(BufferedReader aReader) throws IOException {
		StringWriter aWriter = new StringWriter();
		JunOpenGLFluxObject anObject = null;

		try {
			int ch;

			while ((ch = aReader.read()) > 0) {
				aWriter.write(ch);
			}

			aWriter.flush();
			anObject = JunOpenGLFluxObject.LoadFrom_(aWriter.toString());
		} finally {
			aWriter.close();
		}

		return anObject;
	}

	/**
	 * Called by menu of 'Open...'
	 */
	public void openLFT() {
		this.openLFT10();
	}

	/**
	 * Open a LFT1.0 file and get a JunOpenGLFluxObject.
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 */
	public void openLFT10() {
		File aFile = StDialog.OpenFileDialog();

		if (aFile == null) {
			return;
		}

		//
		JunOpenGLFluxObject anObject;

		try {
			anObject = this.readFromLFT10_(aFile);
		} catch (IOException e) {
			throw new SmalltalkException(e);
		}

		if (anObject == null) {
			return;
		}

		this.fluxObject_(anObject);
		this.resetView();

		if (this.showModel() != null) {
			this.showModel().resetView();
		}
	}

	/**
	 * Enumerate the points and evaluate the Block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public void pointsDo_(final StBlockClosure aBlock) {
		if (this.fluxObject() == null) {
			this.pointsDo_(aBlock);
		} else {
			this.fluxObject().pointsDo_(aBlock);
		}
	}

	/**
	 * Read a LFT1.0 file and create a flux object.
	 * 
	 * @param aFile java.io.File
	 * 
	 * @return sra.jun.opengl.flux.JunOpenGLFluxObject
	 * 
	 * @exception IOException
	 */
	public JunOpenGLFluxObject readFromLFT10_(File aFile) throws IOException {
		FileReader aFileReader;

		try {
			aFileReader = new FileReader(aFile);
		} catch (FileNotFoundException e) {
			return null;
		}

		BufferedReader aReader = new BufferedReader(aFileReader);
		JunOpenGLFluxObject anObject = null;

		try {
			anObject = this.loadFromLFT10_(aReader);
		} finally {
			aReader.close();
		}

		return anObject;
	}

	/**
	 * Invoked when the application will be closed.
	 */
	public void release() {
		this.trackerModel().end();
	}

	/**
	 * Called by menu of 'Save...'
	 */
	public void saveLFT() {
		this.saveLFT10();
	}

	/**
	 * Save the flux object to a file.
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 */
	public void saveLFT10() {
		File file = StDialog.SaveFileDialog_default_("", this.fluxObject().name() + ".lft");

		if (file == null) {
			return;
		}

		try {
			this.writeToLFT10_object_(file, this.fluxObject());
		} catch (IOException e) {
			throw new SmalltalkException(e);
		}
	}

	/**
	 * Write the JunOpenGLFluxObject to the writer.
	 * 
	 * @param aWriter java.io.BufferedWriter
	 * @param anObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * 
	 * @exception IOException
	 */
	public void saveToLFT10_object_(BufferedWriter aWriter, JunOpenGLFluxObject anObject) throws IOException {
		aWriter.write(this.defaultStampForLFT10());
		anObject.saveOn_(aWriter);
	}

	/**
	 * Called by menu of 'spawn'
	 */
	public void spawnObject() {
		JunOpenGLDisplayModel displayModel = new JunOpenGLDisplayModel(this.spawningObject());
		displayModel.displayProjection_((JunOpenGLProjection) this.displayProjection().copy());
		displayModel.open();
		displayModel.changed_($("object"));
	}

	/**
	 * Start to TrackerModel.
	 */
	public void start() {
		this.trackerModel().start();
	}

	/**
	 * Answer trackerModel.
	 * 
	 * @return sra.jun.goodies.track.JunTrackerModel
	 */
	public JunTrackerModel trackerModel() {
		final JunOpenGLFluxModel this_ = this;

		if (trackerModel == null) {
			trackerModel = new JunTrackerModel();
			trackerModel.compute_(new StBlockClosure() {
				public Object value_(Object o) {
					double n = ((Number) o).doubleValue();

					if (fluxObject != null) {
						this_.displayObject_(fluxObject.at_(n));
						this_.changed_($("object"));
					}

					return null;
				}
			});
		}

		return trackerModel;
	}

	/**
	 * Update the indication of the misc menu.
	 */
	public void updateMiscMenuIndication() {
		// Nothing to do.
	}

	/**
	 * Write the JunOpenGLFluxObject to the file.
	 * 
	 * @param aFile java.io.File
	 * @param anObject sra.jun.opengl.flux.JunOpenGLFluxObject
	 * 
	 * @exception IOException
	 */
	public void writeToLFT10_object_(File aFile, JunOpenGLFluxObject anObject) throws IOException {
		if (aFile == null) {
			return;
		}

		BufferedWriter aWriter = new BufferedWriter(new FileWriter(aFile));

		try {
			this.saveToLFT10_object_(aWriter, anObject);
		} finally {
			aWriter.flush();
			aWriter.close();
		}
	}

	/**
	 * Answer the default file menu for JunOpenGLFluxModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultFileMenu() {
		StMenu defaultFileMenu = new StMenu("File");

		defaultFileMenu.add(createMenuItem("New", "newModel"));
		defaultFileMenu.add(createMenuItem("Open...", "openLFT"));
		defaultFileMenu.addSeparator();
		defaultFileMenu.add(createMenuItem("Save...", "saveLFT"));
		defaultFileMenu.addSeparator();
		defaultFileMenu.add(createMenuItem("Quit", "quitDoing"));

		return defaultFileMenu;
	}

	/**
	 * Answer the default misc menu for JunOpenGLFluxModel.
	 * 
	 * @return sra.smalltalk.StMenu
	 */
	protected StMenu _defaultMiscMenu() {
		StMenu defaultMiscMenu = new StMenu("Misc");

		defaultMiscMenu.add(createMenuItem("Spawn", "spawnObject"));
		defaultMiscMenu.add(createMenuItem("Viewport", "spawnViewport"));
		defaultMiscMenu.add(createMenuItem("Bounds", "showBounds"));

		return defaultMiscMenu;
	}

	/**
	 * Create the menubar for JunOpenGLDisplay.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();

		rootMenu.add(this._fileMenu());
		rootMenu.add(this._editMenu());
		rootMenu.add(this._viewMenu());
		rootMenu.add(this._lightMenu());
		rootMenu.add(this._miscMenu());

		return rootMenu;
	}

	/**
	 * Please refer to the corresponding method in Smalltalk.
	 * 
	 * @param miscMenu DOCUMENT ME!
	 */
	protected void updateMiscMenu_(Menu miscMenu) {
		// Nothing to do.
	}

	/**
	 * Answer a window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "OpenGL Flux Viewfinder";
	}
}
