package sra.jun.opengl.microcosm;

import java.awt.Color;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.curves.Jun3dLine;
import sra.jun.geometry.surfaces.JunSphereSurface;
import sra.jun.geometry.transformations.Jun3dTransformation;
import sra.jun.goodies.button.JunButtonModel;
import sra.jun.goodies.cursors.JunCursors;
import sra.jun.opengl.display.*;
import sra.jun.opengl.objects.JunOpenGL3dObject;
import sra.jun.opengl.projection.*;
import sra.smalltalk.*;

/**
 * JunOpenGLMicrocosmModel class
 * 
 *  @author    nisinaka
 *  @created   1999/08/17 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLMicrocosmModel.java,v 6.0 2003/10/28 10:52:18 nisinaka Exp $
 */
public class JunOpenGLMicrocosmModel extends JunOpenGLDisplayModel {

	/**
	 * Answer the default light color.
	 * 
	 * @return java.awt.Color
	 */
	public Color defaultLightColor() {
		return Color.getHSBColor(0f, 0f, 0.75f);
	}

	/**
	 * Answer the default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunOpenGLMicrocosmViewAwt(this);
		} else {
			return new JunOpenGLMicrocosmViewSwing(this);
		}
	}

	/**
	 * Answer the array of all display lights.
	 * 
	 * @return sra.jun.opengl.display.JunOpenGLDisplayLight[]
	 */
	public JunOpenGLDisplayLight[] displayLights() {
		if (displayLights == null) {
			displayLights = new JunOpenGLDisplayLight[5];
			displayLights[0] = JunOpenGLDisplayLight.ParallelLight_color_position_(true, this.defaultLightColor(), this.defaultLightPoint());
			displayLights[1] = JunOpenGLDisplayLight.PointLight_color_position_(true, Color.white, new Jun3dPoint(0, 0, 0));
			displayLights[2] = new JunOpenGLDisplayLight();
			displayLights[3] = new JunOpenGLDisplayLight();
			displayLights[4] = JunOpenGLDisplayLight.AmbientLight_color_(true, this.defaultLightColor());

			final JunOpenGLDisplayModel self = this;

			for (int i = 0; i < displayLights.length; i++) {
				displayLights[i].compute_(new StBlockClosure() {
					public Object value() {
						self.updateLightMenuIndication();
						self.changed_($("light"));

						return null;
					}
				});
			}
		}

		return displayLights;
	}

	/**
	 * Set a display object.
	 * 
	 * @param a3dObject sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public void displayObject_(JunOpenGL3dObject a3dObject) {
		super.displayObject_(a3dObject);

		//
		Jun3dBoundingBox box = this.boundingBox();
		Jun3dPoint center = (Jun3dPoint) box.center();
		Jun3dPoint sightPoint = (Jun3dPoint) center.minus_(new Jun3dPoint(0, 0, box.depth() * 2));
		this.defaultSightPoint_(sightPoint);
		this.defaultEyePoint_(center);
		this.defaultUpVector_(new Jun3dPoint(0, 1, 0));
		this.defaultViewFactor_(10);
		this.displayProjector();
	}

	/**
	 * Fit the 3d object to the current view size.
	 */
	public void fitSilently() {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjector projector = this.displayProjector();
		projector.sightPoint_(this.computeSightPoint());
		projector.eyePoint_(this.computeEyePoint());
		projector.upVector_(this.computeUpVector());
		projector.viewFactor_(this.computeViewFactor());
		projector.zoomHeight_(this.computeZoomHeight());
	}

	/**
	 * Grab and rotate the 3d object.
	 * 
	 * @param from2dPoint sra.jun.grometry.basic.Jun2dPoint
	 * @param to2dPoint sra.jun.grometry.basic.Jun2dPoint
	 */
	public void grab_xy_(Jun2dPoint from2dPoint, Jun2dPoint to2dPoint) {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjection aProjection = this.displayProjection();
		JunSphereSurface grabSphere = this.grabSphere();
		grabSphere.center_(aProjection.eyePoint());

		Jun3dPoint fromPoint = this.grab3dPoint_(new Jun2dPoint(from2dPoint.x(), from2dPoint.y()));
		Jun3dPoint toPoint = this.grab3dPoint_(new Jun2dPoint(to2dPoint.x(), to2dPoint.y()));
		JunAngle rotationAngle = (new Jun3dLine(grabSphere.center(), toPoint)).angleWithLine_(new Jun3dLine(grabSphere.center(), fromPoint));

		if (Math.abs(rotationAngle.rad()) > ThetaAccuracy) {
			Jun3dLine rotationAxis = new Jun3dLine(grabSphere.center(), (Jun3dPoint) grabSphere.center().minus_(((Jun3dPoint) fromPoint.minus_(grabSphere.center())).product_((Jun3dPoint) toPoint.minus_(grabSphere.center()))));
			Jun3dTransformation transformation = Jun3dTransformation.Rotate_around_(rotationAngle, rotationAxis);
			Jun3dTransformation transformationInv = Jun3dTransformation.Rotate_around_(rotationAngle.mul_(-1), rotationAxis);
			Jun3dPoint upPoint = (Jun3dPoint) aProjection.eyePoint().plus_(aProjection.unitUpVector());
			aProjection.sightPoint_(transformationInv.applyTo_(aProjection.sightPoint()));
			aProjection.upVector_((Jun3dPoint) transformation.applyTo_(upPoint).minus_(aProjection.sightPoint()));
			this.displayProjection_(aProjection);
			this.changed_($("projection"));
		}
	}

	/**
	 * Answer my grab button model.
	 * 
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	public JunButtonModel grabButton() {
		if (grabButton == null) {
			final JunOpenGLDisplayModel self = this;
			JunButtonModel button = new JunButtonModel();
			button.value_(false);
			button.visual_(JunCursors.ReadCursorImage());
			button.action_(new StBlockClosure() {
				public Object value_(Object o) {
					JunButtonModel model = (JunButtonModel) o;
					model.value_(!model.value());

					if (model.value()) {
						self.dragButton().value_(false);
						self.pickButton().value_(false);
					}

					self.changed_($("state"));

					return model;
				}
			});
			grabButton = button;
		}

		return grabButton;
	}

	/**
	 * Compute the current eye point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	protected Jun3dPoint computeEyePoint() {
		return (Jun3dPoint) this.boundingBox().center();
	}

	/**
	 * Compute the current sight point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	protected Jun3dPoint computeSightPoint() {
		Jun3dBoundingBox box = this.boundingBox();
		double z = Math.min(box.width(), Math.min(box.height(), box.depth())) * 2;

		return (Jun3dPoint) box.center().minus_(new Jun3dPoint(0, 0, z));
	}

	/**
	 * Compute the current up vector.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	protected Jun3dPoint computeUpVector() {
		return new Jun3dPoint(0, 1, 0);
	}

	/**
	 * Compute the current view factor.
	 * 
	 * @return double
	 */
	protected double computeViewFactor() {
		return 25;
	}

	/**
	 * Compute the current zoom height.
	 * 
	 * @return double
	 */
	protected double computeZoomHeight() {
		Jun3dBoundingBox box = this.boundingBox();

		return Math.min(box.width(), Math.min(box.height(), box.depth())) * 2.5;
	}

	/**
	 * Answer the window title.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "OpenGL Microcosm";
	}
}
