package sra.jun.opengl.objects;

import java.io.Writer;
import sra.jun.geometry.basic.Jun3dPoint;
import sra.jun.geometry.curves.JunNurbsCurve;
import sra.jun.geometry.transformations.JunTransformation;
import sra.jun.goodies.lisp.*;
import sra.jun.opengl.support.JunOpenGLRenderingContext;
import sra.smalltalk.*;

/**
 * JunOpenGL3dNurbsCurve class
 * 
 *  @author    nisinaka
 *  @created   1998/10/29 (by nisinaka)
 *  @updated   1999/08/04 (by nisinaka)
 *  @updated   2001/11/20 (by nisinaka)
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGL3dNurbsCurve.java,v 6.0 2003/10/28 10:52:19 nisinaka Exp $
 */
public class JunOpenGL3dNurbsCurve extends JunOpenGL3dPrimitiveObject {
	/** The line width. */
	protected float lineWidth;

	/** The stipple factor. */
	protected int stippleFactor = 1;

	/** The stipple pattern. */
	protected short stipplePattern;

	/** The array of the control points. */
	protected Jun3dPoint[] controlPoints = null;

	/** The array of weight numbers. */
	protected double[] weights = null;

	/** The knot vector. */
	protected double[] knotVector = null;

	/**
	 * Create an instance of JunOpenGL3dNurbsCurve and initialize it with the
	 * arguments.
	 * 
	 * @param anArrayOfPoints sra.jun.geometry.basic.Jun3dPoint[]
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dNurbsCurve
	 */
	public static final JunOpenGL3dNurbsCurve BezierControlPoints_(Jun3dPoint[] anArrayOfPoints) {
		int numberOfPoints = anArrayOfPoints.length;
		double[] knotVector = new double[numberOfPoints * 2];

		for (int i = 0; i < numberOfPoints; i++) {
			knotVector[i] = 0;
			knotVector[i + numberOfPoints] = 1;
		}

		return BSplineControlPoints_knotVector_(anArrayOfPoints, knotVector);
	}

	/**
	 * Create an instance of JunOpenGL3dNurbsCurve and initialize it with the
	 * arguments.
	 * 
	 * @param anArrayOfPoints sra.jun.geometry.basic.Jun3dPoint[]
	 * @param anArrayOfNumbers double[]
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dNurbsCurve
	 */
	public static final JunOpenGL3dNurbsCurve BSplineControlPoints_knotVector_(Jun3dPoint[] anArrayOfPoints, double[] anArrayOfNumbers) {
		int numberOfPoints = anArrayOfPoints.length;
		double[] weights = new double[numberOfPoints];

		for (int i = 0; i < numberOfPoints; i++) {
			weights[i] = 1.0;
		}

		return ControlPoints_weights_knotVector_(anArrayOfPoints, weights, anArrayOfNumbers);
	}

	/**
	 * Create an instance of JunOpenGL3dNurbsCurve and initialize it with the
	 * arguments.
	 * 
	 * @param anArrayOfPoints sra.jun.geometry.basic.Jun3dPoint[]
	 * @param anArrayOfNumber1 double[]
	 * @param anArrayOfNumber2 double[]
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dNurbsCurve
	 */
	public static final JunOpenGL3dNurbsCurve ControlPoints_weights_knotVector_(Jun3dPoint[] anArrayOfPoints, double[] anArrayOfNumber1, double[] anArrayOfNumber2) {
		JunOpenGL3dNurbsCurve nurbsCurve = new JunOpenGL3dNurbsCurve();
		nurbsCurve.controlPoints_(anArrayOfPoints);
		nurbsCurve.weights_(anArrayOfNumber1);
		nurbsCurve.knotVector_(anArrayOfNumber2);

		return nurbsCurve;
	}

	/**
	 * Convert the receiver as a JunNurbsCurve.
	 * 
	 * @return sra.jun.geometry.curves.JunNurbsCurve
	 */
	public JunNurbsCurve _asJunNurbsCurve() {
		return JunNurbsCurve.ControlPoints_weights_knotVector_(controlPoints, weights, knotVector);
	}

	/**
	 * Answer the control points of the receiver.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint[]
	 */
	public Jun3dPoint[] controlPoints() {
		return controlPoints;
	}

	/**
	 * Set the control points of the receiver.
	 * 
	 * @param anArray sra.jun.geometry.basic.Jun3dPoint[]
	 */
	public void controlPoints_(Jun3dPoint[] anArray) {
		controlPoints = anArray;
	}

	/**
	 * Convert the receiver's contol points as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList controlsToLispList() {
		Object[] collection = new Object[controlPoints.length + 1];
		collection[0] = $("controls");

		for (int i = 0; i < controlPoints.length; i++) {
			collection[i + 1] = controlPoints[i];
		}

		return JunLispCons.List_(collection);
	}

	/**
	 * Enumerate every geometries and evaluate the Block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return java.lang.Object
	 */
	public Object geometriesDo_(StBlockClosure aBlock) {
		return aBlock.value_(JunNurbsCurve.ControlPoints_weights_knotVector_(controlPoints, weights, knotVector));
	}

	/**
	 * Answer the StSymbol which represents the kind of the receiver.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public StSymbol kindName() {
		return $("NurbsCurve");
	}

	/**
	 * Convert the receiver's knot vector as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList knotsToLispList() {
		Object[] collection = new Object[knotVector.length + 1];
		collection[0] = $("knots");

		for (int i = 0; i < knotVector.length; i++) {
			collection[i + 1] = new Double(knotVector[i]);
		}

		return JunLispCons.List_(collection);
	}

	/**
	 * Answer the knot vector.
	 * 
	 * @return double[]
	 */
	public double[] knotVector() {
		return knotVector;
	}

	/**
	 * Set the knot vector.
	 * 
	 * @param anArray double[]
	 */
	public void knotVector_(double[] anArray) {
		knotVector = anArray;
	}

	/**
	 * Answer the line width.
	 * 
	 * @return float
	 */
	public float lineWidth() {
		return lineWidth;
	}

	/**
	 * Set a number as the line width.
	 * 
	 * @param aNumber float
	 */
	public void lineWidth_(float aNumber) {
		lineWidth = aNumber;
	}

	/**
	 * Convert the receiver's stipple as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList lineWidthToLispList() {
		if (this.lineWidth <= 1) {
			return this.lispNil();
		}

		JunLispCons list = this.lispCons();
		list.head_($("width"));
		list.tail_(new Float(this.lineWidth()));

		return list;
	}

	/**
	 * Answer the number of polygons.
	 * 
	 * @return int
	 */
	public int numberOfPolygons() {
		return 0;
	}

	/**
	 * Enumerate every points and evaluate the Block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public void pointsDo_(StBlockClosure aBlock) {
		for (int i = 0; i < controlPoints.length; i++) {
			aBlock.value_(controlPoints[i]);
		}
	}

	/**
	 * Render the OpenGL object on a rendering context.
	 * 
	 * @param aRenderingContext sra.jun.opengl.support.JunRenderingContext
	 */
	public void renderOn_(JunOpenGLRenderingContext aRenderingContext) {
		if (this.hasPaint()) {
			aRenderingContext.paint_(this.paint());
		}

		if (this.hasAlpha()) {
			aRenderingContext.alpha_(this.alpha());
		}

		aRenderingContext.lineWidth_(lineWidth);

		if (this.hasStipple()) {
			aRenderingContext.lineStippleFactor_pattern_(this.stippleFactor(), this.stipplePattern());
			aRenderingContext.enableLineStipple();
		}

		aRenderingContext.displayNurbsCurve_(this._asJunNurbsCurve());

		if (this.hasStipple()) {
			aRenderingContext.disableLineStipple();
		}
	}

	/**
	 * Anser the reversed JunOpenGL3dObject of the receiver.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 */
	public JunOpenGL3dObject reversed() {
		throw new SmalltalkException("not implemented yet");
	}

	/**
	 * Answer the stipple factor.
	 * 
	 * @return int
	 */
	public int stippleFactor() {
		return stippleFactor;
	}

	/**
	 * Set a number as the stipple factor.
	 * 
	 * @param aNumber int
	 */
	public void stippleFactor_(int aNumber) {
		stippleFactor = aNumber;
	}

	/**
	 * Answer the stipple pattern
	 * 
	 * @return short
	 */
	public short stipplePattern() {
		return stipplePattern;
	}

	/**
	 * Set a number as the stipple pattern.
	 * 
	 * @param aNumber short
	 */
	public void stipplePattern(short aNumber) {
		stipplePattern = aNumber;
	}

	/**
	 * Convert the receiver's stipple as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList stippleToLispList() {
		if (!this.hasStipple()) {
			return this.lispNil();
		}

		JunLispCons list = this.lispCons();
		list.head_($("stipple"));
		list.add_(new Integer(this.stippleFactor()));
		list.add_(new Short(this.stipplePattern()));

		return list;
	}

	/**
	 * Convert the receiver's texture as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList textureToLispList() {
		throw SmalltalkException.ShouldNotImplement();
	}

	/**
	 * Convert the receiver as a JunLispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList toLispList() {
		JunLispCons list = this.lispCons();
		list.head_(this.kindName());

		if (this.hasName()) {
			list.add_(this.nameToLispList());
		}

		if (this.hasColor()) {
			list.add_(this.colorToLispList());
		}

		if (this.lineWidth() > 1) {
			list.add_(this.lineWidthToLispList());
		}

		if (this.hasStipple()) {
			list.add_(this.stippleToLispList());
		}

		if (this.hasTexture()) {
			list.add_(this.textureToLispList());
		}

		list.add_(this.controlsToLispList());
		list.add_(this.weightsToLispList());
		list.add_(this.knotsToLispList());

		return list;
	}

	/**
	 * Answer the new JunOpen3dObject transformed with aTransformation.
	 * 
	 * @param aTransformation
	 *        sra.jun.geometry.transformations.JunTransformation
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 * 
	 * @throws SmalltalkException DOCUMENT ME!
	 */
	public JunOpenGL3dObject transform_(JunTransformation aTransformation) {
		throw new SmalltalkException("not implemented yet");
	}

	/**
	 * not supported in original smalltalk jun
	 * 
	 * @param writer java.io.Writer
	 */
	public void vrml10On_(Writer writer) {
	}

	/**
	 * not supported in original smalltalk jun
	 * 
	 * @param writer java.io.Writer
	 */
	public void vrml20On_(Writer writer) {
	}

	/**
	 * Answer the array of weight number.
	 * 
	 * @return double[]
	 */
	public double[] weights() {
		return weights;
	}

	/**
	 * Set the weight numbers.
	 * 
	 * @param anArrayOfDouble double[]
	 */
	public void weights_(double[] anArrayOfDouble) {
		weights = anArrayOfDouble;
	}

	/**
	 * Convert the receiver's weight numbers as a LispList.
	 * 
	 * @return sra.jun.goodies.lisp.JunLispList
	 */
	public JunLispList weightsToLispList() {
		Object[] collection = new Object[weights.length + 3];
		collection[0] = $("weights");

		for (int i = 0; i < weights.length; i++) {
			collection[i + 1] = new Double(weights[i]);
		}

		return JunLispCons.List_(collection);
	}
}
