package sra.jun.opengl.projection;

import java.util.*;
import sra.smalltalk.*;
import sra.jun.geometry.abstracts.JunGeometry;
import sra.jun.geometry.basic.Jun3dPoint;
import sra.jun.opengl.lights.*;
import sra.jun.opengl.objects.JunOpenGL3dObject;
import sra.jun.opengl.support.JunOpenGLRenderingContext;

/**
 * JunOpenGLProjector class
 * 
 *  @author    MATSUDA Ryouichi
 *  @created   1998/10/29 (by MATSUDA Ryouichi)
 *  @updated   1999/06/25 (by nisinaka)
 *  @updated   2003/03/07 (by nisinaka)
 *  @updated   2003/05/13 (by nisinaka)
 *  @version   406b (with StPL6.0) based on Jun481 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunOpenGLProjector.java,v 6.0 2003/10/28 10:52:21 nisinaka Exp $
 */
public class JunOpenGLProjector extends StObject implements StBlockValued {

	/** The projection. */
	protected JunOpenGLProjection projection;

	/** The symbol represents a shading. */
	protected StSymbol shading = null;

	/** The symbol represents a presentation. */
	protected StSymbol presentation = null;

	/** The lights is a Vector. */
	protected Vector lights; // JunOpenGLLight or StBlockValue

	/** Whether do a line smoothing effect or not. */
	protected boolean lineSmoothBoolean;

	/** Whether do a polygon smoothing effect or not. */
	protected boolean polygonSmoothBoolean;

	/**
	 * Constainer.
	 */
	public JunOpenGLProjector() {
		super();
		this.initialize();
	}

	/**
	 * Add all Lights.
	 * 
	 * @param aCollectionOfJunOpenGLLight java.util.Vector
	 */
	public void addAllLights_(Vector aCollectionOfJunOpenGLLight) {
		lights.addAll(aCollectionOfJunOpenGLLight);
	}

	/**
	 * Add head light.
	 */
	public void addHeadLight() {
		final JunOpenGLSpotLight headLight = JunOpenGLSpotLight.At3dPoint_(this.eyePoint());
		this.addLight_(new StBlockValue(new StBlockClosure() {
			public Object value_(Object projector) {
				headLight.position_(((JunOpenGLProjector) projector).eyePoint());
				return headLight;
			}
		}, this));
	}

	/**
	 * Add a light.
	 * 
	 * @param aJunOpenGLLight sra.smalltalk.StBlockValue
	 */
	public void addLight_(JunOpenGLLight aJunOpenGLLight) {
		lights.add(aJunOpenGLLight);
	}

	/**
	 * Add a light.
	 * 
	 * @param aJunOpenGLLight sra.smalltalk.StBlockValue
	 */
	public void addLight_(StBlockValue aJunOpenGLLight) {
		lights.add(aJunOpenGLLight);
	}

	/**
	 * Answer a BlockValue that computes aBlock with the receiver's value  as
	 * the argument. aBlock will become a dependent of the receiver, and will
	 * be sent the message value: when the receiver is sent the message
	 * value:.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 * 
	 * @return sra.smalltalk.StBlockValue
	 */
	public StBlockValue compute_(StBlockClosure aBlock) {
		return new StBlockValue(aBlock, this);
	}

	/**
	 * Distance.
	 * 
	 * @return double
	 */
	public double distance() {
		return projection.distance();
	}

	/**
	 * Answer eyePoint.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint eyePoint() {
		return projection.eyePoint();
	}

	/**
	 * Set a 3dPoint to eyePoint.
	 * 
	 * @param aJun3dPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public void eyePoint_(Jun3dPoint aJun3dPoint) {
		if ((projection.eyePoint() == null) || (this.sightPoint().distance_(aJun3dPoint) > JunGeometry.Accuracy())) {
			projection.eyePoint_(aJun3dPoint);
		}
	}

	/**
	 * Answer far.
	 * 
	 * @return double
	 */
	public double far() {
		return projection.far();
	}

	/**
	 * Set a number to far.
	 * 
	 * @param aNumber double
	 */
	public void far_(double aNumber) {
		projection.far_(aNumber);
	}

	/**
	 * Set flatShading mode.
	 */
	public void flatShading() {
		this.shading_($("flatShading"));
	}

	/**
	 * Set the presentation mode as hidden-line.
	 */
	public void hiddenlinePresentation() {
		this.presentation_($("hiddenlinePresentation"));
	}

	/**
	 * Enumerate every lights and evaluate a Block.
	 * 
	 * @param aBlock sra.smalltalk.StBlockClosure
	 */
	public void lightsDo_(StBlockClosure aBlock) {
		StValued[] values = new StValued[lights.size()];
		lights.copyInto(values);
		for (int i = 0; i < values.length; i++) {
			aBlock.value_(values[i].value());
		}
	}

	/**
	 * Answer whether do a line smoothing effect or not.
	 * 
	 * @return boolean
	 */
	public boolean lineSmooth() {
		return lineSmoothBoolean;
	}

	/**
	 * Set whether do a line smoothing effect or not.
	 * 
	 * @param aBoolean boolean
	 */
	public void lineSmooth_(boolean aBoolean) {
		lineSmoothBoolean = aBoolean;
	}

	/**
	 * Answer near.
	 * 
	 * @return double
	 */
	public double near() {
		return projection.near();
	}

	/**
	 * Set a number to near.
	 * 
	 * @param aNumber double
	 */
	public void near_(double aNumber) {
		projection.near_(aNumber);
	}

	/**
	 * Arrange to receive a message with aSymbol when the value aspect  changes
	 * on anObject.
	 * 
	 * @param aSymbol java.lang.String
	 * @param anObject sra.smalltalk.StObject
	 */
	public void onChangeProjectionSend_to_(String aSymbol, StObject anObject) {
		this.expressInterestIn_for_sendBack_($("projection"), anObject, aSymbol);
	}

	/**
	 * Arrange to receive a message with aSymbol when the value aspect  changes
	 * on anObject.
	 * 
	 * @param aSymbol java.lang.String
	 * @param anObject sra.smalltalk.StObject
	 */
	public void onChangeSend_to_(String aSymbol, StObject anObject) {
		this.expressInterestIn_for_sendBack_($("value"), anObject, aSymbol);
	}

	/**
	 * Pan.
	 */
	public void pan() {
		projection.pan();
	}

	/**
	 * Pan with factor.
	 * 
	 * @param factor double
	 */
	public void pan_(double factor) {
		projection.pan_(factor);
	}

	/**
	 * Answer a parallelProjection.
	 */
	public void parallelProjection() {
		this.projection_(projection.asParallelProjection());
	}

	/**
	 * Answer a Perspective projection.
	 *
	 * @deprecated since Jun481
	 */
	public void perspective() {
		this.projection_(projection.asPerspective());
	}

	/**
	 * Answer a Perspective projection.
	 */
	public void perspectiveProjection() {
		this.projection_(projection.asPerspective());
	}

	/**
	 * Answer whether do a polygon smoothing effect or not.
	 * 
	 * @return boolean
	 */
	public boolean polygonSmooth() {
		return polygonSmoothBoolean;
	}

	/**
	 * Set whether do a polygon smoothing effect or not.
	 * 
	 * @param aBoolean boolean
	 */
	public void polygonSmooth_(boolean aBoolean) {
		polygonSmoothBoolean = aBoolean;
	}

	/**
	 * Do not copy the dependents list.
	 * 
	 * @return sra.smalltalk.StObject
	 */
	public StObject postCopy() {
		super.postCopy();
		this.breakDependents();

		return this;
	}

	/**
	 * Answer a symbol of the presentation.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public final StSymbol presentation() {
		return presentation;
	}

	/**
	 * Set a symbol to the presentation.
	 * 
	 * @param aSymbol sra.smalltalk.StSymbol
	 */
	public final void presentation_(StSymbol aSymbol) {
		presentation = aSymbol;
	}

	/**
	 * Project on a RenderingContext.
	 * 
	 * @param a3dObjectOrModel sra.jun.opengl.objects.JunOpenGL3dObject
	 * @param aRenderingContext
	 *        sra.jun.opengl.support.JunOpenGLRenderingContext
	 */
	public void project_on_(JunOpenGL3dObject a3dObjectOrModel, JunOpenGLRenderingContext aRenderingContext) {
		this.project_withLights_on_(a3dObjectOrModel, null, aRenderingContext);
	}

	/**
	 * Project on a RenderingContext with lights.
	 * 
	 * @param a3dObjectOrModel sra.jun.opengl.objects.JunOpenGL3dObject
	 * @param aCollectionOfJunOpenGLLight sra.jun.opengl.lights.JunOpenGLLight[]
	 * @param aRenderingContext sra.jun.opengl.support.JunOpenGLRenderingContext
	 */
	public void project_withLights_on_(JunOpenGL3dObject a3dObjectOrModel, JunOpenGLLight[] aCollectionOfJunOpenGLLight, JunOpenGLRenderingContext aRenderingContext) {
		this.projectOn_(aRenderingContext);

		if (aCollectionOfJunOpenGLLight != null) {
			for (int index = 0; index < aCollectionOfJunOpenGLLight.length; index++) {
				aCollectionOfJunOpenGLLight[index].lightingOn_(aRenderingContext);
			}
		}

		try {
			if (a3dObjectOrModel != null) {
				a3dObjectOrModel.renderOn_(this.renderingContextFrom_(aRenderingContext));
			}
		} finally {
			aRenderingContext.flush();
		}
	}

	/**
	 * Answer a projection.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjection
	 */
	public JunOpenGLProjection projection() {
		return (JunOpenGLProjection) projection.copy();
	}

	/**
	 * Set a projection.
	 * 
	 * @param aProjection sra.jun.opengl.projection.JunOpenGLProjection
	 */
	public void projection_(JunOpenGLProjection aProjection) {
		projection = (JunOpenGLProjection) aProjection.copy();
	}

	/**
	 * Project on a RenderingContext.
	 * 
	 * @param aRenderingContext
	 *        sra.jun.opengl.support.JunOpenGLRenderingContext
	 */
	public void projectOn_(final JunOpenGLRenderingContext aRenderingContext) {
		try {
			projection.projectOn_(aRenderingContext);
			this.lightsDo_(new StBlockClosure() {
				public Object value_(Object light) {
					((JunOpenGLLight) light).lightingOn_(aRenderingContext);
					return null;
				}
			});
		} finally {
			aRenderingContext.flush();
		}
	}

	/**
	 * Remove all Lights.
	 * 
	 * @param aCollectionOfJunOpenGLLight java.util.Vector
	 */
	public void removeAllLight_(Vector aCollectionOfJunOpenGLLight) {
		lights.removeAll(aCollectionOfJunOpenGLLight);
	}

	/**
	 * Remove a light.
	 * 
	 * @param aJunOpenGLLight sra.jun.opengl.lights.JunOpenGLLight
	 */
	public void removeLight_(JunOpenGLLight aJunOpenGLLight) {
		lights.remove(aJunOpenGLLight);
	}

	/**
	 * Undo a send of onChangeSend:to:.
	 * 
	 * @param anObject sra.smalltalk.StObject
	 */
	public void retractInterestsFor_(StObject anObject) {
		this.retractInterestIn_for_($("value"), anObject);
	}

	/**
	 * Undo a send of onChangeSend:to:.
	 * 
	 * @param anObject sra.smalltalk.StObject
	 */
	public void retractProjectionInterestsFor_(StObject anObject) {
		this.retractInterestIn_for_($("projection"), anObject);
	}

	/**
	 * Answer a symbol of shading.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public final StSymbol shading() {
		return shading;
	}

	/**
	 * Set a symbol to the shading.
	 * 
	 * @param aSymbol sra.smalltalk.StSymbol
	 */
	public final void shading_(StSymbol aSymbol) {
		shading = aSymbol;
	}

	/**
	 * Answer sightPoint.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint sightPoint() {
		return projection.sightPoint();
	}

	/**
	 * Set sightPoint.
	 * 
	 * @param aJun3dPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public void sightPoint_(Jun3dPoint aJun3dPoint) {
		if ((projection.sightPoint() == null) || (this.eyePoint().distance_(aJun3dPoint) > JunGeometry.Accuracy())) {
			projection.sightPoint_(aJun3dPoint);
		}
	}

	/**
	 * Set smoothShading.
	 */
	public void smoothShading() {
		this.shading_($("smoothShading"));
	}

	/**
	 * Set the presentation mode as solid.
	 */
	public void solidPresentation() {
		this.presentation_($("solidPresentation"));
	}

	/**
	 * Answer unit RightVector.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint unitRightVector() {
		return projection.unitRightVector();
	}

	/**
	 * Answer unit SightVector.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint unitSightVector() {
		return projection.unitSightVector();
	}

	/**
	 * Answer unit UpVector.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint unitUpVector() {
		return projection.unitUpVector();
	}

	/**
	 * Answer upVector.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint upVector() {
		return projection.upVector();
	}

	/**
	 * Set upVector.
	 * 
	 * @param a3dPoint sra.jun.geometry.basic.Jun3dPoint
	 */
	public void upVector_(Jun3dPoint a3dPoint) {
		projection.upVector_(a3dPoint);
	}

	/**
	 * Answer a receiver.
	 * 
	 * @return java.lang.Object
	 */
	public Object value() {
		return this;
	}

	/**
	 * Answer viewFactor.
	 * 
	 * @return double
	 */
	public double viewFactor() {
		return projection.viewFactor();
	}

	/**
	 * Set viewFactor.
	 * 
	 * @param factor double
	 */
	public void viewFactor_(double factor) {
		factor = (factor < 2) ? 2 : factor;
		factor = (1000 < factor) ? 1000 : factor;
		projection.viewFactor_(factor);
	}

	/**
	 * Set the presentation mode as wireframe.
	 */
	public void wireframePresentation() {
		this.presentation_($("wireframePresentation"));
	}

	/**
	 * Zoom.
	 */
	public void zoom() {
		projection.zoom();
	}

	/**
	 * Zoom with factor.
	 * 
	 * @param factor double
	 */
	public void zoom_(double factor) {
		projection.zoom_(factor);
	}

	/**
	 * Answer my zoom height.
	 *
	 * @return double
	 */
	public double zoomHeight() {
		return projection.zoomHeight();
	}

	/**
	 * Zoom height.
	 * 
	 * @param aNumber double
	 */
	public void zoomHeight_(double aNumber) {
		projection.zoomHeight_(aNumber);
	}

	/**
	 * Answer a symbol of default Presentation.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	protected StSymbol defaultPresentation() {
		return $("solidPresentation");
	}

	/**
	 * Answer a default projection.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjection
	 */
	protected JunOpenGLProjection defaultProjection() {
		return JunOpenGLProjection.Default();
	}

	/**
	 * Answer a symbol of default Shading.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	protected StSymbol defaultShading() {
		return $("flatShading");
	}

	/**
	 * Initialize the projector.
	 */
	protected void initialize() {
		shading = this.defaultShading();
		presentation = this.defaultPresentation();
		projection = this.defaultProjection();
		lights = new Vector();
	}

	/**
	 * Set projector mode to a RenderingContext and answer it.
	 * 
	 * @return sra.jun.opengl.support.JunOpenGLRenderingContext
	 * @param aRenderingContext sra.jun.opengl.support.JunOpenGLRenderingContext
	 */
	protected JunOpenGLRenderingContext renderingContextFrom_(JunOpenGLRenderingContext aRenderingContext) {
		if (presentation == $("solidPresentation")) {
			aRenderingContext.solidPresentation();
		} else if (presentation == $("wireframePresentation")) {
			aRenderingContext.wireframePresentation();
		} else if (presentation == $("hiddenlinePresentation")) {
			aRenderingContext.hiddenlinePresentation();
		}

		if (shading == $("smoothShading")) {
			aRenderingContext.smoothShading();
		} else if (shading == $("flatShading")) {
			aRenderingContext.flatShading();
		}

		if (this.lineSmooth()) {
			aRenderingContext.lineSmoothNicest();
			aRenderingContext.enableLineSmooth();
		} else {
			aRenderingContext.disableLineSmooth();
		}

		if (this.polygonSmooth()) {
			aRenderingContext.polygonSmoothFastest();
			aRenderingContext.enablePolygonSmooth();
		} else {
			aRenderingContext.disablePolygonSmooth();
		}

		return aRenderingContext;
	}

	/**
	 * Set a projection.
	 * 
	 * @param aProjection sra.jun.opengl.projection.JunOpenGLProjection
	 */
	protected final void setProjection_(JunOpenGLProjection aProjection) {
		projection = aProjection;
	}
}
