package sra.jun.system.framework;

import java.awt.*;
import java.awt.event.*;
import java.text.SimpleDateFormat;
import java.util.*;
import javax.swing.*;
import sra.smalltalk.*;

/**
 * JunApplicationModel class
 * 
 *  @author    Mitsuhiro Asada
 *  @created   2002/11/07 (by Mitsuhiro Asada)
 *  @updated   2003/03/03 (by Mitsuhiro Asada)
 *  @version   406b (with StPL6.0) based on Jun475 for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunApplicationModel.java,v 6.0 2003/10/28 10:52:22 nisinaka Exp $
 */
public abstract class JunApplicationModel extends StApplicationModel {
	/**
	 * An utility method to make a rectangle visible on the screen.
	 * This method was defined in Screen class.
	 *
	 * @return java.awt.Rectangle
	 * @param aRectangle java.awt.Rectangle
	 * @category Utilities
	 */
	public static Rectangle _MakeRectangleVisible(Rectangle aRectangle) {
		int dx = 0;
		int dy = 0;

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		if (aRectangle.x < 0) {
			dx = aRectangle.x;
		} else if (aRectangle.x + aRectangle.width >= screenSize.width) {
			dx = screenSize.width - (aRectangle.x + aRectangle.width);
		}
		if (aRectangle.y < 0) {
			dy = aRectangle.y;
		} else if (aRectangle.y + aRectangle.height >= screenSize.height) {
			dy = screenSize.height - (aRectangle.y + aRectangle.height);
		}

		aRectangle.translate(dx, dy);
		return aRectangle;
	}

	/**
	 * Answer the string of default base name.
	 *
	 * @return java.lang.String
	 * @category Defaults
	 */
	public static String DefaultBaseName() {
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyyMMddHHmmss");
		return dateFormat.format(new Date());
	}

	/**
	 * Add the KeyListener to the Component.
	 * If the Component is a Container, do the setting to its components.
	 *
	 * @param aKeyListener java.awt.event.KeyListener
	 * @param aComponent java.awt.Component
	 * @category Events
	 */
	private static void _SetKeyListener(KeyListener aKeyListener, Component aComponent) {
		if (aComponent instanceof Container && ((Container) aComponent).getComponentCount() > 0) {
			Component[] components = ((Container) aComponent).getComponents();
			for (int i = 0; i < components.length; i++) {
				_SetKeyListener(aKeyListener, components[i]);
			}
		} else {
			aComponent.addKeyListener(aKeyListener);

			// As part of the new focus architecture introduced in 1.4.0, 
			// focus traversal keys (in most cases this means the <code>TAB</code> key) are now consumed,
			// which can cause problems if a program depends on a key listener being notified of these key events.
			// @see http://developer.java.sun.com/developer/bugParade/bugs/4650902.html
			if (System.getProperty("java.version").startsWith("1.4")) {
				try {
					java.lang.reflect.Method method = Component.class.getMethod("setFocusTraversalKeysEnabled", new Class[] { Boolean.TYPE });
					method.invoke(aComponent, new Object[] { Boolean.FALSE });
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 * Set my KeyListener to the view.
	 *
	 * @param aView sra.smalltalk.StView
	 * @category interface opening
	 */
	public void _setKeyListener(StView aView) {
		KeyListener aKeyListener = this._keyListener();
		if (aKeyListener != null) {
			_SetKeyListener(aKeyListener, aView.toComponent());
		}
	}

	/**
	 * Answer the string of default base name.
	 *
	 * @return java.lang.String
	 * @category defaults
	 */
	public String defaultBaseName() {
		return DefaultBaseName();
	}

	/**
	 * Answer the point of default offset.
	 * 
	 * @return java.awt.Point
	 * @category defaults
	 */
	public Point defaultOffsetPoint() {
		return new Point(25, 25);
	}

	/**
	 * Answer a controller.
	 * 
	 * @return sra.smalltalk.StController
	 * @category accessing
	 */
	public StController getController() {
		StView aView = this.getView();
		if (aView == null) {
			return null;
		}
		return aView.controller();
	}

	/**
	 * Open this application window.
	 * 
	 * @return java.awt.Frame
	 * @category interface opening
	 */
	public Frame open() {
		Frame aFrame = super.open();
		this.updateMenuIndication();
		this.setWindowLabel();
		return aFrame;
	}

	/**
	 * Create a Frame for a default view and open it in the specified area.
	 *
	 * @return java.awt.Frame
	 * @param aRectangle java.awt.Rectangle
	 * @category interface opening
	 */
	public Frame openIn_(Rectangle aRectangle) {
		return this.openViewIn_(this.defaultView(), aRectangle);
	}

	/**
	 * Create a Frame for the view and open it at the specified point.
	 *
	 * @return java.awt.Frame
	 * @param aView sra.smalltalk.StView
	 * @param aPoint java.awt.Point
	 * @category interface opening
	 */
	public Frame openViewAt_(StView aView, Point aPoint) {
		Frame aFrame = this.allButOpenView_(aView);
		aFrame.setLocation(aPoint);
		aFrame.show();
		return aFrame;
	}

	/**
	 * Create a Frame for the view and open it in the specified area.
	 *
	 * @return java.awt.Frame
	 * @param aView sra.smalltalk.StView
	 * @param aRectangle java.awt.Rectangle
	 * @category interface opening
	 */
	public Frame openViewIn_(StView aView, Rectangle aRectangle) {
		Frame aFrame = this.allButOpenView_(aView);
		aFrame.setBounds(_MakeRectangleVisible(aRectangle));
		aFrame.show();
		return aFrame;
	}

	/**
	 * Set window label.
	 * 
	 * @category interface opening
	 */
	public void setWindowLabel() {
		Window aWindow = this.getWindow();
		if (aWindow == null) {
			return;
		}

		if (aWindow instanceof Frame) {
			((Frame) aWindow).setTitle(this.windowTitle());
		}
	}

	/**
	 * Update the menu indication.
	 * 
	 * @category menu accessing
	 */
	public void updateMenuIndication() {
		return;
	}

	/**
	 * Update this window.
	 * 
	 * @category updating
	 */
	public void updateWindow() {
		this.flushMenus();
		this.updateMenus();
		this.setWindowLabel();
		this.flushVisuals();
		Window aWindow = this.getWindow();
		if (aWindow != null) {
			aWindow.repaint();
		}
	}

	/**
	 * Answer the collection of an instance of frame with the menu.
	 *
	 * @return java.util.Vector
	 * @param aComponent java.awt.Component
	 * @param aCollection java.uitl.Vector
	 * @category menu accessing
	 */
	private Vector _getFrameHavingMenu_(Component aComponent, Vector aCollection) {
		if (aComponent instanceof Frame && ((Frame) aComponent).getMenuBar() != null) {
			aCollection.add(aComponent);
		}
		if (aComponent instanceof JFrame && ((JFrame) aComponent).getJMenuBar() != null) {
			aCollection.add(aComponent);
		}
		if (aComponent instanceof Container) {
			Component[] components = ((Container) aComponent).getComponents();
			for (int i = 0; i < components.length; i++) {
				this._getFrameHavingMenu_(components[i], aCollection);
			}
		}
		return aCollection;
	}

	/**
	 * Answer my KeyListener.
	 * If not null, the KeyListener will be added to all components of the view.
	 * The subclasses may override this method to handle key events.
	 *
	 * @return java.awt.event.KeyListener
	 * @category events
	 */
	protected KeyListener _keyListener() {
		return null;
	}

	/**
	 * Create a Frame for the application model.
	 * Do everything but open the frame.
	 *
	 * @return java.awt.Frame
	 * @param aView sra.smalltalk.StView
	 * @category interface opening
	 */
	protected Frame allButOpenView_(StView aView) {
		this._setKeyListener(aView);
		return super.allButOpenView_(aView);
	}

	/**
	 * Clean up this menus.
	 * 
	 * @category flushing
	 */
	protected void flushMenus() {
	}

	/**
	 * Clean up this visuals.
	 * 
	 * @category flushing
	 */
	protected void flushVisuals() {
	}

	/**
	 * Answer a view.
	 * 
	 * @return sra.smalltalk.StView
	 * @category accessing
	 */
	public StView getView() {
		Object[] dependents = this.dependents();
		for (int i = 0; i < dependents.length; i++) {
			Object each = dependents[i];
			if (each instanceof StView && ((StView) each).model() == this) {
				return (StView) each;
			}
		}

		return null;
	}

	/**
	 * Answer my window.
	 * 
	 * @return java.awt.Window
	 * @category accessing
	 */
	public Window getWindow() {
		StView aView = this.getView();
		if (aView == null) {
			return null;
		}
		return aView.topComponent();
	}

	/**
	 * Answer a margin rectangle for visible screen area.
	 * 
	 * @return java.awt.Rectangle
	 * @category private
	 */
	protected Rectangle marginRectangleForVisibleScreenArea() {
		return new Rectangle(10, 30, 0, -20);
	}

	/**
	 * Set the size for the view.
	 * 
	 * @param extent java.awt.Dimension
	 * @category private
	 */
	protected void setSize_(Dimension extent) {
		StView aView = this.getView();
		if (aView == null) {
			return;
		}

		Window aWindow = aView.topComponent();
		if (aWindow == null) {
			return;
		}

		Dimension oldExtent = ((Component) aView).getSize();
		Dimension newExtent = extent;
		int width = aWindow.getWidth() - oldExtent.width + newExtent.width;
		int height = aWindow.getHeight() - oldExtent.height + newExtent.height;
		aWindow.setSize(width, height);
		aWindow.validate();
	}

	/**
	 * Update this menus.
	 * 
	 * @category private
	 */
	protected void updateMenus() {
		this.flushMenus();
		Window aWindow = this.getWindow();
		if (aWindow == null) {
			return;
		}
		Vector aCollection = new Vector();
		this._getFrameHavingMenu_(aWindow, aCollection);
		for (Enumeration enum = aCollection.elements(); enum.hasMoreElements();) {
			Frame aFrame = (Frame) enum.nextElement();
			StMenuModel menuModel = this.menuModel();
			if (menuModel != null) {
				if (aFrame instanceof JFrame) {
					JMenuBar aMenuBar = new JMenuBar();
					StMenuViewForSwing aMenuView = new StMenuViewForSwing(aMenuBar);
					aMenuView.setModel(menuModel);
					((JFrame) aFrame).setJMenuBar(aMenuBar);
				} else {
					MenuBar aMenuBar = new MenuBar();
					StMenuViewForAwt aMenuView = new StMenuViewForAwt(aMenuBar);
					aMenuView.setModel(menuModel);
					aFrame.setMenuBar(aMenuBar);
				}
				aFrame.pack();
			}
		}
	}
}
