package sra.jun.terrain.editor;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.*;
import sra.smalltalk.*;
import sra.jun.delaunay.twoD.Jun2dDelaunayProcessor;
import sra.jun.geometry.basic.*;
import sra.jun.geometry.boundaries.JunBorderGenerator;
import sra.jun.geometry.curves.Jun2dLine;
import sra.jun.goodies.button.JunButtonModel;
import sra.jun.goodies.cursors.JunCursors;
import sra.jun.goodies.progress.JunCursorAnimator;
import sra.jun.goodies.wheels.JunThumbWheel;
import sra.jun.opengl.display.JunOpenGL3dModel;
import sra.jun.opengl.objects.*;
import sra.jun.opengl.projection.*;
import sra.jun.terrain.display.JunTerrainDisplayModel;
import sra.jun.terrain.support.*;

/**
 * JunTerrainEditModel class
 * 
 *  @author    nisinaka
 *  @created   2001/11/19 (by nisinaka)
 *  @updated   N/A
 *  @version   406b (with StPL6.0) based on JunXXX for Smalltalk
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 1999-2003 Information-technology Promotion Agency,Japan (IPA)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: JunTerrainEditModel.java,v 6.0 2003/10/28 10:52:23 nisinaka Exp $
 */
public class JunTerrainEditModel extends JunOpenGL3dModel {
	protected Vector peaks = new Vector();
	protected Jun3dPoint[][] contours = null;
	protected double contourStep = Double.NaN;
	protected double contourGrid = Double.NaN;
	protected JunTerrainField terrainField = null;
	protected Hashtable pushButtons = new Hashtable();
	protected double zoomHeight = Double.NaN;
	protected Jun2dBoundingBox mapBounds = null;
	protected Jun2dPoint sight = null;

	/**
	 * Add a 3D point as a peak.
	 * 
	 * @param aJun3dPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint addPeak_(Jun3dPoint aJun3dPoint) {
		peaks.addElement(aJun3dPoint);
		this.flushContours();
		this.updateProjection();
		this.updateDisplayObject();

		return aJun3dPoint;
	}

	/**
	 * Answer my current button state.
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public StSymbol buttonState() {
		if (pushButtons.size() <= 3) {
			this.dragButton();
			this.peakButton();
			this.showButton();
		}

		Enumeration enum = pushButtons.keys();

		while (enum.hasMoreElements()) {
			StSymbol key = (StSymbol) enum.nextElement();
			JunButtonModel buttonModel = (JunButtonModel) pushButtons.get(key);

			if (buttonModel.value()) {
				return key;
			}
		}

		return null;
	}

	/**
	 * menu message: change my contour grid.
	 */
	public void changeContourGrid() {
		String numberString = StDialog.Request_("Please enter contour grid", String.valueOf(this.contourGrid()));

		if (numberString == null) {
			return;
		}

		try {
			double grid = Double.parseDouble(numberString);
			this.contourGrid_(grid);
		} catch (NumberFormatException e) {
			StDialog.Warn_("invalid input");
		}
	}

	/**
	 * menu message: change my contour step.
	 */
	public void changeContourStep() {
		String numberString = StDialog.Request_("Please enter contour step", String.valueOf(this.contourStep()));

		if (numberString == null) {
			return;
		}

		try {
			double step = Double.parseDouble(numberString);
			this.contourStep_(step);
		} catch (NumberFormatException e) {
			StDialog.Warn_("invalid input");
		}
	}

	/**
	 * Answer the contour grid.
	 * 
	 * @return double
	 */
	public double contourGrid() {
		if (Double.isNaN(contourGrid)) {
			contourGrid = this.defaultContourGrid();
		}

		return contourGrid;
	}

	/**
	 * Set the contour grid.
	 * 
	 * @param aNumber double
	 */
	public void contourGrid_(double aNumber) {
		contourGrid = aNumber;
		this.flushContours();
		this.changed();
	}

	/**
	 * Answer the array of contours.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint[][]
	 */
	public Jun3dPoint[][] contours() {
		if (contours == null) {
			contours = this.computeContours();
		}

		return contours;
	}

	/**
	 * Answer the contour step.
	 * 
	 * @return double
	 */
	public double contourStep() {
		if (Double.isNaN(contourStep)) {
			contourStep = this.defaultContourStep();
		}

		return contourStep;
	}

	/**
	 * Set the contour step.
	 * 
	 * @param aNumber double
	 */
	public void contourStep_(double aNumber) {
		contourStep = aNumber;
		this.flushContours();
		this.changed();
	}

	/**
	 * Answer the default value of 'contourGrid'.
	 * 
	 * @return double
	 */
	public double defaultContourGrid() {
		return Math.min(this.mapBounds().width(), this.mapBounds().height()) / 20.0d;
	}

	/**
	 * Answer the default value of 'contourStep'.
	 * 
	 * @return double
	 */
	public double defaultContourStep() {
		return 10.0d;
	}

	/**
	 * Answer the default value of 'mapBounds'.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 */
	public Jun2dBoundingBox defaultMapBounds() {
		return Jun2dBoundingBox.Origin_corner_(new Jun2dPoint(-1000, -1000), new Jun2dPoint(1000, 1000));
	}

	/**
	 * Answer the default sight point.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint defaultSight() {
		return this.mapBounds().center();
	}

	/**
	 * Answer my default up vector.
	 * 
	 * @return sra.jun.opengl.objects.Jun3dPoint
	 */
	public Jun3dPoint defaultUpVector() {
		return new Jun3dPoint(0, 1, 0);
	}

	/**
	 * Answer a default view.
	 * 
	 * @return sra.smalltalk.StView
	 */
	public StView defaultView() {
		if (GetDefaultViewMode() == VIEW_AWT) {
			return new JunTerrainEditViewAwt(this);
		} else {
			return new JunTerrainEditViewSwing(this);
		}
	}

	/**
	 * Answer my default view factor.
	 * 
	 * @return double
	 */
	public double defaultViewFactor() {
		return 10;
	}

	/**
	 * Answer the default value of 'zoomHeight'.
	 * 
	 * @return double
	 */
	public double defaultZoomHeight() {
		return this.mapBounds().height() * 1.1d;
	}

	/**
	 * Answer my display object.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject displayObject() {
		return this.openGL3dObject();
	}

	/**
	 * Answer my current projection.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjection
	 */
	public JunOpenGLProjection displayProjection() {
		return this.displayProjector().projection();
	}

	/**
	 * Answer my display projector.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjector
	 */
	public JunOpenGLProjector displayProjector() {
		return this.openGLProjector();
	}

	/**
	 * Answer my drag button.
	 * 
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	public JunButtonModel dragButton() {
		if (pushButtons.containsKey($("drag")) == false) {
			JunButtonModel button = new JunButtonModel();
			button.value_(false);
			button.visual_(JunCursors.QuartersCursorImage());
			button.action_(new StBlockClosure() {
				public Object value_(Object obj) {
					JunButtonModel model = (JunButtonModel) obj;
					model.value_(model.value() == false);

					if (model.value() == true) {
						peakButton().value_(false);
						showButton().value_(false);
					}

					changed_($("state"));

					return null;
				}
			});
			pushButtons.put($("drag"), button);
		}

		return (JunButtonModel) pushButtons.get($("drag"));
	}

	/**
	 * Answer the current eye point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint eyePoint() {
		return new Jun3dPoint(this.sight().x(), this.sight().y(), (this.topPeak() + 1) * 2);
	}

	/**
	 * Fit the 3d object to the current view.
	 */
	public void fitSilently() {
		// Avoid to fit when opening.
	}

	/**
	 * Answer the current value of 'mapBounds'.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dBoundingBox
	 */
	public Jun2dBoundingBox mapBounds() {
		if (mapBounds == null) {
			mapBounds = this.defaultMapBounds();
		}

		return mapBounds;
	}

	/**
	 * Answer my display object. Override the method of JunOpenGL3dModel for my
	 * own initialization.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	public JunOpenGL3dObject openGL3dObject() {
		if (super.openGL3dObject() == null) {
			this.computeDisplayObject();
		}

		return super.openGL3dObject();
	}

	/**
	 * Anser my OpenGL projector. Override the method of JunOpenGL3dModel for
	 * my own initialization.
	 * 
	 * @return sra.jun.opengl.projection.JunOpenGLProjector
	 */
	public JunOpenGLProjector openGLProjector() {
		if (openGLProjector == null) {
			openGLProjector = new JunOpenGLProjector();
			openGLProjector.parallelProjection();
			openGLProjector.sightPoint_(this.sightPoint());
			openGLProjector.eyePoint_(this.eyePoint());
			openGLProjector.upVector_(this.defaultUpVector());
			openGLProjector.viewFactor_(this.defaultViewFactor());
			openGLProjector.zoomHeight_(this.defaultZoomHeight());
		}

		return openGLProjector;
	}

	/**
	 * Answer my peak button.
	 * 
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	public JunButtonModel peakButton() {
		if (pushButtons.containsKey($("peak")) == false) {
			BufferedImage image = JunCursors.UpCursorImage();
			int width = image.getWidth();
			int height = image.getHeight();
			BufferedImage peakButtonImage = new BufferedImage(width, height, image.getType());
			peakButtonImage.setData(image.getData(new Rectangle(width - 3, height)).createTranslatedChild(3, 0));

			JunButtonModel button = new JunButtonModel();
			button.value_(false);
			button.visual_(peakButtonImage);
			button.action_(new StBlockClosure() {
				public Object value_(Object obj) {
					JunButtonModel model = (JunButtonModel) obj;
					model.value_(model.value() == false);

					if (model.value() == true) {
						dragButton().value_(false);
						showButton().value_(false);
					}

					changed_($("state"));

					return null;
				}
			});
			pushButtons.put($("peak"), button);
		}

		return (JunButtonModel) pushButtons.get($("peak"));
	}

	/**
	 * Answer the array of peaks.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint[]
	 */
	public Jun3dPoint[] peaks() {
		Jun3dPoint[] peaksArray = new Jun3dPoint[peaks.size()];
		peaks.copyInto(peaksArray);

		return peaksArray;
	}

	/**
	 * Answer my show button.
	 * 
	 * @return sra.jun.goodies.button.JunButtonModel
	 */
	public JunButtonModel showButton() {
		if (pushButtons.containsKey($("show")) == false) {
			BufferedImage image = JunCursors.ReadCursorImage();
			int width = image.getWidth();
			int height = image.getHeight();
			BufferedImage showButtonImage = new BufferedImage(width, height, image.getType());
			showButtonImage.setData(image.getData(new Rectangle(width - 3, height)).createTranslatedChild(3, 0));

			JunButtonModel button = new JunButtonModel();
			button.value_(false);
			button.visual_(showButtonImage);
			button.action_(new StBlockClosure() {
				public Object value_(Object obj) {
					JunButtonModel model = (JunButtonModel) obj;
					model.value_(model.value() == false);

					if (model.value() == true) {
						peakButton().value_(false);
						dragButton().value_(false);
					}

					changed_($("state"));

					return null;
				}
			});
			pushButtons.put($("show"), button);
		}

		return (JunButtonModel) pushButtons.get($("show"));
	}

	/**
	 * menu message: show the contour.
	 */
	public void showContour() {
		this.displayObject().show();
	}

	/**
	 * menu message: show the terrain field.
	 */
	public void showTerrainField() {
		JunOpenGL3dObject anObject = this.terrainField().asJunOpenGL3dObject();
		anObject.paint_(Color.green);
		anObject.show();
	}

	/**
	 * Answer the current sight point.
	 * 
	 * @return sra.jun.geometry.basic.Jun2dPoint
	 */
	public Jun2dPoint sight() {
		if (sight == null) {
			sight = this.defaultSight();
		}

		return sight;
	}

	/**
	 * Set my sight point.
	 * 
	 * @param aPoint sra.jun.geometry.basic.Jun2dPoint
	 */
	public void sight_(Jun2dPoint aPoint) {
		sight = aPoint;
		this.updateProjection();
	}

	/**
	 * menu message: align the sight to center.
	 */
	public void sightCenter() {
		this.sight_(this.mapBounds().center());
	}

	/**
	 * Answer the current sight point.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	public Jun3dPoint sightPoint() {
		return new Jun3dPoint(this.sight().x(), this.sight().y(), 0.0d);
	}

	/**
	 * Answer my terrain field.
	 * 
	 * @return sra.jun.terrain.support.JunTerrainField
	 */
	public JunTerrainField terrainField() {
		if (terrainField == null) {
			terrainField = this.computeTerrainField();
		}

		return terrainField;
	}

	/**
	 * Update the menu indications.
	 */
	public void updateMenuIndication() {
		// do nothing.
	}

	/**
	 * Zoom in/out.
	 * 
	 * @param factor double
	 */
	public void zoom_(double factor) {
		double zoomFactor = factor;

		if (false) { // InputState default shiftDown
			if (factor >= 1.0) {
				zoomFactor = 1.0 + ((factor - 1.0) * 0.1);
			} else {
				zoomFactor = 1.0 - ((1.0 - factor) * 0.1);
			}
		}

		this.zoomHeight_(this.zoomHeight() * zoomFactor);
	}

	/**
	 * Answer the current value of 'zoomHeight'.
	 * 
	 * @return double
	 */
	public double zoomHeight() {
		if (Double.isNaN(zoomHeight)) {
			zoomHeight = this.defaultZoomHeight();
		}

		return zoomHeight;
	}

	/**
	 * Set the new zoomHeight value.
	 * 
	 * @param newZoomHeight double
	 */
	public void zoomHeight_(double newZoomHeight) {
		zoomHeight = newZoomHeight;
		this.updateProjection();
	}

	/**
	 * menu message: zoom x1.
	 */
	public void zoomX1() {
		this.zoomHeight_(this.mapBounds().height() * 1.1d);
	}

	/**
	 * menu message: zoom x2.
	 */
	public void zoomX2() {
		this.zoomHeight_(this.mapBounds().height() / 2.0d * 1.1d);
	}

	/**
	 * menu message: zoom x4.
	 */
	public void zoomX4() {
		this.zoomHeight_(this.mapBounds().height() / 4.0d * 1.1d);
	}

	/**
	 * menu message: zoom x8.
	 */
	public void zoomX8() {
		this.zoomHeight_(this.mapBounds().height() / 8.0d * 1.1d);
	}

	/**
	 * Answer a zThumbWheel.
	 * 
	 * @return sra.jun.goodies.wheels.JunThumbWheel
	 */
	public JunThumbWheel zThumbWheel() {
		final JunTerrainEditModel self = this;
		JunThumbWheel thumbWheel = new JunThumbWheel();
		thumbWheel.compute_(new StBlockClosure() {
			public Object value_(Object value) {
				double factor = 1 + (((Number) value).doubleValue() / 100);
				self.zoom_(factor);

				return null;
			}
		});

		return thumbWheel;
	}

	/**
	 * Answer the basic 3D point which corresponds to the specified 2D point.
	 * 
	 * @param aJun2dPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	protected Jun3dPoint basic3dPointAt_(Jun2dPoint aJun2dPoint) {
		return new Jun3dPoint(aJun2dPoint.x(), aJun2dPoint.y(), this.basicHeightAt_(aJun2dPoint));
	}

	/**
	 * Answer the basic height at the specified 2D point.
	 * 
	 * @param aJun2dPoint sra.jun.geometry.basic.Jun2dPoint
	 * 
	 * @return double
	 */
	protected double basicHeightAt_(Jun2dPoint aJun2dPoint) {
		if (peaks.isEmpty()) {
			return 0.0d;
		}

		double maxDistance = this.mapBounds().width() + this.mapBounds().height();
		double distance1 = maxDistance;
		double distance2 = maxDistance;
		double height1 = Double.NaN;
		double height2 = Double.NaN;

		Enumeration enum = peaks.elements();

		while (enum.hasMoreElements()) {
			Jun3dPoint peak = (Jun3dPoint) enum.nextElement();
			double distance = aJun2dPoint.distance_(new Jun2dPoint(peak.x(), peak.y()));

			if (distance < distance1) {
				distance2 = distance1;
				height2 = height1;
				distance1 = distance;
				height1 = this.heightWithPeak_distance_(peak, distance);
			} else if (distance < distance2) {
				distance2 = distance;
				height2 = this.heightWithPeak_distance_(peak, distance);
			}
		}

		return (Double.isNaN(height2) ? height1 : (distance2 * height1 + distance1 * height2) / (distance1 + distance2));
	}

	/**
	 * Create the current contours.
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint[][]
	 */
	protected Jun3dPoint[][] computeContours() {
		final StBlockClosure evaluationBlock = new StBlockClosure() {
			public Object value_value_(Object objX, Object objY) {
				double x = ((Number) objX).doubleValue();
				double y = ((Number) objY).doubleValue();

				return new Double(basicHeightAt_(new Jun2dPoint(x, y)));
			}
		};

		final Vector contourStream = new Vector();
		JunCursorAnimator.ClockCursors().showWhile_(new StBlockClosure() {
			public Object value() {
				final Jun2dPoint mapOrigin = mapBounds().origin();
				final Jun2dPoint mapCorner = mapBounds().corner();
				double topPeak = topPeak();
				double contourStep = contourStep();

				for (double z = 0.0d; z <= topPeak; z += contourStep) {
					double contourGrid = contourGrid();
					Jun2dPoint[][] polylines =
						JunBorderGenerator.ContourPolylinesFrom_at_xInterval_yInterval_interim_(
							evaluationBlock,
							z,
							new StInterval(mapOrigin.x(), mapCorner.x() + (contourGrid / 2.0d), contourGrid),
							new StInterval(mapOrigin.y(), mapCorner.y() + (contourGrid / 2.0d), contourGrid),
							null);

					for (int i = 0; i < polylines.length; i++) {
						Jun3dPoint[] points = new Jun3dPoint[polylines[i].length];

						for (int j = 0; j < points.length; j++) {
							Jun2dPoint point = polylines[i][j];
							points[j] = new Jun3dPoint(point.x(), point.y(), z);
						}

						contourStream.addElement(points);
					}
				}

				return null;
			}
		});

		Jun3dPoint[][] contourPoints = new Jun3dPoint[contourStream.size()][];
		contourStream.copyInto(contourPoints);

		return contourPoints;
	}

	/**
	 * Create the current display object.
	 */
	protected void computeDisplayObject() {
		Vector objects = new Vector();
		objects.addElement(this.mapBorderLine());

		Jun3dPoint[] peaks = this.peaks();

		for (int i = 0; i < peaks.length; i++) {
			objects.addElement(this.peakMarkAt_(peaks[i]));
		}

		Jun3dPoint[][] contours = this.contours();

		for (int i = 0; i < contours.length; i++) {
			objects.addElement(new JunOpenGL3dPolyline(contours[i], Color.red));
		}

		this.openGL3dObject_(JunOpenGL3dCompoundObject.Components_(objects));
	}

	/**
	 * Create my terrain field.
	 * 
	 * @return sra.jun.terrain.support.JunTerrainField
	 */
	protected JunTerrainField computeTerrainField() {
		Jun2dDelaunayProcessor delaunayProcessor = new Jun2dDelaunayProcessor(this.mapBounds().expandedBy_(this.mapBounds().extent().dividedBy_(4.0d)));
		delaunayProcessor.insertPoint_(this.basic3dPointAt_(new Jun2dPoint(this.mapBounds().origin().x(), this.mapBounds().origin().x())));
		delaunayProcessor.insertPoint_(this.basic3dPointAt_(new Jun2dPoint(this.mapBounds().corner().x(), this.mapBounds().origin().x())));
		delaunayProcessor.insertPoint_(this.basic3dPointAt_(new Jun2dPoint(this.mapBounds().corner().x(), this.mapBounds().corner().x())));
		delaunayProcessor.insertPoint_(this.basic3dPointAt_(new Jun2dPoint(this.mapBounds().origin().x(), this.mapBounds().corner().x())));

		Jun3dPoint[][] contours = this.contours();

		for (int i = 0; i < contours.length; i++) {
			for (int j = 0; j < contours[i].length; j++) {
				delaunayProcessor.insertPoint_(contours[i][j]);
			}
		}

		final Vector triangles = new Vector();
		delaunayProcessor.convexHullTrianglesDo_(new StBlockClosure() {
			public Object value_value_value_(Object p1, Object p2, Object p3) {
				triangles.addElement(new JunTerrainTriangle((Jun3dPoint) p1, (Jun3dPoint) p2, (Jun3dPoint) p3));

				return null;
			}
		});

		JunTerrainTriangle[] anArrayOfJunTerrainTriangle = new JunTerrainTriangle[triangles.size()];
		triangles.copyInto(anArrayOfJunTerrainTriangle);

		return new JunTerrainField(anArrayOfJunTerrainTriangle);
	}

	/**
	 * Answer the default menu model.
	 * 
	 * @return sra.smalltalk.StMenuModel
	 */
	protected StMenuModel defaultMenu() {
		StMenuModel rootMenu = new StMenuModel();

		//
		StMenu contoursMenu = new StMenu("Contours");
		contoursMenu.add(createMenuItem("Grid...", "changeContourGrid"));
		contoursMenu.add(createMenuItem("Step...", "changeContourStep"));
		rootMenu.add(contoursMenu);

		//
		StMenu viewMenu = new StMenu("View");
		viewMenu.add(createMenuItem("Center", "sightCenter"));
		viewMenu.addSeparator();
		viewMenu.add(createMenuItem("x 1", "zoomX1"));
		viewMenu.add(createMenuItem("x 2", "zoomX2"));
		viewMenu.add(createMenuItem("x 4", "zoomX4"));
		viewMenu.add(createMenuItem("x 8", "zoomX8"));
		rootMenu.add(viewMenu);

		//
		StMenu miscMenu = new StMenu("Misc");
		miscMenu.add(createMenuItem("Show", "showContour"));
		miscMenu.add(createMenuItem("Show TerrainField", "showTerrainField"));
		rootMenu.add(miscMenu);

		//
		return rootMenu;
	}

	/**
	 * Called when a mouse is dragged.
	 * 
	 * @param from2dPoint sra.jun.geometry.basic.Jun2dPoint
	 * @param to2dPoint sra.jun.geometry.basic.Jun2dPoint
	 */
	protected void drag_xy_(Jun2dPoint from2dPoint, Jun2dPoint to2dPoint) {
		if (this.displayObject() == null) {
			return;
		}

		JunOpenGLProjection aProjection = this.displayProjection();
		Jun3dPoint fromPoint = aProjection.translateTo3dPointFromPoint_(from2dPoint);
		Jun3dPoint toPoint = aProjection.translateTo3dPointFromPoint_(to2dPoint);
		Jun3dPoint vector = fromPoint.minus_(toPoint);
		this.sight_(new Jun2dPoint(this.sight().x() + vector.x(), this.sight().y() + vector.y()));
	}

	/**
	 * Flush the contours information.
	 */
	protected void flushContours() {
		if (contours != null) {
			contours = this.computeContours();
		}

		this.flushTerrainField();
		this.flushDisplayObject();
	}

	/**
	 * Flush the current display object.
	 */
	protected void flushDisplayObject() {
		super.flushObject();
	}

	/**
	 * Flush the current terrain field.
	 */
	protected void flushTerrainField() {
		terrainField = null;
	}

	/**
	 * Answer the height calculated with the specified peak and distance.
	 * 
	 * @param aJun3dPoint sra.jun.geometry.basic.Jun3dPoint
	 * @param aNumber double
	 * 
	 * @return double
	 */
	protected double heightWithPeak_distance_(Jun3dPoint aJun3dPoint, double aNumber) {
		return Math.max(aJun3dPoint.z() - (aNumber / 3.0d), 0.0d);
	}

	/**
	 * Create a map border line.
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dPolylineLoop
	 */
	protected JunOpenGL3dPolylineLoop mapBorderLine() {
		Jun2dBoundingBox bounds = this.mapBounds();
		Jun2dPoint mapOrigin = bounds.origin();
		Jun2dPoint mapCorner = bounds.corner();

		return new JunOpenGL3dPolylineLoop(
			new Jun3dPoint[] { new Jun3dPoint(mapOrigin.x(), mapOrigin.y(), 0.0d), new Jun3dPoint(mapCorner.x(), mapOrigin.y(), 0.0d), new Jun3dPoint(mapCorner.x(), mapCorner.y(), 0.0d), new Jun3dPoint(mapOrigin.x(), mapCorner.y(), 0.0d)},
			Color.black);
	}

	/**
	 * Convert the mouse point to a Jun3dPoint with the current projection.
	 * 
	 * @param aController sra.jun.terrain.editor.JunTerrainEditController
	 * 
	 * @return sra.jun.geometry.basic.Jun3dPoint
	 */
	protected Jun3dPoint mousePointFrom_(JunTerrainEditController aController) {
		return this.displayProjection().translateTo3dPointFromPoint_(aController.mouse2dPoint());
	}

	/**
	 * Create a new peak mark at the specified point.
	 * 
	 * @param aJun3dPoint sra.jun.geometry.basic.Jun3dPoint
	 * 
	 * @return sra.jun.opengl.objects.JunOpenGL3dObject
	 */
	protected JunOpenGL3dObject peakMarkAt_(Jun3dPoint aJun3dPoint) {
		double scale = this.mapBounds.height() / 20.0d;
		JunOpenGL3dObject peakMark = new JunOpenGL3dPolylineLoop(new Jun3dPoint[] { new Jun3dPoint(0.0d, 1.0d, 0.0d), new Jun3dPoint(-0.5d, 0.0d, 0.0d), new Jun3dPoint(0.5d, 0.0d, 0.0d)}, Color.blue);
		peakMark = peakMark.scaledBy_(scale);

		return peakMark.translatedBy_(aJun3dPoint);
	}

	/**
	 * Plot a peak with the information of the controller.
	 * 
	 * @param aController sra.jun.terrain.editor.JunTerrainEditController
	 */
	protected void plotPeak_(JunTerrainEditController aController) {
		String numberString = StDialog.Request_("Please enter height", "100.0");

		if (numberString == null) {
			return;
		}

		try {
			double height = Double.parseDouble(numberString);
			Jun3dPoint point = this.mousePointFrom_(aController);
			this.addPeak_(new Jun3dPoint(point.x(), point.y(), height));
		} catch (NumberFormatException e) {
			StDialog.Warn_("invalid height");
		}
	}

	/**
	 * Show the terrain.
	 * 
	 * @param aController sra.jun.terrain.editor.JunTerrainEditController
	 * 
	 * @return sra.jun.terrain.display.JunTerrainDisplayModel
	 */
	protected JunTerrainDisplayModel showTerrain_(JunTerrainEditController aController) {
		JunAngle zoomAngle = JunAngle.FromDeg_(45);
		Jun2dLine sightLineSegment = aController.mouse2dLineSegment();

		if (sightLineSegment == null) {
			return null;
		}

		Jun3dPoint eyePoint = this.displayProjection().translateTo3dPointFromPoint_(sightLineSegment.from());
		Jun3dPoint sightPoint = this.displayProjection().translateTo3dPointFromPoint_(sightLineSegment.to());

		if (eyePoint.distance_(sightPoint) < (1.0d - 12)) {
			return null;
		}

		JunTerrainDisplayModel displayModel = new JunTerrainDisplayModel(this.terrainField());
		displayModel.defaultZoomHeight_(eyePoint.minus_(sightPoint).length() * zoomAngle.tan());
		displayModel.defaultUpVector_(new Jun3dPoint(0, 0, 1));
		displayModel.moveTo_sight_(new Jun2dPoint(eyePoint.x(), eyePoint.y()), new Jun2dPoint(sightPoint.x(), sightPoint.y()));

		JunOpenGLProjection projection = displayModel.displayProjection();
		displayModel.defaultSightPoint_(projection.sightPoint());
		displayModel.defaultEyePoint_(projection.eyePoint());
		displayModel.open();

		return displayModel;
	}

	/**
	 * Answer the value of the top peak.
	 * 
	 * @return double
	 */
	protected double topPeak() {
		double top = 0.0d;
		Jun3dPoint[] peaks = this.peaks();

		for (int i = 0; i < peaks.length; i++) {
			Jun3dPoint peak = peaks[i];

			if (peak.z() > top) {
				top = peak.z();
			}
		}

		return top;
	}

	/**
	 * Update the current display object.
	 */
	protected void updateDisplayObject() {
		this.flushDisplayObject();
		this.changed_($("object"));
	}

	/**
	 * Update the projection.
	 */
	protected void updateProjection() {
		this.openGLProjector().sightPoint_(this.sightPoint());
		this.openGLProjector().eyePoint_(this.eyePoint());
		this.openGLProjector().upVector_(this.defaultUpVector());
		this.openGLProjector().viewFactor_(this.defaultViewFactor());
		this.openGLProjector().zoomHeight_(this.zoomHeight());
		this.changed_($("projection"));
	}

	/**
	 * Answer the title string.
	 * 
	 * @return java.lang.String
	 */
	protected String windowTitle() {
		return "Terrain Editor";
	}
}
