package sra.smalltalk;

import java.awt.*;
import java.awt.font.*;
import java.io.*;
import java.text.AttributedString;
import java.util.Vector;

/**
 * StComposedText class
 * 
 *  @author    nisinaka
 *  @created   2002/11/20 (by nisinaka)
 *  @updated   N/A
 *  @version   6.0
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StComposedText.java,v 6.0 2003/10/28 10:52:32 nisinaka Exp $
 */
public class StComposedText extends StObject implements StDisplayable {

	protected String string;
	protected Font font;
	protected int compositionWidth;
	protected int compositionHeight;
	protected boolean fitWidth;
	protected TextLayout[] layouts;

	public final static int DefaultCompositionWidth = Integer.MAX_VALUE;

	/**
	 * Create a new instance of StComposedText with a string and the default font.
	 *
	 * @param aString java.lang.String
	 */
	public StComposedText(String aString) {
		this(aString, null);
	}

	/**
	 * Create a new instance of StComposedText and initialize it.
	 *
	 * @param aString java.lang.String
	 * @param width int
	 */
	public StComposedText(String aString, int width) {
		this();
		this.compositionWidth_string_font_fitWidth_(width, aString, null, false);
	}

	/**
	 * Create a new instance of StComposedText and initialize it.
	 *
	 * @param aString java.lang.String
	 * @param aFont java.awt.Font
	 */
	public StComposedText(String aString, Font aFont) {
		this();
		this.compositionWidth_string_font_fitWidth_(DefaultCompositionWidth, aString, aFont, true);
	}

	/**
	 * Create a new instance of StComposedText and initialize it.
	 *
	 * @param aString java.lang.String
	 * @param aFont java.awt.Font
	 * @param width int
	 */
	public StComposedText(String aString, Font aFont, int width) {
		this();
		this.compositionWidth_string_font_fitWidth_(width, aString, aFont, false);
	}

	/**
	 * Create a new instance of StComposedText and initialize it.
	 */
	protected StComposedText() {
		string = null;
		font = null;
		compositionWidth = -1;
		compositionHeight = -1;
		fitWidth = true;
		layouts = null;
	}

	/**
	 * Answer the default font.
	 *
	 * @return java.awt.Font
	 */
	public static Font DefaultFont() {
		return SystemResourceSupport.getFont();
	}

	/**
	 * Answer the default FontRenderContext.
	 *
	 * @return java.awt.font.FontRenderContext
	 */
	public static FontRenderContext DefaultFontRenderContext() {
		return new FontRenderContext(null, false, false);
	}

	/**
	 * Convert to a StImage.
	 *
	 * @return java.awt.Image
	 */
	public StImage asImage() {
		int width = this.width();
		int height = this.height();
		StImage anImage = new StImage(width, height);
		Graphics2D aGraphics = null;
		try {
			aGraphics = (Graphics2D) anImage.image().getGraphics();
			aGraphics.setColor(Color.white);
			aGraphics.fillRect(0, 0, width, height);
			aGraphics.setColor(Color.black);
			this.displayOn_(aGraphics);
		} finally {
			if (aGraphics != null) {
				aGraphics.dispose();
				aGraphics = null;
			}
		}
		return anImage;
	}

	/**
	 * Answer the bounding box of the receiver.
	 *
	 * @return java.awt.Rectangle
	 */
	public Rectangle bounds() {
		return new Rectangle(0, 0, this.width(), this.height());
	}

	/**
	 * Display the receiver on aGraphics.
	 * 
	 * @param aGraphics java.awt.Graphics
	 */
	public void displayOn_(Graphics aGraphics) {
		this.displayOn_at_(aGraphics, new Point(0, 0));
	}

	/**
	 * Display the receiver on the graphics at the specified point.
	 *
	 * @param aGraphics java.awt.Graphics
	 * @param aPoint java.awt.Point
	 */
	public void displayOn_at_(Graphics aGraphics, Point aPoint) {
		Graphics2D aGraphics2D = (Graphics2D) aGraphics;
		int x = aPoint.x;
		int y = aPoint.y;
		TextLayout[] layouts = this.layouts();
		for (int i = 0; i < layouts.length; i++) {
			y += layouts[i].getAscent();
			layouts[i].draw(aGraphics2D, x, y);
			y += layouts[i].getDescent() + layouts[i].getLeading();
		}
	}

	/**
	 * Answer the extent around the receiver.
	 *
	 * @return java.awt.Dimension
	 */
	public Dimension extent() {
		return new Dimension(this.width(), this.height());
	}

	/**
	 * Answer the current font.
	 *
	 * @return java.awt.Font
	 */
	public Font font() {
		if (font == null) {
			return DefaultFont();
		}
		return font;
	}

	/**
	 * Set the new font.
	 *
	 * @param newFont java.awt.Font
	 */
	public void font_(Font newFont) {
		Font current = this.font();
		font = newFont;
		if (current.equals(this.font()) == false) {
			this.composeAll();
		}
	}

	/**
	 * Ansewr the height of the receiver's bounding box.
	 *
	 * @return int
	 */
	public int height() {
		if (compositionHeight < 0) {
			this.updateCompositionHeight();
		}
		return compositionHeight;
	}

	/**
	 * Answer true if the receiver is composed, otherwise false.
	 *
	 * @return boolean
	 */
	public boolean isComposed() {
		return (layouts != null);
	}

	/**
	 * Answer the number of lines of text in the receiver.
	 *
	 * @return int
	 */
	public int numberOfLines() {
		return this.layouts().length;
	}

	/**
	 * Answer the current string.
	 *
	 * @return java.lang.String
	 */
	public String string() {
		return string;
	}

	/**
	 * Set the new string.
	 *
	 * @param newString java.lang.String
	 */
	public void string_(String newString) {
		string = newString;
		this.composeAll();
	}

	/**
	 * Convert to an image.
	 *
	 * @return java.awt.Image
	 */
	public Image toImage() {
		return this.asImage().image();
	}

	/**
	 * Ansewr the width of the receiver's bounding box.
	 *
	 * @return int
	 */
	public int width() {
		return compositionWidth;
	}

	/**
	 * Compose the text from the current string and the font.
	 *
	 * @return int
	 */
	protected int composeAll() {
		if (this.string().length() == 0) {
			layouts = new TextLayout[0];
			this.setHeight_(0);
			return 0;
		}

		float formatWidth = this.width();
		Vector layoutCollection = new Vector();
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new StringReader(this.string()));
			String line;
			while ((line = reader.readLine()) != null) {
				int length = line.length();
				if (length == 0) {
					// AttributedString does not allow to add an attribute to a 0-length string.
					// Don't worry about the width.  TextLayout#getVisibleAdvance() will return 0.0.
					line = " ";
					length = 1;
				}
				AttributedString anAttributedString = new AttributedString(line);
				anAttributedString.addAttribute(TextAttribute.FONT, this.font());
				LineBreakMeasurer aLineBreakMeasurer = new LineBreakMeasurer(anAttributedString.getIterator(), DefaultFontRenderContext());
				aLineBreakMeasurer.setPosition(0);
				while (aLineBreakMeasurer.getPosition() < length) {
					layoutCollection.add(aLineBreakMeasurer.nextLayout(formatWidth));
				}
			}
		} catch (IOException e) {
			this.setHeight_(0);
			return 0;
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					this.setHeight_(0);
					return 0;
				}
			}
		}

		layouts = new TextLayout[layoutCollection.size()];
		layoutCollection.copyInto(layouts);
		float maximumRightX = layouts[0].getAdvance();
		float height = layouts[0].getAscent() + layouts[0].getDescent();
		for (int i = 1; i < layouts.length; i++) {
			maximumRightX = Math.max(maximumRightX, layouts[i].getVisibleAdvance());
			height += layouts[i].getLeading() + layouts[i].getAscent() + layouts[i].getDescent();
		}
		this.setHeight_((int) height);

		return (int) maximumRightX;
	}

	/**
	 * Initialize the receiver.
	 *
	 * @param newWidth int
	 * @param newString java.lang.String
	 * @param newFont java.awt.Font
	 * @param aBoolean boolean
	 */
	protected void compositionWidth_string_font_fitWidth_(int newWidth, String newString, Font newFont, boolean aBoolean) {
		fitWidth = aBoolean;
		string = newString;
		font = newFont;
		this.setCompositionWidth_(newWidth);

		if (aBoolean) {
			compositionWidth = this.composeAll();
		} else {
			this.composeAll();
		}
	}

	/**
	 * Answer the layouts.  Initialize it if necessary.
	 *
	 * @return java.awt.font.TextLayout[]
	 */
	protected TextLayout[] layouts() {
		if (layouts == null) {
			this.composeAll();
		}
		return layouts;
	}

	/**
	 * Set the width for composition.
	 * Make sure the composition width in not negative.
	 *
	 * @param newCompositionWidth int
	 */
	protected void setCompositionWidth_(int newCompositionWidth) {
		compositionWidth = Math.max(newCompositionWidth, 0);
	}

	/**
	 * Set the height of the receiver's bounding box.
	 *
	 * @param newHeight int
	 */
	protected void setHeight_(int newHeight) {
		compositionHeight = newHeight;
	}

	/**
	 * Update the height of the receiver's bounding box.
	 */
	protected void updateCompositionHeight() {

	}
}
