package sra.smalltalk;

import java.awt.*;
import java.io.*;

/**
 * StRectangle class
 * 
 *  @author    nisinaka
 *  @created   2002/10/28 (by nisinaka)
 *  @updated   N/A
 *  @version   6.0
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StRectangle.java,v 6.0 2003/10/28 10:52:32 nisinaka Exp $
 */
public class StRectangle extends StObject {

	protected int originX;
	protected int originY;
	protected int cornerX;
	protected int cornerY;

	/**
	 * Create a new instance of StRectangle.
	 * 
	 * @param newOriginX int
	 * @param newOriginY int
	 * @param newCornerX int
	 * @param newCornerY int
	 */
	public StRectangle(int newOriginX, int newOriginY, int newCornerX, int newCornerY) {
		originX = newOriginX;
		originY = newOriginY;
		cornerX = newCornerX;
		cornerY = newCornerY;
	}

	/**
	 * Create a new instance of StRectangle.
	 * 
	 * @param aDimension java.awt.Dimension
	 */
	public StRectangle(Dimension aDimension) {
		this(0, 0, aDimension.width, aDimension.height);
	}

	/**
	 * Create a new instance of StRectangle.
	 * 
	 * @param aRectangle java.awt.Rectangle
	 */
	public StRectangle(Rectangle aRectangle) {
		this(aRectangle.x, aRectangle.y, aRectangle.x + aRectangle.width, aRectangle.y + aRectangle.height);
	}

	/**
	 * Create a new instance of StRectangle with the origin and the corner.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param newOrigin java.awt.Point
	 * @param newCorner java.awt.Point
	 */
	public static StRectangle Origin_corner_(Point newOrigin, Point newCorner) {
		return new StRectangle(newOrigin.x, newOrigin.y, newCorner.x, newCorner.y);
	}

	/**
	 * Create a new instance of StRectangle with the origin and the extent.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param originPoint java.awt.Point
	 * @param extentPoint java.awt.Point
	 */
	public static StRectangle Origin_extent_(Point originPoint, Point extentPoint) {
		return new StRectangle(originPoint.x, originPoint.y, originPoint.x + extentPoint.x, originPoint.y + extentPoint.y);
	}

	/**
	 * Create a new instance of StRectangle with two points.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param point1 java.awt.Point
	 * @param point2 java.awt.Point
	 */
	public static StRectangle Vertex_vertex_(Point point1, Point point2) {
		return new StRectangle(Math.min(point1.x, point2.x), Math.min(point1.y, point2.y), Math.max(point1.x, point2.x), Math.max(point1.y, point2.y));
	}

	/**
	 * Answer a new StRectangle that is a translated by aPoint2 - aPoint1.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param aPoint1 java.awt.Point
	 * @param aPoint2 java.awt.Point
	 */
	public StRectangle align_with_(Point aPoint1, Point aPoint2) {
		return this.translatedBy_(aPoint2.x - aPoint1.x, aPoint2.y - aPoint1.y);
	}

	/**
	 * Answer my current bottom horizontal line.
	 * 
	 * @return int
	 */
	public int bottom() {
		return cornerY;
	}

	/**
	 * Answer the point at the center of the bottom horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point bottomCenter() {
		return new Point((originX + cornerX) / 2, cornerY);
	}

	/**
	 * Answer the point at the left edge of the bottom horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point bottomLeft() {
		return new Point(originX, cornerY);
	}

	/**
	 * Answer the point at the right edge of the bottom horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point bottomRight() {
		return this.corner();
	}

	/**
	 * Answer my center point.
	 * 
	 * @return java.awt.Point
	 */
	public Point center() {
		return new Point((originX + cornerX) / 2, (originY + cornerY) / 2);
	}

	/**
	 * Answer true if the point is contained in the rectangle, otherwise false.
	 * 
	 * @return boolean
	 * @param aPoint java.awt.Point
	 */
	public boolean containsPoint_(Point aPoint) {
		return (originX <= aPoint.x) && (originY <= aPoint.y) && (aPoint.x <= cornerX) && (aPoint.y <= cornerY);
	}

	/**
	 * Answer the point at the right edge of the bottom horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point corner() {
		return new Point(cornerX, cornerY);
	}

	/**
	 * Answer my current x-coordinate of the corner point.
	 * 
	 * @return int
	 */
	public int cornerX() {
		return cornerX;
	}

	/**
	 * Answer my current y-coordinate of the corner point.
	 * 
	 * @return int
	 */
	public int cornerY() {
		return cornerY;
	}

	/**
	 * Answer an StRectangle that is outset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta int
	 */
	public StRectangle expandedBy_(int delta) {
		return new StRectangle(originX - delta, originY - delta, cornerX + delta, cornerY + delta);
	}

	/**
	 * Answer an StRectangle that is outset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param deltaX int
	 * @param deltaY int
	 */
	public StRectangle expandedBy_(int deltaX, int deltaY) {
		return new StRectangle(originX - deltaX, originY - deltaY, cornerX + deltaX, cornerY + deltaY);
	}

	/**
	 * Answer an StRectangle that is outset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta java.awt.Point
	 */
	public StRectangle expandedBy_(Point delta) {
		return new StRectangle(originX - delta.x, originY - delta.y, cornerX + delta.x, cornerY + delta.y);
	}

	/**
	 * Answer an StRectangle that is outset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta sra.smalltalk.StRectangle
	 */
	public StRectangle expandedBy_(StRectangle delta) {
		return new StRectangle(originX - delta.originX, originY - delta.originY, cornerX + delta.cornerX, cornerY + delta.cornerY);
	}

	/**
	 * Answer my height.
	 * 
	 * @return double
	 */
	public int height() {
		return cornerY - originY;
	}

	/**
	 * Answer an StRectangle that is inset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta int
	 */
	public StRectangle insetBy_(int delta) {
		return new StRectangle(originX + delta, originY + delta, cornerX - delta, cornerY - delta);
	}

	/**
	 * Answer an StRectangle that is inset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta java.awt.Point
	 */
	public StRectangle insetBy_(Point delta) {
		return new StRectangle(originX + delta.x, originY + delta.y, cornerX - delta.x, cornerY - delta.y);
	}

	/**
	 * Answer an StRectangle that is inset from the receiver by delta.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param delta sra.smalltalk.StRectangle
	 */
	public StRectangle insetBy_(StRectangle delta) {
		return new StRectangle(originX + delta.originX, originY + delta.originY, cornerX - delta.cornerX, cornerY - delta.cornerY);
	}

	/**
	 * Answer the intersect with the rectangle.
	 * 
	 * @return sra.smalltalk.StRectangle
	 * @param aRectangle sra.smalltalk.StRectangle
	 */
	public StRectangle intersect_(StRectangle aRectangle) {
		return new StRectangle(Math.max(originX, aRectangle.originX), Math.max(originY, aRectangle.originY), Math.min(cornerX, aRectangle.cornerX), Math.min(cornerY, aRectangle.cornerY));
	}

	/**
	 * Answer whether aRectangle intersects the receiver anywhere.
	 * 
	 * @return boolean
	 * @param aRectangle sra.smalltalk.StRectangle
	 */
	public boolean intersects_(StRectangle aRectangle) {
		return (this.originX() < aRectangle.cornerX() && this.originY() < aRectangle.cornerY())
			&& (aRectangle.originX() < this.cornerX() && aRectangle.originY() < this.cornerY())
			&& (this.originX() < aRectangle.cornerX() && this.originY() < aRectangle.cornerY())
			&& (aRectangle.originX() < aRectangle.cornerX() && aRectangle.originY() < aRectangle.cornerY());
	}

	/**
	 * Answer my current left vertical line.
	 * 
	 * @return int
	 */
	public int left() {
		return originX;
	}

	/**
	 * Answer the point at the center of the receiver's left vertical line.
	 * 
	 * @return java.awt.Point
	 */
	public Point leftCenter() {
		return new Point(this.left(), this.center().y);
	}

	/**
	 * Answer a new StRectangle that contains both the receiver and the argument rectangle.
	 * 
	 * @return StRectangle
	 * @param aRectangle
	 */
	public StRectangle merge_(StRectangle aRectangle) {
		if (aRectangle != null) {
			return new StRectangle(Math.min(originX, aRectangle.originX), Math.min(originY, aRectangle.originY), Math.max(cornerX, aRectangle.cornerX), Math.max(cornerY, aRectangle.cornerY));
		} else {
			return new StRectangle(this.left(), this.top(), this.right(), this.bottom());
		}
	}

	/**
	 * Answer the point at the left edge of the top horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point origin() {
		return new Point(originX, originY);
	}

	/**
	 * Set my new origin and corner.
	 * 
	 * @param newOrigin java.awt.Point
	 * @param newCorner java.awt.Point
	 */
	public void origin_corner_(Point newOrigin, Point newCorner) {
		originX = newOrigin.x;
		originY = newOrigin.y;
		cornerX = newCorner.x;
		cornerY = newCorner.y;
	}

	/**
	 * Answer my current x-coordinate of the origin point.
	 * 
	 * @return int
	 */
	public int originX() {
		return originX;
	}

	/**
	 * Answer my current y-coordinate of the origin point.
	 * 
	 * @return int
	 */
	public int originY() {
		return originY;
	}

	/**
	 * Print my string representation on aWriter.
	 * 
	 * @param aWriter java.io.Writer
	 * 
	 * @throws IOException 
	 */
	public void printOn_(Writer aWriter) throws IOException {
		aWriter.write(String.valueOf(originX));
		aWriter.write('@');
		aWriter.write(String.valueOf(originY));
		aWriter.write(" corner: ");
		aWriter.write(String.valueOf(cornerX));
		aWriter.write('@');
		aWriter.write(String.valueOf(cornerY));
	}

	/**
	 * Answer my current right vertical line.
	 * 
	 * @return int
	 */
	public int right() {
		return cornerX;
	}

	/**
	 * Answer the point at the center of the receiver's right vertical line.
	 * 
	 * @return Point
	 */
	public Point rightCenter() {
		return new Point(this.right(), this.center().y);
	}

	/**
	 * Answer my current top horizontal line.
	 * 
	 * @return int
	 */
	public int top() {
		return originY;
	}

	/**
	 * Answer the point at the center of the top horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point topCenter() {
		return new Point((originX + cornerX) / 2, originY);
	}

	/**
	 * Answer the point at the left edge of the top horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point topLeft() {
		return this.origin();
	}

	/**
	 * Answer the point at the right edge of the top horizontal line of the receiver.
	 * 
	 * @return java.awt.Point
	 */
	public Point topRight() {
		return new Point(cornerX, originY);
	}

	/**
	 * Convert to a Rectangle.
	 * 
	 * @return java.awt.Rectangle
	 */
	public Rectangle toRectangle() {
		return new Rectangle(originX, originY, cornerX - originX, cornerY - originY);
	}

	/**
	 * Answer a new StRectangle translated by delta.
	 * 
	 * @param deltaX
	 * @param deltaY
	 * 
	 * @return StRectangle
	 */
	public StRectangle translatedBy_(int deltaX, int deltaY) {
		return new StRectangle(originX + deltaX, originY + deltaY, cornerX + deltaX, cornerY + deltaY);
	}

	/**
	 * Answer my width.
	 * 
	 * @return double
	 */
	public int width() {
		return cornerX - originX;
	}

	/**
	 * Answer my current x-coordinate of the origin point.
	 * 
	 * @return int
	 */
	public int x() {
		return this.originX();
	}

	/**
	 * Answer my current y-coordinate of the origin point.
	 * 
	 * @return int
	 */
	public int y() {
		return this.originY();
	}
}
