package sra.smalltalk;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * StUIBuilder class
 * 
 *  @author    He Weijie
 *  @created   1998/08/27 (by He Weijie)
 *  @updated   2003/03/04 (by nisinaka)
 *  @version   6.0
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StUIBuilder.java,v 6.0 2003/10/28 10:52:32 nisinaka Exp $
 */
public class StUIBuilder extends StObject implements WindowListener {

	protected StApplicationModel model;
	protected Vector windows;
	protected Hashtable views;
	protected Hashtable namedComponents;

	/**
	 * Create a new instance of StUIBuilder and initialize it with the StApplicationModel.
	 * 
	 * @param aModel sra.smalltalk.StApplicationModel
	 * @category Instance creation
	 */
	public StUIBuilder(StApplicationModel aModel) {
		super();

		model = aModel;
		windows = new Vector();
		views = new Hashtable();
		namedComponents = new Hashtable();
	}

	/**
	 * Answer a window.
	 * 
	 * @return java.awt.Window
	 * @category accessing
	 */
	public Window window() {
		return windows.isEmpty() ? null : (Window) windows.get(0);
	}

	/**
	 * Answer the current windows.
	 * 
	 * @return java.awt.Window[]
	 * @category accessing
	 */
	public synchronized Window[] windows() {
		Window[] anArray = new Window[windows.size()];
		windows.toArray(anArray);
		return anArray;
	}

	/**
	 * Answer the current frames, not windows.
	 * 
	 * @return java.awt.Frame[]
	 * @category accessing
	 */
	public Frame[] frames() {
		Window[] myWindows = this.windows();
		Vector frames = new Vector(myWindows.length);
		for (int i = 0; i < myWindows.length; i++) {
			if (myWindows[i] instanceof Frame) {
				frames.add(myWindows[i]);
			}
		}
		Frame[] myFrames = new Frame[frames.size()];
		frames.copyInto(myFrames);
		return myFrames;
	}

	/**
	 * Answer my StApplicationModel.
	 * 
	 * @return sra.smalltalk.StApplicationModel
	 * @category accessing
	 */
	protected StApplicationModel myModel() {
		return model;
	}

	/**
	 * Add a frame.
	 * 
	 * @param aFrame java.awt.Frame
	 * @deprecated since Jun379, use addWindow_(Window)
	 * @category adding
	 */
	public void addFrame_(Frame aFrame) {
		this.addWindow_(aFrame, null);
	}

	/**
	 * Add the window as a member of the builder.
	 *
	 * @param aWindow java.awt.Window
	 * @param aView sra.smalltalk.StView
	 * @category adding
	 */
	public void addWindow_(Window aWindow, StView aView) {
		aWindow.addWindowListener(this);
		windows.add(aWindow);
		views.put(aWindow, aView);
	}

	/**
	 * Remove the frame from the members of the builder.
	 * 
	 * @param aFrame java.awt.Frame
	 * @deprecated since Jun379, use removeWindow_(Window)
	 * @category removing
	 */
	public void removeFrame_(Frame aFrame) {
		this.removeWindow_(aFrame);
	}

	/**
	 * Remove the window from the members of the builder.
	 *
	 * @param aWindow java.awt.Window
	 * @category removing
	 */
	public void removeWindow_(Window aWindow) {
		aWindow.removeWindowListener(this);
		windows.remove(aWindow);
		StView aView = (StView) views.remove(aWindow);
		if (aView != null) {
			aView.release();
		}
	}

	/**
	 * Retrieve the Component of the indicated component.
	 *
	 * @return java.awt.Component
	 * @param aKey sra.smalltalk.StSymbol
	 * @category component handling
	 */
	public Component componentAt_(StSymbol aKey) {
		return (Component) namedComponents.get(aKey);
	}

	/**
	 * Associate aKey with aComponent.
	 *
	 * @param aKey sra.smalltalk.StSymbol
	 * @param aComponent java.awt.Component
	 * @category component handling
	 */
	public void componentAt_put_(StSymbol aKey, Component aComponent) {
		namedComponents.put(aKey, aComponent);
	}

	/**
	 * Invoked when a window is activated.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowActivated(WindowEvent e) {
	}

	/**
	 * Invoked when a window has been closed.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowClosed(WindowEvent e) {
		this.removeWindow_(e.getWindow());

		if (windows.isEmpty()) {
			this.myModel().allWindowsClosed();
		}
	}

	/**
	 * Invoked when a window is in the process of being closed.
	 * The close operation can be overridden at this point.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowClosing(WindowEvent e) {
		this.myModel().noticeOfWindowClose(e);
	}

	/**
	 * Invoked when a window is de-activated.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowDeactivated(WindowEvent e) {
	}

	/**
	 * Invoked when a window is de-iconified.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowDeiconified(WindowEvent e) {
	}

	/**
	 * Invoked when a window is iconified.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowIconified(WindowEvent e) {
	}

	/**
	 * Invoked when a window has been opened.
	 * 
	 * @param e java.awt.event.WindowEvent
	 * @category event handling
	 */
	public void windowOpened(WindowEvent e) {
	}

}
