package sra.smalltalk;

import java.awt.*;

/**
 * StViewCanvas class
 * 
 *  @author    MATSUDA Ryouichi
 *  @created   1999/11/11 (by MATSUDA Ryouichi)
 *  @updated   2002/04/12 (by nisinaka)
 *  @updated   2002/11/11 (by nisinaka)
 *  @updated   2002/11/21 (by nisinaka)
 *  @version   6.0
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: StViewCanvas.java,v 6.0 2003/10/28 10:52:32 nisinaka Exp $
 */
public abstract class StViewCanvas extends Canvas implements StView, StDisplayable {

	protected StModel model;
	protected StController controller;

	/**
	 * Create a new instance of StViewCanvas and initialize it.
	 */
	public StViewCanvas() {
		this(null);
	}

	/**
	 * Create a new instance of StViewCanvas and initialize it.
	 * 
	 * @param newModel
	 */
	public StViewCanvas(StModel newModel) {
		super();
		this.initialize();
		model = (newModel == null) ? this.defaultModel() : newModel;
		model.addDependent_(this);
		this.buildComponent();
	}

	/**
	 * Replacement for the # notation of Smalltalk. :-)
	 * 
	 * @param aString java.lang.String
	 * 
	 * @return sra.smalltalk.StSymbol
	 */
	public static final StSymbol $(String aString) {
		return StObject.$(aString);
	}

	/**
	 * Remove the receiver as a dependent of its model.
	 */
	public void release() {
		this.model().removeDependent_(this);
	}

	/**
	 * Create a Frame for the view.
	 *
	 * @return java.awt.Frame
	 * @param anApplicationModel sra.smalltalk.StApplicationModel
	 */
	public Frame _createFrame(StApplicationModel anApplicationModel) {
		return StMenuViewForAwt.CreateFrame(anApplicationModel, this);
	}

	/**
	 * Create a PopupMenu for the view.
	 *
	 * @return java.awt.PopupMenu
	 * @param aMenu sra.smalltalk.StMenu
	 */
	public PopupMenu _createPopupMenu(StMenu aMenu) {
		return StMenuViewForAwt.CreatePopupMenu(aMenu);
	}

	/**
	 * Answer the bordered panel which contains the receiver.
	 *
	 * @return java.awt.Panel
	 */
	public Panel _onBorderedPanel() {
		Panel aPanel = StApplicationModel._PanelWithBorder();
		aPanel.setLayout(new BorderLayout());
		aPanel.add(this, BorderLayout.CENTER);
		return aPanel;
	}

	/**
	 * Show the popup menu at the specified point on the view.
	 *
	 * @param x int
	 * @param y int
	 */
	public void _showPopupMenu(int x, int y) {
		PopupMenu popupMenu = this.popupMenu();
		if (popupMenu != null) {
			popupMenu.show(this, x, y);
		}
	}

	/**
	 * Make sure to have both a model and a controller before showing.
	 */
	public void addNotify() {
		if (model == null) {
			this.model_(this.defaultModel());
		}

		if (controller == null) {
			this.controller_(this.defaultController());
		}

		super.addNotify();
	}

	/**
	 * Answer the receiver as StImage.
	 *
	 * @return sra.smalltalk.StImage
	 */
	public StImage asImage() {
		int width = this.getWidth();
		int height = this.getHeight();
		StImage anImage = new StImage(width, height);
		Graphics2D aGraphics = null;
		try {
			aGraphics = (Graphics2D) anImage.image().getGraphics();
			this.displayOn_(aGraphics);
		} finally {
			if (aGraphics != null) {
				aGraphics.dispose();
				aGraphics = null;
			}
		}
		return anImage;
	}

	/**
	 * Answer my current controller.
	 * 
	 * @return sra.smalltalk.StController
	 */
	public StController controller() {
		if (controller == null) {
			this.controller_(this.defaultController());
		}

		return controller;
	}

	/**
	 * Set my new controller.
	 * 
	 * @param newController sra.smalltalk.StController
	 */
	public void controller_(StController newController) {
		if (newController != null) {
			newController.view_(this);
		}

		controller = newController;
	}

	/**
	 * Paints this component.
	 * 
	 * @param aGraphics java.awt.Graphics
	 */
	public void displayOn_(Graphics aGraphics) {
		// Do nothing as a default.
	}

	/**
	 * Display the receiver on the graphics at the specified point.
	 *
	 * @param aGraphics java.awt.Graphics
	 * @param aPoint java.awt.Point
	 */
	public void displayOn_at_(Graphics aGraphics, Point aPoint) {
		aGraphics.translate(aPoint.x, aPoint.y);
		this.displayOn_(aGraphics);
		aGraphics.translate(-aPoint.x, -aPoint.y);
	}

	/**
	 * Answer my current model.
	 * 
	 * @return sra.smalltalk.StModel
	 */
	public StModel model() {
		if (model == null) {
			this.model_(this.defaultModel());
		}

		return model;
	}

	/**
	 * Set my new model.
	 * 
	 * @param newModel sra.smalltalk.StModel
	 */
	public void model_(StModel newModel) {
		if (model != newModel) {
			if (model != null) {
				model.removeDependent_(this);
			}

			StModel oldModel = model;

			model = newModel;

			if (model != null) {
				model.addDependent_(this);
			}

			rebuildComponent(oldModel, newModel);
			repaint();
			firePropertyChange("model", oldModel, newModel);
		}
	}

	/**
	 * Paints this component.
	 * 
	 * @param g java.awt.Graphics
	 */
	public void paint(Graphics g) {
		this.displayOn_(g);
	}

	/**
	 * Answer the view as a java component.
	 *
	 * @return java.awt.Component
	 */
	public Component toComponent() {
		return (Component) this;
	}

	/**
	 * Answer the top component of the view.
	 *
	 * @return java.awt.Window
	 */
	public Window topComponent() {
		Component aComponent = this.toComponent();
		do {
			if (aComponent instanceof Window) {
				return (Window) aComponent;
			}
			aComponent = aComponent.getParent();
		} while (aComponent != null);

		return null;
	}

	/**
	 * Receive a change notice from an object of whom the receiver is a
	 * dependent.  The argument anAspectSymbol is typically a Symbol that
	 * indicates what change has occurred.
	 * 
	 * @param evt sra.smalltalk.DependentEvent
	 */
	public void update_(DependentEvent evt) {
		repaint();
	}

	/**
	 * Build this component.
	 */
	protected void buildComponent() {
		// Do nothing as a default.
	}

	/**
	 * Answer my default controller. Subclasses may override this.
	 * 
	 * @return sra.smalltalk.StController
	 */
	protected StController defaultController() {
		return new StController();
	}

	/**
	 * Answer a default model. Subclasses may override this.
	 * 
	 * @return sra.smalltalk.StModel
	 */
	protected StModel defaultModel() {
		return new StModel();
	}

	/**
	 * DOCUMENT ME!
	 * 
	 * @param propertyName DOCUMENT ME!
	 * @param oldValue DOCUMENT ME!
	 * @param newValue DOCUMENT ME!
	 */
	protected void firePropertyChange(String propertyName, Object oldValue, Object newValue) {
		// Do nothing as a default.
	}

	/**
	 * Initialize the receiver.
	 */
	protected void initialize() {
		// Do nothing as a default.
	}

	/**
	 * Answer the popup menu if exist.
	 *
	 * @return java.awt.PopupMenu
	 */
	protected PopupMenu popupMenu() {
		return null;
	}

	/**
	 * Rebuild this component.
	 * 
	 * @param oldModel sra.smalltalk.StModel
	 * @param newModel sra.smalltalk.StModel
	 */
	protected void rebuildComponent(StModel oldModel, StModel newModel) {
		// Do nothing as a default.
	}
}
