package sra.smalltalk;

import java.awt.Point;
import java.awt.Rectangle;

/**
 * SystemInterface class
 * 
 *  @author    Hoshi Takanori
 *  @created   2001/11/30 (by Hoshi Takanori)
 *  @updated   N/A
 *  @version   6.0
 *  @copyright 1999-2003 SRA (Software Research Associates, Inc.)
 *  @copyright 2001-2003 SRA/KTL (SRA Key Technology Laboratory, Inc.)
 * 
 * $Id: SystemInterface.java,v 6.0 2003/10/28 10:52:32 nisinaka Exp $
 */
public final class SystemInterface {
    
    public static final String libraryName = "StPL";
    public static final String requestVersion = "1.0";
    public static final int requestBuildNo = 5;
    
    private static SystemInterface Current = new SystemInterface();

    static {
        try {
            System.loadLibrary(libraryName);

            String findVersion = getVersion();
            String buildPrefix = requestVersion + " build ";
            int buildNo;

            try {
                buildNo = Integer.parseInt(findVersion.substring(buildPrefix.length()));
            } catch (NumberFormatException e) {
                buildNo = 0;
            }

            if (findVersion.startsWith(buildPrefix) && (buildNo >= requestBuildNo)) {
                System.out.println("JNI library '" + libraryName + "' version " + findVersion);
            } else {
                System.out.println("JNI library '" + libraryName + "' version error.");
                System.out.println("Find library is version " + findVersion);
                System.out.println("Request library is version " + requestVersion + " build " + requestBuildNo);
                throw new Error("JNI library '" + libraryName + "' version error");
            }
        } catch (UnsatisfiedLinkError e) {
            System.out.println("Can't find JNI librally '" + libraryName + "'.");
            throw e;
        }
    }

    /**
     * Constainer Method.
     */
    protected SystemInterface() {
        // Singleton pattern
        super();
    }

    /**
     * Answer the only instance of SystemInterface. Call native methods in a
     * block of synchronized. sample:
     * <pre>
     *   SystemInterface StPLInterface = SystemInterface.Current();
     *   synchronized (StPLInterface) {
     *     pixels = StPLInterface.utilImageFromUser(dimension);
     *   }
     * </pre>
     * 
     * @return sra.smalltalk.SystemInterface
     */
    public static SystemInterface Current() {
        return Current;
    }

    /**
     * Answer a rectangle from user.
     * 
     * @return java.awt.Rectangle
     */
    public static Rectangle _RectangleFromUser() {
        SystemInterface StPLInterface = SystemInterface.Current();
        int[] rect;
        synchronized (StPLInterface) {
            rect = StPLInterface.utilRectangleFromUser();
        }
    
        if (rect != null) {
            return new Rectangle(rect[0], rect[1], rect[2], rect[3]);
        } else {
            return null;
        }
    }

    /**
     * Answer the current mouse position.
     * 
     * @return java.awt.Point
     */
    public static Point _MousePoint() {
        SystemInterface StPLInterface = SystemInterface.Current();
        int[] point;
        synchronized (StPLInterface) {
            point = StPLInterface.utilMousePoint();
        }
    
        if (point != null) {
            return new Point(point[0], point[1]);
        } else {
            return null;
        }
    }

    /**
     * Set the mouse position.
     * 
     * @param x int
     * @param y int
     */
    public static void _SetMousePoint(int x, int y) {
        SystemInterface StPLInterface = SystemInterface.Current();
        synchronized (StPLInterface) {
            StPLInterface.utilSetMousePoint(x, y);
        }
    }

    /**
     * Native Methods.
     */
    public static native String getVersion();
    public native int[] utilImageFromUser(int[] dimension);
    public native int[] utilImageOfArea(int x, int y, int width, int height);
    public native int[] utilImageOfWholeScreen(int[] dimension);
    public native int[] utilRectangleFromUser();
    public native int utilColorSpuit();
    public native int[] utilMousePoint();
    public native void utilSetMousePoint(int x, int y);
    public native void utilBrowse(String url);
}
